/* dbAccess.h	*/
/* share/epicsH @(#)dbAccess.h	1.9     6/7/91 */

/*
 *      Original Author: Bob Dalesio
 *      Current Author:  Marty Kraimer
 *      Date:            6-1-90
 *
 *      Experimental Physics and Industrial Control System (EPICS)
 *
 *      Copyright 1991, the Regents of the University of California,
 *      and the University of Chicago Board of Governors.
 *
 *      This software was produced under  U.S. Government contracts:
 *      (W-7405-ENG-36) at the Los Alamos National Laboratory,
 *      and (W-31-109-ENG-38) at Argonne National Laboratory.
 *
 *      Initial development by:
 *              The Controls and Automation Group (AT-8)
 *              Ground Test Accelerator
 *              Accelerator Technology Division
 *              Los Alamos National Laboratory
 *
 *      Co-developed with
 *              The Controls and Computing Group
 *              Accelerator Systems Division
 *              Advanced Photon Source
 *              Argonne National Laboratory
 *
 * Modification Log:
 * -----------------
 * .01  mm-dd-yy        iii     Comment
 * .02  06-04-93	ems	added S_db_noRecType
 * .03  07-06-93        ems     chg DBRenumStrs, strs from 24 to 45
 */


#ifndef INCdbAccessh
#define INCdbAccessh

#include "dbAddr.h"

#ifndef INCdbFldTypesh
#include "dbFldTypes.h"
#endif

#ifndef INC_tsDefs_h
#include "tsDefs.h"
#endif
#include "db_field_log.h"

#define MAX_STRING_SIZE	80

/* event masks */
/* NOTE: these have the same value as the AGNT versions
 *	 in cadef.h, DO NOT CHNAGE THESE
 */
#ifndef	DBE_VALUE
#define DBE_VALUE       (1<<0)
#define DBE_LOG         (1<<1)
#define DBE_ALARM       (1<<2)
#endif	!DBE_VALUE

/*  The database field and request types are defined in dbFldTypes.h*/
/* Data Base Request Options	*/
#define DBR_STATUS      0x00000001
#define DBR_UNITS       0x00000002
#define DBR_PRECISION   0x00000004
#define DBR_TIME        0x00000008
#define DBR_ENUM_STRS   0x00000010
#define DBR_GR_LONG     0x00000020
#define DBR_GR_DOUBLE   0x00000040
#define DBR_CTRL_LONG   0x00000080
#define DBR_CTRL_DOUBLE 0x00000100
#define DBR_AL_LONG     0x00000200
#define DBR_AL_DOUBLE   0x00000400

/**********************************************************************
 * The next page contains macros for defining requests.
 * As an example the following defines a buffer to accept an array
 * of 10 float values + DBR_STATUS and DBR_TIME options
 *
 * struct {
 *	DBRstatus
 *	DBRtime
 *	float value[10]
 * } buffer;
 *
 * IMPORTANT!! The DBRoptions must be given in the order that they
 *             appear in the Data Base Request Options #defines
 *
 * The associated dbGetField call is:
 *
 * long options,number_elements;
 * ...
 * options = DBR_STATUS|DBR_TIME;
 * number_elements = 10;
 * rtnval=dbGetField(paddr,DBR_FLOAT,&buffer,&options,&number_elements);
 *	
 * When dbGetField returns:
 *	rtnval is error status (0 means success)
 *	options has a bit set for each option that was accepted
 *	number_elements is actual number of elements obtained
 *
 * The individual items can be refered to by the expressions::
 *
 * buffer.status
 * buffer.severity
 * buffer.err_status
 * buffer.epoch_seconds
 * buffer.nano_seconds
 * buffer.value[i]
 *
 * The following is also a valid declaration:
 *
 * typedef struct {
 *	DBRstatus
 *	DBRtime
 *	float value[10]
 * } MYBUFFER;
 *
 * With this definition you can give definitions such as the following:
 *
 * MYBUFFER *pbuf1;
 * MYBUFFER buf;
 *************************************************************************/

/* Macros for defining each option */
#define DBRstatus \
	unsigned short	status;		/* alarm status */\
	unsigned short	severity;	/* alarm severity*/\
        long            err_status;     /* error value  */
#define DBRunits \
	char		units[8];	/* units	*/
#define DBRprecision \
        long            precision;      /* number of decimal places*/\
        long            field_width;    /* field width             */
#define DBRtime \
	TS_STAMP	time;		/* time stamp*/
#define DBRenumStrs \
	unsigned long	no_str;		/* number of strings*/\
	long		padenumStrs;	/*padding to force 8 mbyte align*/\
	char		strs[64][26];	/* string values    */
#define DBRgrLong \
        long            upper_disp_limit;       /*upper limit of graph*/\
        long            lower_disp_limit;       /*lower limit of graph*/
#define DBRgrDouble \
        double          upper_disp_limit;       /*upper limit of graph*/\
        double          lower_disp_limit;       /*lower limit of graph*/
#define DBRctrlLong \
        long            upper_ctrl_limit;       /*upper limit of graph*/\
        long            lower_ctrl_limit;       /*lower limit of graph*/
#define DBRctrlDouble \
        double          upper_ctrl_limit;       /*upper limit of graph*/\
        double          lower_ctrl_limit;       /*lower limit of graph*/
#define DBRalLong \
        long            upper_alarm_limit;\
        long            upper_warning_limit;\
        long            lower_warning_limit;\
        long            lower_alarm_limit;
#define DBRalDouble \
        double          upper_alarm_limit;\
        double          upper_warning_limit;\
        double          lower_warning_limit;\
        double          lower_alarm_limit;

/* sizes for value associated with each DBR request type */
#ifdef NEED_DBR_VALUE_SIZE
static short dbr_value_size[] = {
	MAX_STRING_SIZE,		/* STRING	*/
	sizeof(char),			/* CHAR		*/
	sizeof(unsigned char),		/* UCHAR	*/
	sizeof(short),			/* SHORT	*/
	sizeof(unsigned short),		/* USHORT	*/
	sizeof(long),			/* LONG		*/
	sizeof(unsigned long),		/* ULONG	*/
	sizeof(float),			/* FLOAT	*/
	sizeof(double),			/* DOUBLE	*/
	sizeof(unsigned short)};	/* ENUM		*/
#endif NEED_DBR_VALUE_SIZE

/*  structures for each option type             */
struct dbr_status       {DBRstatus};
struct dbr_units        {DBRunits};
struct dbr_precision    {DBRprecision};
struct dbr_time         {DBRtime};
struct dbr_enumStrs     {DBRenumStrs};
struct dbr_grLong       {DBRgrLong};
struct dbr_grDouble     {DBRgrDouble};
struct dbr_ctrlLong     {DBRctrlLong};
struct dbr_ctrlDouble   {DBRctrlDouble};
struct dbr_alLong       {DBRalLong};
struct dbr_alDouble     {DBRalDouble};
/* sizes for each option structure              */
#define dbr_status_size sizeof(struct dbr_status)
#define dbr_units_size sizeof(struct dbr_units)
#define dbr_precision_size sizeof(struct dbr_precision)
#define dbr_time_size sizeof(struct dbr_time)
#define dbr_enumStrs_size sizeof(struct dbr_enumStrs)
#define dbr_grLong_size sizeof(struct dbr_grLong)
#define dbr_grDouble_size sizeof(struct dbr_grDouble)
#define dbr_ctrlLong_size sizeof(struct dbr_ctrlLong)
#define dbr_ctrlDouble_size sizeof(struct dbr_ctrlDouble)
#define dbr_alLong_size sizeof(struct dbr_alLong)
#define dbr_alDouble_size sizeof(struct dbr_alDouble)

#ifndef INCerrMdefh
#include <errMdef.h>
#endif
#define S_db_notFound 	(M_dbAccess| 1) /*Process Variable Not Found*/
#define S_db_badDbrtype	(M_dbAccess| 3) /*Illegal Database Request Type*/
#define S_db_noMod 	(M_dbAccess| 5) /*Attempt to modify noMod field*/
#define S_db_badLset 	(M_dbAccess| 7) /*Illegal Lock Set*/
#define S_db_precision 	(M_dbAccess| 9) /*get precision failed */
#define S_db_onlyOne 	(M_dbAccess|11) /*Only one element allowed*/
#define S_db_badChoice 	(M_dbAccess|13) /*Illegal choice*/
#define S_db_badField 	(M_dbAccess|15) /*Illegal field value*/
#define S_db_lsetLogic 	(M_dbAccess|17) /*Logic error generating lock sets*/
#define S_db_noRSET 	(M_dbAccess|31) /*missing record support entry table*/
#define S_db_noSupport 	(M_dbAccess|33) /*RSET routine not defined*/
#define S_db_BadSub 	(M_dbAccess|35) /*Subroutine not found*/
#define S_db_noRecType	(M_dbAccess|37) /* can't find record type */

/* Global Database Access Routines, used in ioc */
void dbScanLock (char * precord);
void dbScanUnlock (char * precord);
void dbScanLockInit (int nset);
long dbScanPassive (struct dbAddr *paddr);
long dbProcess (struct dbAddr *paddr);
long dbNameToAddr (char *pname, struct dbAddr *paddr);
long dbGetLink (struct db_link *pdblink, struct dbCommon *pdest,
	        short dbrType, char * pbuffer,
	        long *options, long *nRequest);
long dbPutField (struct dbAddr *paddr, short dbrType,
		 char * pbuffer, long nRequest);
long dbGetField (struct dbAddr *paddr, short dbrType, char * pbuffer,
		 long *options, long *nRequest, db_field_log *pfl);
long dbPutLink (struct db_link *pdblink, struct dbCommon *psource,
		short dbrType, char * pbuffer, long nRequest);
long dbBufferSize (short dbr_type, long options, long no_elements);

#endif INCdbAccessh
