/****************************************************************/
/*								*/
/*		REDUNDANT UPDATE MEMORY QUEUE MSGS		*/
/*								*/
/*	The following data structures are used to control	*/
/*	the queue between Channel Access, the various scan	*/
/*	tasks, and dbServer.					*/
/*								*/
/****************************************************************/

#include <rngLib.h>
#include <fast_lock.h>
#include <dbRecDes.h>

/******
	All people need this variable to use the Queue
******/

extern RING_ID		redundQueue;
extern FAST_LOCK	dbsrvLock;

/****************************************************************/
/*								*/
/*	These routines are for people that put messages into	*/
/*	the queue. A message is for one point but can contain	*/
/*	any number of fields.					*/
/*								*/
/****************************************************************/

int redQInit(struct dbCommon *pRecord, char *buffer,int buffSize,int recType);
/******
	This routine is called at the start of processing for each
	point. It sets up the header of the message buffer so that
	new fields can be "stuffed" into it one at a time.

	Syntax:
		redQInit(pRecord, buffer, size, type)
		struct dbCommon	*pRecord;
		char		*buffer;
		int		size;
	Arguments:
		pRecord is the address of the current record. The
		caller is responsible for allocation the message
		area whose address is passed in buffer. The total
		size of the buffer is size. The record type is passed
		in type.
	Returns:
		zero on success, -1 if any of the pointers are null,
		the size is less than or equal to zero, or if the
		pRecord parameter points to an empty record.
******/

int redQField(char *buffer, struct fldDes *pfldDes);
/******
	This routine is used to put a single field into the message
	buffer. It may be called any number of times after redQInit.
	It does NOT NOT NOT look to see if the same field has been
	queued multiple times.

	Syntax:
		redQField(buffer, pfldDes)
		char		*buffer;
		struct fldDes	*pfldDes;
	Arguments:
		The user's buffer address is in the paramter buffer.
		pfldDes is a pointer to the struct fldDes (see the
		header file dbRecDes.h) that describes the field.
		These pointers are usually found during initiali-
		zation for the record type and are then retained in
		an array internal to the record type module.
	Returns:
		zero on success, -1 if the pfldDes argument is zero
		or if there is not enough room in the buffer for
		the field. In the latter case, the routine will print
		an error message on the system console so that these
		occurrences can be discovered and then fixed.
******/

int redQSend(char *buffer);
/******
	This routine is used to put the user's buffer into the queue.
	It is called once per point, at the end of processing.

	Syntax:
		redQSend(buffer)
		char	*buffer;
	Arguments:
		The address of the user's buffer is passed in buffer.
	Returns:
		zero on success, -1 if the pointer to the user's buffer
		is null. Note that this routine will not wait for
		space to become available in the queue. If the queue
		is full, then the routine will set the "Full update
		required next pass" flag in the record and return
		a success code. If the "Full update required next
		pass flag is set when this routine is called, then
		any field changes that the user may have previously
		enqueued this pass will be discarded and all fields
		for the point that are required for redundant update
		will be enqueued. When a message that has been filled
		for all redundant fields has been successfully enqueued
		off to dbServer, then the "Full update required next
		pass" flag will be cleard by this routine.
******/


void redQPost (void);
/******

	This routine is to be called at the end of a list
	by any one of the scan tasks. If the message Queue
	isn't empty, then post the semaphore to make dbServer
	process the queue.

******/

/****************************************************************/
/*								*/
/*		REDUNDANT UPDATE QUEUE BUFFER MANAGEMENT	*/
/*								*/
/*	The buffer that is managed by the above routines	*/
/*	starts with a struct redQBuffHdr. The first two		*/
/*	fields are used to control the buffer itself.		*/
/*	They contain the total size of the buffer and the	*/
/*	current number of bytes that have been stuffed into	*/
/*	the buffer up to the current time. The point type	*/
/*	is also recorded in the buffer in case these libary	*/
/*	routines decide that they want to update all redundant	*/
/*	fields just before sending the message.			*/
/*								*/
/*	The start of the message that is actually sent to 	*/
/*	dbServer is the point type of the database record	*/
/*	itself. For each field that is contained in the		*/
/*	record, there is a pointer to the struct fldDes		*/
/*	followed by the new value for that field. The		*/
/*	number of bytes for the new value depends on the	*/
/* 	field, but is contained in the struct fldDes.		*/
/*								*/
/****************************************************************/
#define	MC68000	1
#if CPU == MC68000
#define REDQ_KBYTES	8
#else
#define REDQ_KBYTES	32
#endif
#define REDQ_MAX_MSGLEN	896

struct redQBuffHdr
	{
	int		totalSize;
	int		currentSize;
	short		pointType;
	short		promptFlag;	/* Number of prompted fields */
	int		numFields;
	struct dbCommon	*recordAddr;
	};


/*
	The following structure is used by dbServer. It has to be
	the same as redQBuffHdr without the totalSize member. Be
	careful to change this one if you change redQBuffHdr.
*/

struct dbServerQHdr
	{
	short		pointType;
	short		promptFlag;
	int		numFields;
	struct dbCommon	*recordAddr;
	};
