/****
*
*	File name :		q4task.c
*	Author :		Steve Johnson
*	Creation date :	December 29, 1998
*
*	Copyright (c) 1998,1999 Sigma Six Corporation
*
*	Provides the following entry points:
*
*		quad4_init
*		quad4_exit
*		quad4_add_meter
*		quad4_add_mv90
*		quad4_table_addr
*		quad4_loop
*		quad4_stop
*
*	Siemens Meter Quad 4 Protocol Top Level Task Driver
*
****/

#include <sscommon.h>
#include <sstrace.h>
#include <ssqueues.h>
#include <ssnames.h>
#include <sslib.h>
#include "q4task.h"
#include "meter.h"
#include "mv90.h"

#define VERSION			"Version 1.35e, 1:45pm 7/28/1999"
#define COPYRIGHT		"Copyright (c) 1998, 1999 Hathaway Industrial Automation"

#define METER_TIMEOUT		15000		/* 15 seconds */

typedef enum {
	CLASS_PUBLISHED = 1,
	CLASS_ACCEPTED = 2,
} MV90_CLASS;

typedef struct {
	QUEUE_DEF		link;
	METER_HANDLE	lower_handle;
	void			*mv90;
	long			last_command_time;
	long			meter_timeout;
	void			*active_mv90;
	long			unit_id;
	char			name [MAX_STRING];
} METER_DEF;

typedef struct {
	QUEUE_DEF		link;
	char			name [MAX_STRING];
} NEW_METER_DEF;

typedef struct {
	QUEUE_DEF		link;
	MV90_CLASS		mv90_class;
	MV90_HANDLE		lower_handle;
	METER_DEF		*meter;
	char			name [MAX_STRING];
} MV90_DEF;

static char			user_input [MAX_STRING];
static char			last_line [MAX_STRING];
static int			input_index;
static int			looping;
static int			special_key;
static int			quad4_task_id;
static int			read_buffer_size;
static UCHAR		*read_buffer;
static QUEUE_DEF	root_meter;
static QUEUE_DEF	root_new_meter;
static QUEUE_DEF	root_mv90;

static void			close_all (void);
static int			publish_meter (char *);
static METER_DEF	*add_meter (char *);
static int			connect_meter (METER_DEF *);
static MV90_DEF		*add_mv90 (MV90_HANDLE, MV90_CLASS, char *);
static void			delete_mv90 (MV90_DEF *);
static int			check_data (void);
static int			check_mv90 (MV90_DEF **, size_t *);
static int			check_meter (METER_DEF **, size_t *);
static int			scan_key (void);
static int			parse_user (char *);
static METER_DEF	*find_meter_by_handle (METER_HANDLE);
static MV90_DEF		*find_mv90_by_handle (MV90_HANDLE);

int quad4_init (void)
{
	int			status;

	printf ("\n\nQuad 4 Meter Driver Initialization\n");
	printf ("-- %s\n", VERSION);
	printf ("-- %s\n\n", COPYRIGHT);
	quad4_task_id = GET_TASK_ID ();
	TRACE ("quad4_init, begin");
	ERROR_TRACE (VERSION);
	ERROR_TRACE (COPYRIGHT);
	queue_init (&root_meter);
	queue_init (&root_new_meter);
	queue_init (&root_mv90);
	read_buffer_size = (size_t) ini_get_long (QUAD4_MAIN_GROUP
											, "MAX_BUFFER_SIZE"
											, METER_MAX_BUFFER_SIZE);
	read_buffer = (UCHAR *) malloc (read_buffer_size);
	if (!read_buffer) {
		ERROR_TRACE_I ("quad4_init, malloc size = %d", read_buffer_size);
		return FALSE;
	}
	status = meter_init ();
	if (METER_NORMAL != status) {
		printf ("quad4_init failed\n");
		free (read_buffer);
		return FALSE;
	}
	return TRUE;
}

int quad4_add_meter (char *meter_name)
{
	NEW_METER_DEF		*new_meter;

	new_meter = (NEW_METER_DEF *) malloc (sizeof (NEW_METER_DEF));
	if (!new_meter) {
		printf ("quad4_add_meter, malloc failed\n");
		return FALSE;
	}
	strcpy (new_meter -> name, meter_name);
	queue_append (&root_new_meter, &new_meter -> link);
	return TRUE;
}

int quad4_loop (void)
{
	NEW_METER_DEF		*new_meter;

	TRACE ("quad4_loop, entry point");
	looping = TRUE;
	special_key = FALSE;
	while (looping) {
		if (!check_data ()) {
			close_all ();
			return FALSE;
		}
		if (!scan_key ()) {
			close_all ();
			return FALSE;
		}
		new_meter = queue_get (&root_new_meter);
		if (new_meter) {
			TRACE_S ("quad4_add_meter, meter_name = %s", new_meter -> name);
			if (!publish_meter (new_meter -> name)) {
				ERROR_TRACE ("quad4_loop, publish_meter failed");
			}
			free (new_meter);
		}
		taskDelay (15);
	}
	close_all ();
	return TRUE;
}

void quad4_exit (void)
{
	close_all ();
}

void quad4_stop (void)
{
	looping = FALSE;
}

UCHAR *quad4_table_addr (USHORT port, USHORT table_number)
{
	return meter_table_addr (port, table_number);
}

void quad4_debug_on (void)
{
	TRACE_HANDLE	handle;

	meter_debug_on ();
	handle = sslib_get_tracehandle (quad4_task_id);
	trace_debug_on (handle);
}

void quad4_debug_off (void)
{
	TRACE_HANDLE	handle;

	meter_debug_off ();
	handle = sslib_get_tracehandle (quad4_task_id);
	trace_debug_off (handle);
}

static int check_data (void)
{
	size_t			bytes_ready;
	MV90_DEF		*mv90;
	METER_DEF		*meter;
	METER_STATUS	meter_status;

	if (check_mv90 (&mv90, &bytes_ready)) {
		meter = mv90 -> meter;
		if (!meter) {
			ERROR_TRACE ("check_data, no meter ???");
			return TRUE;
		}
		if (!meter -> active_mv90) {
			meter -> active_mv90 = mv90;
			meter_status = meter_lock (meter -> lower_handle);
			if (METER_NORMAL != meter_status) {
				ERROR_TRACE ("check_data, unable to lock meter");
				meter -> active_mv90 = NULL;
				return TRUE;
			}
		}
		if (mv90 == meter -> active_mv90) {
			if (!meter -> lower_handle) {
				if (!connect_meter (meter)) {
					ERROR_TRACE ("check_data, connect_meter failed");
					return TRUE;
				}
			}
			meter_status = meter_write (meter -> lower_handle
									  , bytes_ready
									  , read_buffer);
			if (METER_NORMAL != meter_status) {
				ERROR_TRACE_I ("check_data, meter_write = %d", meter_status);
				meter_shutdown (meter -> lower_handle);
				meter -> lower_handle = (METER_HANDLE) NULL;
				meter -> unit_id = 0;
				return TRUE;
			}
			meter -> last_command_time = GetTickCount ();
		}
	}
	meter = (METER_DEF *) queue_next (&root_meter);
	while (meter) {
		if (meter -> active_mv90) {
			if ((meter -> last_command_time + meter -> meter_timeout)
			< (long) GetTickCount ()) {
				ERROR_TRACE ("check_data, mv90 shutdown by inactivity");
				mv90 = (MV90_DEF *) meter -> active_mv90;
				meter_unlock (meter -> lower_handle);
				meter -> active_mv90 = NULL;
				mv90_shutdown (mv90 -> lower_handle);
				delete_mv90 (mv90);
			}
		}
		meter = (METER_DEF *) queue_next (&meter -> link);
	}
	while (check_meter (&meter, &bytes_ready)) {
		if (!meter -> active_mv90) {
			ERROR_TRACE ("check_data, mv90 shutdown in midstream");
			return TRUE;
		}
		mv90 = (MV90_DEF *) meter -> active_mv90;
		meter_status = mv90_write (mv90 -> lower_handle
								  , bytes_ready
								  , read_buffer);
		if (METER_NORMAL != meter_status) {
			ERROR_TRACE_I ("check_data, mv90_write = %d", meter_status);
			meter -> active_mv90 = NULL;
			meter_unlock (meter -> lower_handle);
			mv90_shutdown (mv90 -> lower_handle);
			delete_mv90 (mv90);
			return TRUE;
		}
	}
	return TRUE;							
}

static int check_mv90 (MV90_DEF **return_mv90, size_t *bytes_ready)
{
	METER_DEF		*meter;
	MV90_DEF		*mv90;
	MV90_DEF		*new_mv90;
	MV90_STATUS		status;
	MV90_HANDLE		handle;
	MV90_HANDLE		new_handle;

	status = mv90_poll (&handle, bytes_ready);
	if (MV90_NO_DATA == status) {
		*bytes_ready = 0;
		return FALSE;
	} else if (MV90_NORMAL != status) {
		ERROR_TRACE_I ("check_mv90, mv90_poll = %d", status);
		*bytes_ready = 0;
		return FALSE;
	}
	mv90 = find_mv90_by_handle (handle);
	if (!mv90) {
		ERROR_TRACE ("check_mv90, unknown handle");
		*bytes_ready = 0;
		return FALSE;
	}
	if (CLASS_PUBLISHED == mv90 -> mv90_class) {
		status = mv90_accept (&new_handle, handle);
		if (MV90_NORMAL != status) {
			ERROR_TRACE_I ("check_mv90, mv90_accept = %d", status);
			*bytes_ready = 0;
			return FALSE;
		}
		new_mv90 = add_mv90 (new_handle, CLASS_ACCEPTED, mv90 -> name);
		if (!new_mv90) {
			ERROR_TRACE ("check_mv90, add_mv90 failed");
			mv90_shutdown (new_handle);
			return FALSE;
		}
		meter = mv90 -> meter;
		new_mv90 -> meter = meter;
		return FALSE;
	}
	status = mv90_read (handle, read_buffer_size, read_buffer, bytes_ready);
	if (MV90_NORMAL != status) {
		TRACE_I ("check_mv90, mv90_read = %d", status);
		meter = mv90 -> meter;
		if (mv90 == meter -> active_mv90) {
			meter_unlock (meter -> lower_handle);
			meter -> active_mv90 = NULL;
		}
		mv90_shutdown (handle);
		delete_mv90 (mv90);
		*bytes_ready = 0;
		return FALSE;
	}
	*return_mv90 = mv90;
	return TRUE;
}

static int check_meter (METER_DEF **meter, size_t *bytes_ready)
{
	METER_STATUS	status;
	METER_HANDLE	meter_handle;

	status = meter_poll (&meter_handle);
	if (METER_NO_DATA == status) {
		*bytes_ready = 0;
		return FALSE;
	} else if (METER_NORMAL != status) {
		ERROR_TRACE_I ("check_meter, meter_poll = %d", status);
		*bytes_ready = 0;
		return FALSE;
	}
	*meter = find_meter_by_handle (meter_handle);
	if (!*meter) {
		ERROR_TRACE ("check_meter, unknown handle");
		meter_shutdown (meter_handle);
		*bytes_ready = 0;
		return FALSE;
	}
	status = meter_read (meter_handle
					   , read_buffer_size
					   , read_buffer
					   , bytes_ready);
	if (METER_NORMAL != status) {
		ERROR_TRACE_I ("check_meter, meter_read = %d", status);
		meter_shutdown (meter_handle);
		(*meter) -> lower_handle = (METER_HANDLE) NULL;
		(*meter) -> unit_id = 0;
		*bytes_ready = 0;
		return FALSE;
	}
	return TRUE;
}

static int publish_meter (char *meter_name)
{
	int			status;
	METER_DEF	*meter;
	MV90_DEF	*mv90;
	MV90_HANDLE	mv90_handle;
	char		public_name [MAX_STRING];
	char		private_name [MAX_STRING];

	TRACE ("publish_meter, entry point");
	if (!ini_get_str (meter_name, "PRIVATE_NAME", private_name)) {
		ERROR_TRACE_S ("publish_meter, no PRIVATE_NAME for (%s)", meter_name);
		return FALSE;
	}
	if (ini_get_int (meter_name, "SLAVE_RIG", 1)) {
		if (!ini_get_str (meter_name, "PUBLIC_NAME", public_name)) {
			ERROR_TRACE_S ("publish_meter, no PUBLIC_NAME for (%s)", meter_name);
			return FALSE;
		}
		status = mv90_publish (&mv90_handle, public_name);
		if (METER_NORMAL != status) {
			ERROR_TRACE ("publish_meter, mv90_publish failed");
			return FALSE;
		}
		mv90 = add_mv90 (mv90_handle, CLASS_PUBLISHED, private_name);
		if (!mv90) {
			ERROR_TRACE ("publish_meter, add_mv90 failed");
			mv90_shutdown (mv90_handle);
			return FALSE;
		}
		meter = add_meter (private_name);
		if (!meter) {
			ERROR_TRACE ("publish_meter, add_meter failed");
			mv90_shutdown (mv90_handle);
			return FALSE;
		}
		meter -> mv90 = mv90;
		mv90 -> meter = meter;
	} else {
		meter = add_meter (private_name);
		if (!meter) {
			ERROR_TRACE ("publish_meter, add_meter failed");
			return FALSE;
		}
		meter -> mv90 = NULL;
	}
	return TRUE;
}

static METER_DEF *add_meter (char *meter_name)
{
	METER_DEF	*meter;

	TRACE ("add_meter, entry point");
	meter = (METER_DEF *) malloc (sizeof (METER_DEF));
	if (!meter) {
		ERROR_TRACE_I ("add_meter, malloc size %d", sizeof (METER_DEF));
		return (METER_DEF *) NULL;
	}
	strcpy (meter -> name, meter_name);
	meter -> last_command_time = 0;
	meter -> active_mv90 = NULL;
	meter -> meter_timeout = METER_TIMEOUT;
	queue_append (&root_meter, &meter -> link);
	if (!connect_meter (meter)) {
		ERROR_TRACE ("add_meter, connect_meter failed");
	}
	return meter;
}

static int connect_meter (METER_DEF *meter)
{
	int			status;

	TRACE ("connect_meter, entry point");
	status = meter_connect (&meter -> lower_handle, meter -> name);
	if (METER_NORMAL != status) {
		ERROR_TRACE_I ("connect_meter, meter_connect = %d", status);
		printf ("Quad4Task, meter (%s) connect failed", meter -> name);
		meter -> lower_handle = (METER_HANDLE) NULL;
		meter -> unit_id = 0;
		return FALSE;
	}
	taskDelay (5);
	status = meter_get_id (meter -> lower_handle, &meter -> unit_id);
	if (METER_NORMAL != status) {
		ERROR_TRACE_I ("connect_meter, meter_get_id = %d", status);
		meter_shutdown (meter -> lower_handle);
		meter -> lower_handle = (METER_HANDLE) NULL;
		meter -> unit_id = 0;
		return FALSE;
	}
	printf ("Meter (%s) added\n", meter -> name);
	return TRUE;
}

/****
*	never done dynamically
*	static void delete_meter (METER_DEF *meter)
*	{
*		TRACE ("delete_meter, entry point");
*		queue_remove (&root_meter, &meter -> link);
*		free (meter);
*	}
****/

static MV90_DEF *add_mv90 (MV90_HANDLE	handle
						 , MV90_CLASS	mv90_class
						 , char			*private_name)
{
	MV90_DEF	*mv90;

	TRACE ("add_mv90, entry point");
	mv90 = (MV90_DEF *) malloc (sizeof (MV90_DEF));
	if (!mv90) {
		ERROR_TRACE_I ("add_mv90, malloc size %d", sizeof (MV90_DEF));
		return (MV90_DEF *) NULL;
	}
	mv90 -> lower_handle = handle;
	mv90 -> mv90_class = mv90_class;
	strcpy (mv90 -> name, private_name);
	queue_append (&root_mv90, &mv90 -> link);
	return mv90;
}

static void delete_mv90 (MV90_DEF *mv90)
{
	TRACE ("delete_mv90, entry point");
	queue_remove (&root_mv90, &mv90 -> link);
	free (mv90);
}

static int scan_key (void)
{
	UCHAR			key_char;

	if (kbhit ()) {
		key_char = (UCHAR) getch ();
		if (0x1B == key_char) {
			return FALSE;
		} else if ((0x0A == key_char) || (0x0D == key_char)) {
			user_input [input_index] = 0;
			strcpy (last_line, user_input);
			if (!parse_user (user_input)) {
				input_index = 0;
				return FALSE;
			}
			input_index = 0;
		} else if (0x08 == key_char) {
			if (input_index) {
				printf ("%c", key_char); fflush (stdout);
				printf (" "); fflush (stdout);
				printf ("%c", key_char); fflush (stdout);
				user_input [input_index] = 0;
				--input_index;
			}
		} else if (0 == key_char) {
			special_key = TRUE;
		} else {
			if (special_key) {
				special_key = FALSE;
				if (0x48 == key_char) {
					strcpy (user_input, last_line);
					printf ("\nCommand:%s", user_input); fflush (stdout);
					input_index = strlen (user_input);
				}
			} else {
				if (!input_index) {
					printf ("\nCommand:"); fflush (stdout);
				}
				if (MAX_STRING >= input_index) {
					printf ("%c", key_char); fflush (stdout);
					user_input [input_index] = key_char;
					++input_index;
				}
			}
		}
	}
	return TRUE;
}

static void close_all (void)
{
	int				status;
	METER_DEF		*meter;
	MV90_DEF		*mv90;

	TRACE ("close_all, entry point");
	TRACE ("close_all, close meter connections");
	while (queue_next (&root_meter)) {
		meter = (METER_DEF *) queue_get (&root_meter);
		status = meter_shutdown (meter -> lower_handle);
		if (METER_NORMAL != status) {
			ERROR_TRACE_I ("close_all, meter_shutdown = %d", status);
		}
		free (meter);
	}
	TRACE ("close_all, close mv90 connections");
	while (queue_next (&root_mv90)) {
		mv90 = (MV90_DEF *) queue_get (&root_mv90);
		status = mv90_shutdown (mv90 -> lower_handle);
		if (METER_NORMAL != status) {
			ERROR_TRACE_I ("close_all, mv90_shutdown = %d", status);
		}
		free (mv90);
	}
	mv90_exit ();
	meter_exit ();
	free (read_buffer);
}

/****
*
*	obsolete for now
*
*	static METER_DEF *find_meter_by_name (char *name)
*	{
*		METER_DEF	*meter;
*	
*		meter = (METER_DEF *) queue_next (&root_meter);
*		while (meter) {
*			if (case_compare (meter -> name, name)) {
*				return meter;
*			}
*			meter = (METER_DEF *) queue_next (&meter -> link);
*		}
*		return (METER_DEF *) NULL;
*	}
*
****/

static METER_DEF *find_meter_by_handle (METER_HANDLE handle)
{
	METER_DEF	*meter;

	meter = (METER_DEF *) queue_next (&root_meter);
	while (meter) {
		if (handle == meter -> lower_handle) {
			return meter;
		}
		meter = (METER_DEF *) queue_next (&meter -> link);
	}
	return (METER_DEF *) NULL;
}

static MV90_DEF *find_mv90_by_handle (MV90_HANDLE handle)
{
	MV90_DEF	*mv90;

	mv90 = (MV90_DEF *) queue_next (&root_mv90);
	while (mv90) {
		if (handle == mv90 -> lower_handle) {
			return mv90;
		}
		mv90 = (MV90_DEF *) queue_next (&mv90 -> link);
	}
	return (MV90_DEF *) NULL;
}

static int parse_user (char *user_input)
{
	METER_STATUS	status;
	int				table_number;
	size_t			table_size;
	METER_DEF		*meter;

	if ('R' == toupper (user_input [0])) {
		table_number = atoi (&user_input [2]);
		meter = (METER_DEF *) queue_next (&root_meter);
		status = meter_read_table (meter -> lower_handle
								 , table_number
								 , read_buffer_size
								 , read_buffer
								 , &table_size);
		if (METER_NORMAL != status) {
			printf ("meter_read_table, failed status = %d", status);
			return FALSE;
		}
		printf ("meter_read_table, table_size = %d", table_size);
	} else {
		printf ("unknown command (%s)\n", user_input);
	}
	return TRUE;
}

