#include <vxWorks.h>
#include <stdio.h>
#include <BootLib.h>
#include <logLib.h>
#include <HostLib.h>
#include <string.h>
#include <netDrv.h>
#include <inetLib.h>
#include <ioLib.h>
#include <usrLib.h>

#define PARAM_STRING_LOCATION (char*)0x1200
#define BOOT_STRING_FILE "/ata0/bootstrg.dat"
#define FD_BOOT_STNG_FILE "/fd0/bootstrg.dat"

#define FTP 1

#define HD 0
#define FD 1

static STATUS InitialiseNetwork(BOOT_PARAMS *params_struct);
static void CrackBootLine(char *params_string, BOOT_PARAMS *params_struct);
static char *LoadBootLine(void);

STATUS boot(void)
{
   BOOT_PARAMS params_struct;
   char *params_string;

   /* Get the boot string */
   params_string = LoadBootLine();
   if (params_string == NULL)
      return (-1);

   /* Convert to a structure */
   CrackBootLine(params_string, &params_struct);

   /* Network initialisation based on the boot line. */
   InitialiseNetwork(&params_struct);

   return OK;
}

/*
   Function    InitialiseNetwork
   Parameters  Boot structure
   returns     STATUS = OK / ERROR
   Comments
*/
STATUS InitialiseNetwork(BOOT_PARAMS *params_struct)
{
   STATUS  status = OK;
   char host_device[21];
   UINT32 netmask;

   status = bootNetmaskExtract (params_struct->ead, &netmask);
   switch (status)
   {
      case 1:
         /* Subnet mask correct */
         status = OK;
         break;
      case 0:
         /* No subnet mask specified */
         logMsg("No subnet mask specified for target IP\n",0,0,0,0,0,0);
         /* Set a reasonable class C default */
         netmask = 0xffffff00;
         break;
      case -1:
         logMsg("Format of subnet for targer IP is incorrect\n",0,0,0,0,0,0);
         netmask = 0xffffff00;
         break;
      default:
         logMsg("Unexpected value %d returned from bootNetmaskExtract\n", status,0,0,0,0,0);
         netmask = 0xffffff00;
         status = ERROR;
         break;
   }
   usrNetIfAttach (params_struct->other, params_struct->ead);
   usrNetIfConfig (params_struct->other, params_struct->ead, 
                       params_struct->targetName, netmask);
   hostAdd (params_struct->hostName, params_struct->had);

   if (strlen(params_struct->gad) > 6)
      routeNetAdd("0.0.0.0", params_struct->gad);

   strcpy (host_device, params_struct->hostName);
   strcat (host_device, ":");
   netDevCreate (host_device, params_struct->hostName, FTP);
   iam (params_struct->usr, params_struct->passwd);

   sethostname(params_struct->targetName, strlen(params_struct->targetName));

   return status;
} 

/*
   Function    CrackBootLine
   Parameters  Pointer to the boot string to crack
               Pointer to the boot structure in which to store the result
   returns     void
   Comments    This function attempts to crack the boot line, and prompts for the
               boot parameters to be changed if the boot line is corrupt.
*/
void CrackBootLine(char *params_string, BOOT_PARAMS *params_struct)
{
   /* Crack the bootline open */
   while (*bootStringToStruct (params_string, params_struct) != EOS)
   {
      /* Error. BootString could not be fully parsed. */
      logMsg("Boot-line could not be fully parsed\n",0,0,0,0,0,0);
      /*
         Since the boot line cannot be parsed, assume that it is corrupt.
         To recover, remove the boot line file, and call LoadBootLine. This
         will ensure that bootChange is called and the user is prompted for
         a new boot line.
      */
      rm (BOOT_STRING_FILE);
      createBoot(0);
      params_string = LoadBootLine();
   }      
}


/*
   Function    LoadBootLine
   Parameters  None
   Returns     Pointer to the boot line
   Comments    This function attempts to load a default boot line from
               a file on the flash drive. If the file does not exist, a
               bootline is constructed by calling the bootChange() function.
               The resulting bootline is stored on the flash drive for
               future use.
*/
char *LoadBootLine(void)
{
   FILE *fp;
   char *params_string = PARAM_STRING_LOCATION;

   fp = fopen(BOOT_STRING_FILE, "r");

   /* Put boot params at location PARAM_STRING_LOCATION */
   if (fp == NULL)
   {
      /* No file available... Use bootChange() */
      logMsg("File bootstrg.dat not found\nTo create run createBoot()\n", 0, 0, 0, 0, 0, 0);
      logout();
      taskDelay(sysClkRateGet() * 3);
      return (NULL);
   }

   else
   {
      /* Load bootline from file */
      do
      {
         if (fread(params_string, 1, 1, fp) != 1)
	 {
	    /* add termination to string */
	    *params_string = 0;
	 }
         params_string++;
      }while (*(params_string - 1) != NULL);
      /* No need to re-write it */

      /* Point params_string back at start of boot line */
      params_string = PARAM_STRING_LOCATION;
   }
   fclose(fp);
    
   return params_string;
}

STATUS createBoot(int dest)
{
   FILE *fp;
   char *params_string = PARAM_STRING_LOCATION;
   int  console_fd, old_fd;

   /* Take a copy of stdin */
   old_fd = ioGlobalStdGet (STD_IN);

   /* Open the console (keyboard) and redirect STD_IN */
   if ((console_fd = open ("/pcConsole/0", O_RDONLY, 0)) == ERROR)
   {
      if ((console_fd = open ("/tyCo/0", O_RDONLY, 0)) == ERROR)
      {
	 perror ("Cannot open console for input");
	 return ERROR;
      }
   }
   ioGlobalStdSet (STD_IN, console_fd);

   bootChange();

   /* Restore the old STD_IN */
   ioGlobalStdSet (STD_IN, old_fd);
   close (console_fd);

   if (dest)
      fp = fopen(FD_BOOT_STNG_FILE, "w");
   else
      fp = fopen(BOOT_STRING_FILE, "w");

   if (fp == NULL)
   {
      /* Error opening for write */
      logMsg("Could not open boot line file to store boot parameters\n",0,0,0,0,0,0);
      return (-1);
   }
   else
   {
      fwrite (params_string, strlen(params_string)+1, 1, fp);
      fclose (fp);
   }
   boot();

   return OK;
}

/*
 * Function:      GetCertFiles
 * Purpose:       Retreives cert files from CA based on IOC name taken
 *                from the bootline parameters
 * Arguments:     IOC device name.
 * Returns:       STATUS... OK or ERROR
 * Preconditions: The necessary calls to hostAdd and netDevCreate have
 *                been made.
 * Environment:   Environment variables must be set for CA_PATH, CA_FILE, 
 *                KEY_FILE, PEM_EXT, 
 */
STATUS GetCertFiles(char *CA_name)
{
    char        *param_ptr;
    BOOT_PARAMS param_struct;
    char        source[64];
    char        dest[64];
    
    param_ptr = LoadBootLine();
    if (param_ptr == NULL)
      return ERROR;
    if ((getenv("CA_CERT_PATH") == NULL) ||
	(getenv("PEM_EXT") == NULL) ||
	(getenv("CA_PATH") == NULL) ||
	(getenv("CERT_FILE") == NULL) ||
	(getenv("CA_KEY_PATH") == NULL) ||
	(getenv("KEY_FILE") == NULL))
    {
	logMsg("The following environment variables must all be set\n%s%s",
	       (int)"CA_CERT_PATH, PEM_EXT, CA_PATH, CERT_FILE",
	       (int)"CA_KEY_PATH, KEY_FILE", 0, 0, 0, 0);
	return ERROR;
    }
	
    CrackBootLine(param_ptr, &param_struct);
    
    /* Set up source path for cert */
    strcpy (source, CA_name);
    strcat (source, getenv("CA_CERT_PATH"));
    strcat (source, param_struct.hostName);
    strcat (source, getenv("PEM_EXT"));
    /* and the destination */
    strcpy (dest, getenv("CA_PATH"));
    strcpy (dest, getenv("CERT_FILE"));
    if (copy (source, dest) == ERROR)
      return ERROR;
    
    /* Set up source path for key */
    strcpy (source, CA_name);
    strcat (source, getenv("CA_KEY_PATH"));
    strcat (source, param_struct.hostName);
    strcat (source, getenv("PEM_EXT"));
    /* and the destination */
    strcpy (dest, getenv("CA_PATH"));
    strcat (dest, getenv("KEY_FILE"));
    if (copy (source, dest) == ERROR)
      return ERROR;
    
    return OK;
}
