#include <vxWorks.h>
#include <dirent.h>
#include <logLib.h>
#include <usrLib.h>
#include <string.h>
#include <hostLib.h>

#include "SimplifyDir.h"

static int debug_level;


/* Static prototypes */
static STATUS MoveCerts (char *from, char *to);
static STATUS MoveDataBase (char *from, char *to);
static STATUS MoveScripts (char *old_base, char *new_dir);
static STATUS MoveBinaries (char *from, char *to);
static STATUS MkDirIfAbsent (char *dir);
static STATUS Lmove(char *from, char *to);

/*
 * Function:      SimplifyDirectory
 * Purpose:       To rearange the directory structure of an IOC
 * Arguments;     void
 * Returns:       STATUS
 * Preconditions: Old directory structure in place
 */
STATUS SimplifyDirectory(int verbosity)
{
    STATUS status = OK;

    debug_level = verbosity;

    /* Change to root directory to prevent deletion of current directory */
    chdir (ROOT_DIR);

    if (MoveBinaries(OLD_BINARIES, NEW_BINARIES) == ERROR)
	status = ERROR;

    if (MoveScripts(OLD_SCRIPT_BASE, NEW_SCRIPTS) == ERROR)
	status = ERROR;

    if (MoveDataBase(OLD_DB, NEW_DB) == ERROR)
	status = ERROR;

    if (MoveCerts(OLD_CERTS, NEW_CERTS) == ERROR)
	status = ERROR;

    /* If all is well change to scripts directory */
    if (status == OK)
	chdir (NEW_SCRIPTS);

    return status;
}

/*
 * Function:      MoveBinaries
 * Purpose:       To move binary files from one directory to another
 * Arguments:     source and destination directories
 * Returns:       STATUS
 * Preconditions: Binaries must be installed in the source directory and parent of destination must exist
 */
static STATUS MoveBinaries (char *from, char *to)
{
    if (MoveFiles (from, to, "binary") == ERROR)
	return ERROR;

    if (CleanDir (BINARY_BASE_DIR) == ERROR)
	return ERROR;

    if (rm (BINARY_BASE_DIR) == ERROR)
	return ERROR;

    return OK;
}

/*
 * Function:      MoveScripts
 * Purpose:       To move script files from one directory to another
 * Arguments:     Source base directory... The source directory is named after the host name with
 *                the source base directory as its parent. and destination directory
 * Returns:       STATUS OK for success or ERROR for (serious) errors
 * Preconditions: Script files must be in the directory outlined above. Parent of destination directory
 *                must exist. Host name must be set correctly.
 */
static STATUS MoveScripts (char *old_base, char *new_dir)
{
    char old_dir[FQN_SIZE];
    char host_name[32] = "";

    /* Get the machine name */
    errno = 0;
    if (((gethostname (host_name, 32) == ERROR) && (errno)) || (!strlen (host_name)))
    {
	logMsg ("Machine host name is not set. ERROR = %s\n", (int)strerror(errno), 0, 0, 0, 0, 0);
	logMsg ("I need this to create the source directory for script files\n", 0, 0, 0, 0, 0, 0);
	return ERROR;
    }
    else
    {
	if (strlen (host_name) > 8)
	    host_name[8] = NULL;
    }

    /* Create old directory name */
    strncpy (old_dir, old_base, FQN_SIZE - 9);
    strcat (old_dir, host_name);
    strcat (old_dir, "/");

    /* Move the files */
    if (MoveFiles (old_dir, new_dir, "script") == ERROR)
	return ERROR;

    if (CleanDir (old_dir) == ERROR)
	return ERROR;

    if (rm (old_dir) == ERROR)
	return ERROR;

    return OK;
}

/*
 * Function:      MoveDataBase
 * Purpose:       Stub function since data base files will not be moved
 * Arguments:     Source and destination directories
 * Returns:       OK
 * Preconditions: none
 */
static STATUS MoveDataBase (char *from, char *to)
{
    return OK;
}

/*
 * Function:      MoveCerts
 * Purpose:       To move cert files from one directory to another
 * Arguments:     source and destination directories
 * Returns:       STATUS OK or ERROR
 * Preconditions: As per MoveFiles
 */
static STATUS MoveCerts (char *from, char *to)
{
    if (MoveFiles (from, to, "certificate") == ERROR)
	return ERROR;

    if (CleanDir (SSL_BASE_DIR) == ERROR)
	return ERROR;

    if (rm (SSL_BASE_DIR) == ERROR)
	return ERROR;

    return OK;
}

/*
 * Function:      MoveFiles
 * Purpose:       To move executable files from one directory to another
 * Arguments:     Source and destination directories and type of file for debug messages
 * Returns:       STATUS OK on success or error
 * Preconditions: Files should be in the source dir and the parent of the destination
 *                directory should exist
 */
STATUS MoveFiles(char *old_dir, char *new_dir, char *file_type)
{
    struct dirent *dent;
    char source[FQN_SIZE], dest[FQN_SIZE];
    DIR *dir_ptr;

    /* Open the source directory */
    dir_ptr = opendir (old_dir);
    if (dir_ptr == NULL)
    {
	logMsg ("Could not open source dir for %s files\n", (int)file_type, 0, 0, 0, 0, 0);
        return ERROR;
    }

    /* Make sure that the destination directory exists and is empty */
    if (MkDirIfAbsent (new_dir) == ERROR)
    {
	closedir (dir_ptr);
	return ERROR;
    }

    /* Copy the files */
    while ((dent = readdir (dir_ptr)) != NULL)
    {
	/* Suppress copy / delete of . and .. */
	if (strcmp (".", dent->d_name) && strcmp ("..", dent->d_name))
	{
	    /* Got a file name. Put FQN into source and dest */
	    strncpy (source, old_dir, FQN_SIZE - 12);
	    if (source[strlen (source) - 1] != '/')
		strcat (source, "/");
	    strcat (source, dent->d_name);
	    strncpy (dest, new_dir, FQN_SIZE - 12);
	    if (dest[strlen (dest) - 1] != '/')
		strcat (dest, "/");
	    strcat (dest, dent->d_name);

	    /* Copy the file. On failure skip deleting the source */
	    if (Lmove (source, dest) == ERROR)
	    {
		logMsg ("ERROR copying %s file %s", (int)file_type, (int)dent->d_name, 0, 0, 0, 0);
		continue;
	    }
	}
    }
    /* Close the directory pointer */
    closedir (dir_ptr);

    return OK;
}

/*
 * Function:      CleanDir
 * Purpose:       Removes all files in a directory
 * Arguments:     Directory name
 * Returns:       STATUS OK if successfull or ERROR if directory does not exist
 * Preconditions: Directory should exist
 */
STATUS CleanDir (char *dir)
{
    DIR    *dir_ptr;
    struct dirent *d_ent;
    char   fqn [FQN_SIZE];

    dir_ptr = opendir (dir);

    /* If directory does not exist return an error */
    if (dir_ptr == NULL)
	return ERROR;

    /* Get the next file in the directory */
    while ((d_ent = readdir (dir_ptr)) != NULL)
    {
	if (strcmp (".", d_ent->d_name) && strcmp ("..", d_ent->d_name))
	{
	    /* Construct fully qualified name */
	    strncpy (fqn, dir, FQN_SIZE - 12);
	    if (fqn[strlen (fqn) - 1] != '/')
		strcat (fqn, "/");
	    strcat (fqn, d_ent->d_name);
            
	    if (rm (fqn) == ERROR)
	    {
		/* Error removing file named by readdir */
		if (debug_level > 1)
		    logMsg ("Could not remove file %s... assuming direcory\n", (int)fqn, 0, 0, 0, 0, 0);

		/* Assume that file was really a directory and recurse */
		if (CleanDir (fqn) == ERROR)
		    return ERROR;

		/* Remove the directory */
		if (rmdir (fqn) == ERROR)
		    logMsg ("ERROR removing directory %s", (int)fqn, 0, 0, 0, 0, 0);
		else
		{
		    if (debug_level > 1)
			logMsg ("Directory %s removed\n", (int)fqn, 0, 0, 0, 0, 0);
		}
	    }
	    else
	    {
		if (debug_level > 1)
		    logMsg ("%s removed\n", (int)d_ent->d_name, 0, 0, 0, 0, 0);
	    }
	}
    }
    return OK;
}

/*
 * Function:      MkDirIfAbsent
 * Purpose:       To ensure that the named directory is in place and empty
 * Arguments:     Directory name (fully qualified)
 * Returns:       STATUS OK or ERROR if directory could not be created or emptied
 * Preconditions: The parent directory should exist
 */
static STATUS MkDirIfAbsent (char *dir)
{
    STATUS status = ERROR;
    DIR *dir_ptr;

    dir_ptr = opendir (dir);
    if (dir_ptr)
    {
	/* Directory already exists */
	if (debug_level > 0)
	    logMsg ("Direcory %s already exists... skipping mkdir\n", (int)dir, 0, 0, 0, 0, 0);
	closedir (dir_ptr);
	status = OK;
    }
    else
    {
	if (mkdir (dir) == OK)
	{
	    if (debug_level > 1)
		logMsg ("Direcory %s created\n", (int)dir, 0, 0, 0, 0, 0);
	    status = OK;
	}
	else
	    logMsg ("ERROR creating directory %s\n", (int)dir, 0, 0, 0, 0, 0);
    }
    return status;
}

/*
 * Function:      Lmove
 * Purpose:       Moves a file with logging (dependant on global variable debug_level)
 * Arguments:     Fully qualified file names for souce (from) and destination (to)
 * Returns:       STATUS OK on success or ERROR
 * Preconditions: debug_level should be set to 0 (quiet), 1 or 2
 *                path for destination file should be in place.
 */
static STATUS Lmove(char *from, char *to)
{
    if (debug_level > 1)
	logMsg ("Moving %s to %s ...\t", (int)from, (int)to, 0, 0, 0, 0);
    if (copy (from, to) == ERROR)
    {
	logMsg ("ERROR copying %s to %s\n", (int)from, (int)to, 0, 0, 0, 0);
	return ERROR;
    }
    else
    {
	if (rm (from) == ERROR)
        {
	    logMsg ("Error removing %s\n", (int)from, 0, 0, 0, 0, 0);
            return ERROR;
        }
	else if (debug_level > 1)
	    logMsg ("...Done\n", 0, 0, 0, 0, 0, 0);
    }
    return OK;
}

