#include <stdio.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#ifdef SOLARIS
#include <sys/fcntl.h>
#else
#include <sys/filio.h>
#endif

#include <cadef.h>
#include <odss.h>
#include <signal.h>
#include <syslog.h>

#ifdef ORACLESNAP
#include <trendDefs.h>
#include <oracleSnap.h>
#include <oracleSnapLib.h>
#include <archiveLib.h>
#endif

char	dbVerbose = 1;
#define	V_DEBUG	3
#define	V_STATUS 2
#define	V_INFO	1
#define	VERBOSE(x) if(dbVerbose >= x) syslog

typedef struct odssCtl ODSSCTL;

struct odssCtl
    {
    ODSSCTL	*nextCtl;
    chid	channel;
    evid	event;
    int		handle;
    short	nativeType;
    short	nativeCount;
    short	writeRdbkPending;
    short	writeRdbkTimer;
    short	readPending;
    short	readTimer;
    short	awaitAck;
    };

ODSSCTL		*firstCtl = 0;
ODSSCTL		*newCtl();
ODSSCTL		*findCtl();

extern char	*optarg;

void		fdRegister();
void		connFunc();
void		monDataFunc();
void		readHandler();
void		writeHandler();

int		caFd;
int		clientFd;
struct timeval	oneSecond = {1, 0};

char		clientBuff[20000];
char		xmitBuff[20000];
int		xmitBuffSize = 0;
int		clientBuffSize;
int		clientBuffIndex;
/*
	Verbosity message level strings
*/
char	*msgLevel[] =
{
  "Silent",
  "Informational",
  "Status",
  "DEBUG!"
};
void	odssSignalHandler(int whichone);


/****************************************************************/
/*								*/
/*			THE MAIN PROGRAM			*/
/*								*/
/****************************************************************/

main(argc, argv)
int	argc;
char	**argv;
{
    int			attachFd;
    short		odssPort = ODSS_PORT;
    int			optVal = 1;
    int			maxFd;
    fd_set		fdMask;
    struct sockaddr_in	portAddr;
    int			portAddrSize;
    long		lastTable = 0;
    short		msgType;
    int			clientRev;
    char		chanName[64];
    ODSSCTL		*currCtl;

#ifdef SOLARIS
    sigset(SIGINT,odssSignalHandler);    		/* handle ^C from keyboard */
    sigset(SIGPIPE,odssSignalHandler);           	/* handle broken pipes on writes */
    sigset(SIGUSR1,odssSignalHandler);		/* increment verbose flag */
    sigset(SIGUSR2,odssSignalHandler);		/* decrement verbose flag */
#else
    signal(SIGINT,odssSignalHandler);	       	/* handle ^C from keyboard */
    signal(SIGPIPE,odssSignalHandler);       	/* handle broken pipes on writes */
    signal(SIGUSR1,odssSignalHandler);		/* increment verbose flag */
    signal(SIGUSR2,odssSignalHandler);		/* decrement verbose flag */
#endif

    openlog("odss",LOG_PID,LOG_LOCAL1);
    /*
    Pick up the port number if one is specified.
    */

    if (getopt(argc, argv, "p:") == 'p')
	odssPort = atoi(optarg);

    /*
    Hook up to Channel Access
    */

    if (ca_task_initialize() != ECA_NORMAL)
	{
	VERBOSE(V_INFO)(LOG_ALERT|LOG_LOCAL1,"ODSS interface can't initialize with Channel Access\n");
	exit (2);
	}
    ca_add_fd_registration(fdRegister, NULL);
#ifdef ORACLESNAP
    if(getenv("ORACLE_HOME") != NULL)
      OpenOracle();					/* start oracle initialization */
#endif


for (;;)
    {

    /*
    Create a socket for a user to attach to.
    */

    attachFd = socket(PF_INET, SOCK_STREAM, 0);
    setsockopt(attachFd, SOL_SOCKET, SO_REUSEADDR, (caddr_t) &optVal, sizeof(optVal));
    portAddr.sin_addr.s_addr = INADDR_ANY;
    portAddr.sin_port = htons(odssPort);
    if (bind(attachFd, (struct sockaddr *) &portAddr, sizeof(portAddr)) == -1)
	{
	VERBOSE(V_INFO)(LOG_ALERT|LOG_LOCAL1,"Can't bind to specified TCP port %d",odssPort);
	exit(2);
	}
    listen(attachFd, 2);
    FD_ZERO(&fdMask);
    FD_SET(attachFd, &fdMask);
    maxFd = attachFd;
    FD_SET(caFd, &fdMask);
    if (caFd > maxFd)
	maxFd = caFd;

    /*
    Wait for an event on this connection socket or from CA. The only event from CA that
    should ever happen here is that we might see him drop away from us. No channels should
    be open at this point in the logic. Here, we can wait forever.
    */

    select(maxFd + 1, &fdMask, NULL, NULL, NULL);
    if (FD_ISSET(caFd, &fdMask))
	ca_pend_io(0.0);
    if (!FD_ISSET(attachFd, &fdMask))
	continue;

    /*
    Someone must be trying to attach to us. Create and "condition" his socket.
    */

    portAddrSize = sizeof(portAddr);
    clientFd = accept(attachFd, (struct sockaddr *) &portAddr, &portAddrSize);
#ifndef SOLARIS
    ioctl(clientFd, FIONBIO, &optVal);
#else
    sockGrow(clientFd);
    fcntl(clientFd, F_SETFL, O_NONBLOCK);
#endif
    setsockopt(clientFd, IPPROTO_TCP, TCP_NODELAY, (char *)&optVal, sizeof(optVal));
    clientRev = 0;		/* Don't know the client's revision level yet. */
    clientBuffIndex = 0;
    clientBuffSize = 0;

    /*
    Closing the original TCP socket guarantees that we only service one client at a time.
    */

    close(attachFd);

    /*
    Now, service both the client and Channel Access.
    */

    maxFd = caFd;
    if (clientFd > maxFd)
	maxFd = clientFd;

    VERBOSE(V_INFO)(LOG_ALERT|LOG_LOCAL1,"Connecting to client on port %d.",odssPort);
    for (;;)
	{

	/*
	Watch both the client and CA sockets with a one second timeout. Do this because 
	there might be timers that we have to count down.
	*/

	FD_ZERO(&fdMask);
	FD_SET(caFd, &fdMask);
	FD_SET(clientFd, &fdMask);
	select(maxFd + 1, &fdMask, NULL, NULL, &oneSecond);
	if (FD_ISSET(caFd, &fdMask))
	  {
	    ca_pend_io(0.0);
	    flushClient();
	  }
	if (FD_ISSET(clientFd, &fdMask))
	    {

	    /*
	    Service the client's socket.
	    */

	    if (!fillClientBuffer())
		break;
	    while (haveCompleteClientMsg())
		{

		/*
		Get the message type to see what follows in the stream.
		*/

		readClient((char *) &msgType, sizeof(msgType));
		msgType = ntohs(msgType);

		if (msgType == ODSS_BUILD)
		    {
		    ODSS_BUILD_MSG	odssBuildMsg;

		    /*
		    Build a channel.
		    */

		    readClient((char *) &odssBuildMsg, sizeof(odssBuildMsg));
		    odssBuildMsg.nameLength = ntohs(odssBuildMsg.nameLength);
		    if((odssBuildMsg.nameLength <= 0) || (odssBuildMsg.nameLength > PVNAME_SZ+FLDNAME_SZ+1))
			{
			VERBOSE(V_INFO)(LOG_ALERT|LOG_LOCAL1,"Invalid channel name length %d... terminating connection.\n",odssBuildMsg.nameLength);
		        closeClientChannels();
		        close(clientFd);
		        clientFd = -1;
		        break;
			}
		    /*
		      pad chanName with blanks so we can tell when the name length is wrong.
		      GCT 18-Mar-1998
		    */
		    memset(chanName,' ',63);
		    chanName[63] = 0;
		    readClient(chanName, odssBuildMsg.nameLength);
		    if (!clientRev)
			continue;
		    /*
		      Compare the length of the string read in by readClient. if nameLength-1 is not EQUAL to
		      strlen(chanName) then announce error and drop socket
		      GCT 18-Mar-1998
		    */
		    if((strlen(chanName) != (odssBuildMsg.nameLength-1)))
			{
			VERBOSE(V_INFO)(LOG_ALERT|LOG_LOCAL1,"Invalid channel name length %d... terminating connection.\n",odssBuildMsg.nameLength);
		        closeClientChannels();
		        close(clientFd);
		        clientFd = -1;
		        break;
			}
		    
		    /*
		    Allocate a control structure and fill it out.
		    */

		    currCtl = newCtl();
		    ca_build_and_connect(chanName, TYPENOTCONN, 0, &(currCtl -> channel), NULL, connFunc, currCtl);
		    currCtl -> handle = odssBuildMsg.ODSShandle;
		    currCtl -> awaitAck = 1;
		    currCtl -> nextCtl = firstCtl;
		    firstCtl = currCtl;
		    }

		else if (msgType == ODSS_MONITOR)
		    {
		    ODSS_MONITOR_MSG	odssMonitorMsg;
		    int			mask = 0;
		    int			type;

		    /*
		    Establish or clear a monitor.
		    */

		    readClient((char *) &odssMonitorMsg, sizeof(odssMonitorMsg));
		    if (!clientRev)
			continue;
		    currCtl = findCtl(odssMonitorMsg.ODSShandle);
		    if (!currCtl)
			continue;
		    odssMonitorMsg.eventMask = htons(odssMonitorMsg.eventMask);
		    if (odssMonitorMsg.eventMask)
			{
			if (odssMonitorMsg.eventMask & 1)
			    mask |= DBE_VALUE;
			if (odssMonitorMsg.eventMask & 2)
			    mask |= DBE_LOG;
			if (odssMonitorMsg.eventMask & 4)
			    mask |= DBE_ALARM;
			type = dbf_type_to_DBR_STS(ca_field_type(currCtl -> channel));
			ca_add_masked_array_event(type, ca_element_count(currCtl -> channel), 
				currCtl -> channel, monDataFunc, currCtl, 0.0, 0.0, 0.0, 
				&(currCtl -> event), mask);
			}
		    else
			ca_clear_event(currCtl -> event);
		    }

		else if (msgType == ODSS_READ)
		    {
		    ODSS_READ_MSG	odssReadMsg;

		    /*
		    Read a channel.
		    */


		    readClient((char *) &odssReadMsg, sizeof(odssReadMsg));
		    if (!clientRev)
			continue;
		    currCtl = findCtl(odssReadMsg.ODSShandle);
		    if (!currCtl)
			continue;

		    /*
		    Check to see if the point is currently connected.
		    */

		    if (ca_state(currCtl -> channel) != currOpen)
			{
			ODSS_READ_DATA_MSG	readReply;

			readReply.ODSShandle = currCtl -> handle;
			readReply.readState = htons(2);		/* Offline */
			readReply.dataLength = 0;
			msgType = htons(ODSS_READ_DATA);
			writeClient(msgType, (char *) &readReply, sizeof(readReply), NULL, 0);
			continue;
			}

		    /*
		    Connected. Issue the read.
		    */

		    odssReadMsg.timeout = ntohs(odssReadMsg.timeout);
		    if (odssReadMsg.timeout <= 0)
			{
			VERBOSE(V_INFO)(LOG_ALERT|LOG_LOCAL1,"Invalid timeout in ODSS_READ message, terminating connection.\n");
		        closeClientChannels();
		        close(clientFd);
		        clientFd = -1;
		        break;
			}
		    currCtl -> readPending = 1;
		    currCtl-> readTimer = odssReadMsg.timeout;
		    ca_array_get_callback(dbf_type_to_DBR_STS(ca_field_type(currCtl -> channel)), 
			    ca_element_count(currCtl -> channel), currCtl -> channel, readHandler, currCtl);
		    }

		else if (msgType == ODSS_WRITE)
		    {
		    ODSS_WRITE_MSG	odssWriteMsg;
		    char		writeData[1024];
		    int typeErr=0;

		    /*
		    Write a channel.
		    */

		    readClient((char *) &odssWriteMsg, sizeof(odssWriteMsg));
		    odssWriteMsg.dataLength = ntohs(odssWriteMsg.dataLength);
		    odssWriteMsg.dataType = ntohs(odssWriteMsg.dataType);
		    if((odssWriteMsg.dataType < AGNT_STRING) || (odssWriteMsg.dataType >  AGNT_DOUBLE))
			{
			VERBOSE(V_INFO)(LOG_ALERT|LOG_LOCAL1,"Unknown data type %d in ODSS_WRITE message, terminating connection.\n",odssWriteMsg.dataType);
		        closeClientChannels();
		        close(clientFd);
		        clientFd = -1;
		        break;
			}
		    if (odssWriteMsg.dataLength <= 0)
			{
			VERBOSE(V_INFO)(LOG_ALERT|LOG_LOCAL1,"Invalid data length %d in ODSS_WRITE message, terminating connection.\n",odssWriteMsg.dataLength);
		        closeClientChannels();
		        close(clientFd);
		        clientFd = -1;
		        break;
			}
		    switch(odssWriteMsg.dataType)
		      {
		      case AGNT_FLOAT:
			if(odssWriteMsg.dataLength != sizeof(float))
			  typeErr = 1;
			break;
		      case AGNT_DOUBLE:
			if(odssWriteMsg.dataLength != sizeof(double))
			  typeErr = 1;
			break;
		      case AGNT_SHORT:
		      case AGNT_ENUM:
			if(odssWriteMsg.dataLength != sizeof(short))
			  typeErr = 1;
			break;
		      case AGNT_LONG:
			if(odssWriteMsg.dataLength != sizeof(int))
			  typeErr = 1;
			break;
		      case AGNT_CHAR:
			if(odssWriteMsg.dataLength != sizeof(char))
			  typeErr = 1;
			break;
		      }
		    if(typeErr == 1)
			{
			VERBOSE(V_INFO)(LOG_ALERT|LOG_LOCAL1,"Invalid data length %d for data type %d, terminating connection.\n",odssWriteMsg.dataType,odssWriteMsg.dataLength);
		        closeClientChannels();
		        close(clientFd);
		        clientFd = -1;
		        break;
			}
		    readClient(writeData, odssWriteMsg.dataLength);
		    if (!clientRev)
			continue;
		    currCtl = findCtl(odssWriteMsg.ODSShandle);
		    if (!currCtl)
			continue;

		    /*
		    Check to see if the point is currently connected.
		    */

		    if (odssWriteMsg.timeout && (ca_state(currCtl -> channel) != currOpen))
			{
			ODSS_WRITE_ACK_MSG	writeReply;

			writeReply.ODSShandle = currCtl -> handle;
			writeReply.status = htons(2);
			msgType = htons(ODSS_WRITE_ACK);
			writeClient(msgType, (char *) &writeReply, sizeof(writeReply), NULL, 0);
			continue;
			}

		    if (ca_state(currCtl -> channel) != currOpen)
			continue;

		    /*
		    Issue the write, then see if the user wanted a timeout.
		    */

		    ca_array_put(ca_field_type(currCtl -> channel), ca_element_count(currCtl -> channel),
			    currCtl -> channel, writeData);
		    odssWriteMsg.timeout = ntohs(odssWriteMsg.timeout);
		    if (odssWriteMsg.timeout < 0)
			{
			VERBOSE(V_INFO)(LOG_ALERT|LOG_LOCAL1,"Invalid timeout in ODSS_WRITE message, terminating connection.\n");
		        closeClientChannels();
		        close(clientFd);
		        clientFd = -1;
		        break;
			}
		    if (odssWriteMsg.timeout)
			{
			currCtl -> writeRdbkPending = 1;
			currCtl -> writeRdbkTimer = odssWriteMsg.timeout;
			ca_array_get_callback(ca_field_type(currCtl -> channel), 1, currCtl -> channel,
				writeHandler, currCtl);
			}
		    }

		else if (msgType == ODSS_HIST_READ)
		    {
		    ODSS_HIST_READ_MSG	odssHistReadMsg;
		    ODSS_HIST_DATA_MSG	odssHistDataMsg;
		    char		*reqBuffer;
		    char		*outputBuff = 0;
		    long		dbRate;
		    char		eguTag[32];
		    double		dispLow;
		    double		dispHigh;
		    int			numReturned = 0;
		    struct tm *timeCheck;
		    char timeString[30];

		    /*
		    Read historical data.
		    */

		    readClient((char *) &odssHistReadMsg, sizeof(odssHistReadMsg));
		    odssHistReadMsg.nameLen = ntohs(odssHistReadMsg.nameLen);
		    if((odssHistReadMsg.nameLen <= 0) || (odssHistReadMsg.nameLen > PVNAME_SZ+FLDNAME_SZ+1))
			{
			VERBOSE(V_INFO)(LOG_ALERT|LOG_LOCAL1,"Invalid channel name length %d... terminating connection.\n",odssHistReadMsg.nameLen);
		        closeClientChannels();
		        close(clientFd);
		        clientFd = -1;
		        break;
			}
		    /*
		      pad chanName with blanks so we can tell when the name length is wrong.
		      GCT 18-Mar-1998
		    */
		    memset(chanName,' ',63);
		    chanName[63] = 0;
		    readClient(chanName, odssHistReadMsg.nameLen);
		    if (!clientRev)
			continue;
		    /*
		      Compare the length of the string read in by readClient. if nameLength-1 is not EQUAL to
		      strlen(chanName) then announce error and drop socket
		      GCT 18-Mar-1998
		    */
		    if((strlen(chanName) != (odssHistReadMsg.nameLen-1)))
			{
			VERBOSE(V_INFO)(LOG_ALERT|LOG_LOCAL1,"Invalid channel name length %d... terminating connection.\n",odssHistReadMsg.nameLen);
		        closeClientChannels();
		        close(clientFd);
		        clientFd = -1;
		        break;
			}
		    odssHistReadMsg.secsPerSample = ntohl(odssHistReadMsg.secsPerSample);
		    if(odssHistReadMsg.secsPerSample <= 0)
			{
			VERBOSE(V_INFO)(LOG_ALERT|LOG_LOCAL1,"Invalid number of seconds/sample %d... terminating connection.\n",odssHistReadMsg.secsPerSample);
		        closeClientChannels();
		        close(clientFd);
		        clientFd = -1;
		        break;
			}
		    odssHistReadMsg.startTime = ntohl(odssHistReadMsg.startTime);
		    timeCheck = localtime((time_t *)&odssHistReadMsg.startTime);
		    if(timeCheck->tm_year < 97)
			{
			  strcpy(timeString,ctime((time_t *)&odssHistReadMsg.startTime));
			  timeString[strlen(timeString)-1] = 0;
			VERBOSE(V_INFO)(LOG_ALERT|LOG_LOCAL1,"Invalid Date/Time %s... terminating connection.\n",timeString);
		        closeClientChannels();
		        close(clientFd);
		        clientFd = -1;
		        break;
			}
		    odssHistReadMsg.numSamples = ntohl(odssHistReadMsg.numSamples);
		    if((odssHistReadMsg.numSamples <= 0) || (odssHistReadMsg.numSamples > 1000))
			{
			VERBOSE(V_INFO)(LOG_ALERT|LOG_LOCAL1,"Invalid number of samples %d... terminating connection.\n",odssHistReadMsg.numSamples);
		        closeClientChannels();
		        close(clientFd);
		        clientFd = -1;
		        break;
			}
		    odssHistReadMsg.collRate = ntohl(odssHistReadMsg.collRate);
		    if(odssHistReadMsg.collRate <= 0)
			{
			VERBOSE(V_INFO)(LOG_ALERT|LOG_LOCAL1,"Invalid rate %d... terminating connection.\n",odssHistReadMsg.collRate);
		        closeClientChannels();
		        close(clientFd);
		        clientFd = -1;
		        break;
			}
		    if(strncmp(odssHistReadMsg.dataReduc,"SMP",3) &&
		       strncmp(odssHistReadMsg.dataReduc,"AVG",3) &&
		       strncmp(odssHistReadMsg.dataReduc,"MIN",3) &&
		       strncmp(odssHistReadMsg.dataReduc,"MAX",3))
			{
			VERBOSE(V_INFO)(LOG_ALERT|LOG_LOCAL1,"Invalid desired reduction method %3.3s... terminating connection.\n",odssHistReadMsg.dataReduc);
		        closeClientChannels();
		        close(clientFd);
		        clientFd = -1;
		        break;
			}
		    if(strncmp(odssHistReadMsg.dbReduc,"SMP",3) &&
		       strncmp(odssHistReadMsg.dbReduc,"AVG",3) &&
		       strncmp(odssHistReadMsg.dbReduc,"MIN",3) &&
		       strncmp(odssHistReadMsg.dbReduc,"MAX",3))
			{
			VERBOSE(V_INFO)(LOG_ALERT|LOG_LOCAL1,"Invalid file reduction method %3.3s... terminating connection.\n",odssHistReadMsg.dbReduc);
		        closeClientChannels();
		        close(clientFd);
		        clientFd = -1;
		        break;
			}
#ifdef ORACLESNAP
		    if(getenv("ORACLE_HOME") == NULL)
		      continue;
		    reqBuffer = (char *)malloc(odssHistReadMsg.numSamples * (sizeof(double) + 4 + sizeof(double)));
		    numReturned = GetSnapValues(chanName, 		/* Point.field */
				      odssHistReadMsg.dbReduc, 		/* trend file reduction method */
				      odssHistReadMsg.dataReduc, 	/* Method for further reduction */
				      odssHistReadMsg.secsPerSample, 	/* Seconds per sample */
				      odssHistReadMsg.startTime, 	/* Secs since 1/1/70 */
		    		      &dbRate, 				/* An output we don't need */
				      eguTag, 				/* also not needed */
				      &dispLow, 			/* also not needed */
				      &dispHigh, 			/* also not needed */
				      odssHistReadMsg.numSamples,	/* Number of samples */
		    		      (double*)reqBuffer, 			/* The data values */
				      (unsigned char **)(reqBuffer + (sizeof(double) + sizeof(double)) *
					    odssHistReadMsg.numSamples),	/* The status array */
				      (double *)(reqBuffer + sizeof(double) * odssHistReadMsg.numSamples));
									/* counts for each cell */
		    if (numReturned)
			outputBuff = reqBuffer;
		    odssHistDataMsg.numReturned = htonl(numReturned);
		    numReturned *= (sizeof(double) + sizeof(double) + 3);
		    odssHistDataMsg.dataLength = htonl(numReturned);
		    msgType = htons(ODSS_HIST_DATA);
		    writeClient(msgType, (char *) &odssHistDataMsg, sizeof(odssHistDataMsg),
			    outputBuff, numReturned);
		    free(reqBuffer);
#endif
		    }

		else if (msgType == ODSS_CHAN_CLEAR)
		    {
		    ODSS_CHAN_CLEAR_MSG	odssChanClearMsg;
		    ODSSCTL		*lookCtl;
		    ODSSCTL		*prevCtl;
		    ODSSCTL		*nextCtl;
		    int			wasDeleted = 0;

		    /*
		    Close a channel.
		    */

		    readClient((char *) &odssChanClearMsg, sizeof(odssChanClearMsg));
		    if (!clientRev)
			continue;
		    currCtl = findCtl(odssChanClearMsg.ODSShandle);
		    if (!currCtl)
		      {
			VERBOSE(V_INFO)(LOG_ALERT|LOG_LOCAL1,"Attempt to clear unknown channel <%d>\n",odssChanClearMsg.ODSShandle);
			continue;
		      }

		    /*
		    Now unchain and free the currCtl block for this channel.
		    Close ALL the channels with this handle just in case the
		    user has given us the same handle multiple times.
		    */

		    prevCtl = 0;
		    nextCtl = firstCtl;
		    while ((lookCtl = nextCtl))
			{
			nextCtl = lookCtl -> nextCtl;
			if (lookCtl -> handle == currCtl -> handle)
			    {
			    if (prevCtl)
				prevCtl -> nextCtl = lookCtl -> nextCtl;
			    else
				firstCtl = lookCtl -> nextCtl;
			    ca_clear_channel(lookCtl -> channel);
			    deleteCtl(lookCtl);
			    wasDeleted = 1;
			    }
			else
			    {
			    prevCtl = lookCtl;
			    }
			}
		    if(wasDeleted == 0)
			VERBOSE(V_INFO)(LOG_ALERT|LOG_LOCAL1,"Attempt to clear unknown channel <%d>\n",odssChanClearMsg.ODSShandle);
		    }

		else if (msgType == ODSS_BOOMERANG)
		    {
		    ODSS_BOOMERANG_MSG	odssBoomerangMsg;

		    /*
		    Boomerang, so he can ping us. Client doesn't have to declare
		    his revision level before using this message.
		    */

		    readClient((char *) &odssBoomerangMsg, sizeof(odssBoomerangMsg));
		    msgType = htons(ODSS_BOOMERANG);
		    if (!writeClient(msgType, (char *) &odssBoomerangMsg, sizeof(odssBoomerangMsg),
				NULL, 0))
			break;
		    }

		else if (msgType == ODSS_REV)
		    {
		    ODSS_REV_MSG	odssRevMsg;

		    /*
		    User declares his protocol revision level.
		    */

		    readClient((char *) &odssRevMsg, sizeof(odssRevMsg));
		    if(ntohl(odssRevMsg.ODSSRev) != ODSS_CURRENT_REV)
			{
			VERBOSE(V_INFO)(LOG_ALERT|LOG_LOCAL1,"Unsupported protocol level %d... terminating connection.\n",ntohl(odssRevMsg.ODSSRev));
		        closeClientChannels();
		        close(clientFd);
		        clientFd = -1;
		        break;
			}
		    clientRev = ntohl(odssRevMsg.ODSSRev);
		    }

		else
		    {

		    /*
		    OOPS! None of the above. The input stream must have gotten
		    out of synchronization.
		    */

		    VERBOSE(V_INFO)(LOG_ALERT|LOG_LOCAL1,"Unknown message code %d received, terminating connection\n",msgType);
		    closeClientChannels();
		    close(clientFd);
		    clientFd = -1;
		    break;
		    }
		}

	    flushClient();
	    if (clientFd < 0)
		break;
	    ca_flush_io();
	    }

	/*
	Look at the table if it is time to do so.
	*/

	if (time(NULL) > lastTable)
	    {

	    /*
	    Traverse the table of channels, seeing if we have to time out anything.
	    */

	    for (currCtl = firstCtl; currCtl; currCtl = currCtl -> nextCtl)
		{
		if (currCtl -> writeRdbkPending)
		    {
		    if (--(currCtl -> writeRdbkTimer) < 0)
			{
			ODSS_WRITE_ACK_MSG	odssWriteAck;

			/*
			Write verification has timed out.
			*/

			currCtl -> writeRdbkPending = 0;
			currCtl -> writeRdbkTimer = 0;

			odssWriteAck.ODSShandle = currCtl -> handle;
			odssWriteAck.status = htons(1);
			msgType = htons(ODSS_WRITE_ACK);
			writeClient(msgType, (char *) &odssWriteAck, sizeof(odssWriteAck),
				NULL, 0);
			}
		    }
		if (currCtl -> readPending)
		    {
		    if (--(currCtl -> readTimer) < 0)
			{
			ODSS_READ_DATA_MSG	odssReaddata;

			/*
			A read has timed out.
			*/

			currCtl -> readPending = 0;
			currCtl -> readTimer = 0;

			odssReaddata.ODSShandle = currCtl -> handle;
			odssReaddata.readState = htons(1);
			odssReaddata.dataLength = 0;
			msgType = htons(ODSS_READ_DATA);
			writeClient(msgType, (char *) &odssReaddata, sizeof(odssReaddata),
				NULL, 0);
			}
		    }
		}
	    lastTable = time(NULL);
	    }
	}
    }
}


/****************************************************************/
/*								*/
/*	CHANNEL ACCESS SOCKET FD REGISTRATION FUNCTION		*/
/*								*/
/****************************************************************/

void fdRegister(arg, fd, direction)
int	arg;
int	fd;
int	direction;
{
    if (direction)
	caFd = fd;
    else
	{
	VERBOSE(V_INFO)(LOG_ALERT|LOG_LOCAL1,"Connection with channel access lost\n");
	exit(2);
	}
}


/****************************************************************/
/*								*/
/*	READ THE CLIENT'S SOCKET TO FILL UP OUR BUFFER		*/
/*								*/
/****************************************************************/

fillClientBuffer()
{
    int		bytesRead;

    /*
    If there is an unprocessed message fragment in the buffer, slide
    it down to the start of the buffer.
    */

    if (clientBuffSize && clientBuffIndex)
	memcpy(clientBuff, clientBuff + clientBuffIndex, clientBuffSize);

    /*
    Read as much additional information as we can.
    */

    bytesRead = recv(clientFd, clientBuff + clientBuffSize, sizeof(clientBuff) - clientBuffSize, 0);
    if (bytesRead <= 0)
	{
	VERBOSE(V_INFO)(LOG_ALERT|LOG_LOCAL1,"EOF or error reading client socket");
	closeClientChannels();
	close(clientFd);
	clientFd = -1;
	return (0);
	}

    clientBuffSize += bytesRead;
    clientBuffIndex = 0;
    return (1);
}


/****************************************************************/
/*								*/
/*	CLOSE ALL THE CHANNELS OPEN ON BEHALF OF A CLIENT	*/
/*								*/
/****************************************************************/

closeClientChannels()
{
    ODSSCTL	*lookCtl;
    ODSSCTL	*nextCtl;

    nextCtl = firstCtl;
    while ((lookCtl = nextCtl))
	{
	nextCtl = lookCtl -> nextCtl;
	ca_clear_channel(lookCtl -> channel);
	deleteCtl(lookCtl);
	}
    firstCtl = 0;
}


/****************************************************************/
/*								*/
/*	DETERMINE IF WE HAVE A COMPLETE MSG IN OUR BUFFER	*/
/*								*/
/****************************************************************/

haveCompleteClientMsg()
{
    short	msgType;
    int32 tempLength;

    /*
    If the socket is gone, there is no point is processing anything.
    */

    if (clientFd < 0)
	return (0);

    /*
    Must have at least a message type.
    */

    if (clientBuffSize < sizeof(short))
	return (0);
    memcpy((char *) &msgType, clientBuff + clientBuffIndex, sizeof(short));
    msgType = htons(msgType);
    if (msgType == ODSS_BUILD)
	{
	ODSS_BUILD_MSG		*odssBuildMsg;

	if (clientBuffSize < (sizeof(ODSS_BUILD_MSG) + sizeof(short)))
	    return (0);
	odssBuildMsg = (ODSS_BUILD_MSG *)(clientBuff + clientBuffIndex + sizeof(short));
	memcpy((char *) &tempLength, (char *) &(odssBuildMsg -> nameLength), sizeof(tempLength));
	tempLength = ntohl(tempLength);
	if (clientBuffSize < (sizeof(ODSS_BUILD_MSG) + tempLength + sizeof(short)))
	    return (0);
	else
	    return (1);
	}
    else if (msgType == ODSS_MONITOR)
	{
	if (clientBuffSize < (sizeof(ODSS_MONITOR_MSG) + sizeof(short)))
	    return (0);
	else
	    return (1);
	}
    else if (msgType == ODSS_READ)
	{
	if (clientBuffSize < (sizeof(ODSS_READ_MSG) + sizeof(short)))
	    return (0);
	else
	    return (1);
	}
    else if (msgType == ODSS_WRITE)
	{
	ODSS_WRITE_MSG		*odssWriteMsg;

	if (clientBuffSize < (sizeof(ODSS_WRITE_MSG) + sizeof(short)))
	    return (0);
	odssWriteMsg = (ODSS_WRITE_MSG *) (clientBuff + clientBuffIndex + sizeof(short));
	memcpy((char *) &msgType, (char *) &(odssWriteMsg -> dataLength), sizeof(msgType));
	msgType = ntohs(msgType);
	if (clientBuffSize < (sizeof(ODSS_WRITE_MSG) + msgType + sizeof(short)))
	    return (0);
	else
	    return (1);
	}
    else if (msgType == ODSS_HIST_READ)
	{
	ODSS_HIST_READ_MSG	*odssHistReadMsg;

	if (clientBuffSize < (sizeof(ODSS_HIST_READ_MSG) + sizeof(short)))
	    return (0);
	odssHistReadMsg = (ODSS_HIST_READ_MSG *) (clientBuff + clientBuffIndex + sizeof(short));
	memcpy((char *) &msgType, (char *) &(odssHistReadMsg -> nameLen), sizeof(msgType));
	msgType = ntohs(msgType);
	if (clientBuffSize < (sizeof(ODSS_HIST_READ_MSG) + msgType + sizeof(short)))
	    return (0);
	else
	    return (1);
	}
    else if (msgType == ODSS_CHAN_CLEAR)
	{
	if (clientBuffSize < (sizeof(ODSS_CHAN_CLEAR_MSG) + sizeof(short)))
	    return (0);
	else
	    return (1);
	}
    else if (msgType == ODSS_BOOMERANG)
	{
	if (clientBuffSize < (sizeof(ODSS_BOOMERANG_MSG) + sizeof(short)))
	    return (0);
	else
	    return (1);
	}
    else if (msgType == ODSS_REV)
	{
	if (clientBuffSize < (sizeof(ODSS_REV_MSG) + sizeof(short)))
	    return (0);
	else
	    return (1);
	}
    else
	{

	/*
	Unknown message code. Let the higher levels bounce this client.
	*/

	return (1);
	}
}


/****************************************************************/
/*								*/
/*		READ BYTES OUT OF THE CLIENT BUFFER		*/
/*								*/
/*	There will always be the number of bytes requested	*/
/*	resident in the buffer because the high level will	*/
/*	have called haveCompleteClientMsg() before calling	*/
/*	this routine.						*/
/*								*/
/****************************************************************/

readClient(buff, nbytes)
char	*buff;
int	nbytes;
{
    memcpy(buff, clientBuff + clientBuffIndex, nbytes);
    clientBuffIndex += nbytes;
    clientBuffSize -= nbytes;
}


/****************************************************************/
/*								*/
/*		WRITE A MESSAGE TO THE CLIENT			*/
/*								*/
/*	Return a zero on failure. This routine will only fail	*/
/*	if we flush the buffer to make room for a message and	*/
/*	that flush fails.					*/
/*								*/
/****************************************************************/

writeClient(msgType, buff1, buff1Size, buff2, buff2Size)
short	msgType;
char	*buff1;
int	buff1Size;
char	*buff2;
int	buff2Size;
{
    if (clientFd < 0)
	return (0);

    if ((xmitBuffSize + sizeof(msgType) + buff1Size + buff2Size) > sizeof(xmitBuff))
	{
	if (!flushClient())
	    return (0);
	}
    memcpy(xmitBuff + xmitBuffSize, (char *) &msgType, sizeof(msgType));
    xmitBuffSize += sizeof(msgType);
    if (buff1 && buff1Size)
	{
	memcpy(xmitBuff + xmitBuffSize, buff1, buff1Size);
	xmitBuffSize += buff1Size;
	}
    if (buff2 && buff2Size)
	{
	memcpy(xmitBuff + xmitBuffSize, buff2, buff2Size);
	xmitBuffSize += buff2Size;
	}
    return (1);
}


/****************************************************************/
/*								*/
/*	    PERFORM A PHYSICAL WRITE TO THE CLIENT		*/
/*								*/
/*	Return a zero on failure. If we do fail, set things	*/
/*	up so that the higher levels don't have to perform	*/
/*	any additional housekeeping.				*/
/*								*/
/****************************************************************/

flushClient()
{
    char	*buffOut;
    int		bytesSent;
    int		sockTxBuff;
    int		sockSize;

    if (clientFd < 0)
	return (0);

    buffOut = xmitBuff;
    while (xmitBuffSize)
	{
	bytesSent = send(clientFd, buffOut, xmitBuffSize, 0);
	if (bytesSent <= 0)
	    {
	    if (bytesSent == 0)
		VERBOSE(V_INFO)(LOG_ALERT|LOG_LOCAL1,"Zero length write to client, terminating connection.\n");
	    else if ((errno == EWOULDBLOCK) || (errno == EAGAIN))
		{
		sockSize = sizeof(sockTxBuff);
		getsockopt(clientFd, SOL_SOCKET, SO_SNDBUF, (char *)&sockTxBuff, &sockSize);
		sockTxBuff += xmitBuffSize;
		sockSize = sizeof(sockTxBuff);
		if (setsockopt(clientFd, SOL_SOCKET, SO_SNDBUF, (char *)&sockTxBuff, sockSize) == -1)
		    {
		    VERBOSE(V_INFO)(LOG_ALERT|LOG_LOCAL1,"Terminating connection with application because of Tx buffer maxout.\n");
		    close(clientFd);
		    closeClientChannels();
		    return (0);
		    }
		continue;
		}
	    else
		VERBOSE(V_INFO)(LOG_ALERT|LOG_LOCAL1,"Terminating client because of I/O error %m.\n");

	    /*
	    Do the housekeeping for a transmission failure.
	    */

	    close(clientFd);
	    clientFd = -1;
	    closeClientChannels();
	    return (0);
	    }

	buffOut += bytesSent;
	xmitBuffSize -= bytesSent;
	}
    return (1);
}


/****************************************************************/
/*								*/
/*		ALLOCATE A NEW CONTROL STRUCTURE		*/
/*								*/
/****************************************************************/

static ODSSCTL		*ctlPool = 0;
#define CTL_ALLOC	50

ODSSCTL *newCtl()
{
    int		index;
    ODSSCTL	*currCtl;

    if (!ctlPool)
	{
	ctlPool = (ODSSCTL *) malloc(sizeof(ODSSCTL) * CTL_ALLOC);
	currCtl = ctlPool;
	index = CTL_ALLOC;
	while (index--)
	    {
	    if (index)
		currCtl -> nextCtl = currCtl + 1;
	    else
		currCtl -> nextCtl = 0;
	    }
	}
    currCtl = ctlPool;
    ctlPool = currCtl -> nextCtl;
    memset((char *) currCtl, 0, sizeof(ODSSCTL));
    return (currCtl);
}


/****************************************************************/
/*								*/
/*		FREE AN EXISTING CONTROL STRUCTURE		*/
/*								*/
/****************************************************************/

deleteCtl(oldCtl)
ODSSCTL		*oldCtl;
{
    oldCtl -> nextCtl = ctlPool;
    ctlPool = oldCtl;
}


/****************************************************************/
/*								*/
/*		CHANNEL ACCESS CONNECTION CALLBACK		*/
/*								*/
/****************************************************************/

void connFunc(connArgs)
struct connection_handler_args	connArgs;
{
    ODSS_BUILD_ACK_MSG		buildAckMsg;
    ODSS_CHAN_CONN_MSG		chanConnMsg;
    ODSSCTL 			*currCtl;
    short			direction;
    short			msgType;

    /*
    The channel has just connected or disconnected. Inform the user.
    */

    if (connArgs.op == CA_OP_CONN_UP)
	direction = 1;
    else
	direction = 0;
    currCtl = (ODSSCTL *) ca_puser(connArgs.chid);

    /*
    He might be awaiting an ACK to the original build.
    */

    if (direction && currCtl -> awaitAck)
	{
	buildAckMsg.ODSShandle = currCtl -> handle;
	currCtl -> nativeType = ca_field_type(connArgs.chid);
	buildAckMsg.nativeType = htons(currCtl -> nativeType);
	currCtl -> nativeCount = ca_element_count(connArgs.chid);
	buildAckMsg.nativeCount = htons(currCtl -> nativeCount);
	msgType = htons(ODSS_BUILD_ACK);
	writeClient(msgType, (char *) &buildAckMsg, sizeof(buildAckMsg), NULL, 0);
	currCtl -> awaitAck = 0;
	}

    msgType = htons(ODSS_CHAN_CONN);
    chanConnMsg.ODSShandle = currCtl -> handle;
    chanConnMsg.connState = htons(direction);
    writeClient(msgType, (char *) &chanConnMsg, sizeof(chanConnMsg), NULL, 0);
}


/****************************************************************/
/*								*/
/*	FIND A CONTROL STRUCTURE GIVEN THE USER'S HANDLE	*/
/*								*/
/****************************************************************/

ODSSCTL *findCtl(handle)
int	handle;
{
    ODSSCTL	*currCtl;

    currCtl = firstCtl;
    while (currCtl)
	{
	if (currCtl -> handle == handle)
	    break;
	currCtl = currCtl -> nextCtl;
	}
    
    return (currCtl);
}


/****************************************************************/
/*								*/
/*		MONITOR DATA EVENT CALLBACK HANDLER		*/
/*								*/
/****************************************************************/

void monDataFunc(args)
struct event_handler_args	args;
{
    ODSS_MONDATA_MSG	odssMondata;
    ODSSCTL		*currCtl = (ODSSCTL *) args.usr;
    char		*pValue;
    short		msgType;
    int			size;
    int			type;

    odssMondata.ODSShandle = currCtl -> handle;

    /*
    Stat and sevr are always in the same place in each data structure.
    */

    odssMondata.status = *(short *)args.dbr;
    odssMondata.status = htons(odssMondata.status);
    odssMondata.severity = *(short *)(((char *)args.dbr) + sizeof(short));
    odssMondata.severity = htons(odssMondata.severity);
    type = ca_field_type(currCtl -> channel);
    odssMondata.dataType = type;
    odssMondata.dataType = htons(odssMondata.dataType);

    /*
    Get a pointer to the value.
    */

    pValue = dbr_value_ptr(args.dbr, args.type);
    if (type == DBF_STRING)
	size = strlen(pValue) + 1;
    else
	size = dbr_size_n(type, args.count);
    odssMondata.dataLength = htons(size);
    msgType = htons(ODSS_MONDATA);
    writeClient(msgType, (char *) &odssMondata, sizeof(odssMondata),
				pValue, size);
}


/****************************************************************/
/*								*/
/*		READ EVENT CALLBACK HANDLER			*/
/*								*/
/*	Very similar in concept to the monitor data event	*/
/*	callback handler above. A characteristic of this	*/
/*	code is that if a user specifices a timeout that is	*/
/*	too small, he will be told that his read timed out	*/
/*	and then the data will be delivered later. The		*/
/*	bottom line is that a read timeout does NOT cancel	*/
/*	processing of the data if it should arrive later.	*/
/*								*/
/****************************************************************/

void readHandler(args)
struct event_handler_args	args;
{
    ODSS_READ_DATA_MSG	odssReaddata;
    ODSSCTL		*currCtl = (ODSSCTL *) args.usr;
    char		*pValue;
    short		msgType;
    int			size;
    int			type;

    /*
    Stat and sevr are always in the same place in each data structure.
    */

    odssReaddata.ODSShandle = currCtl -> handle;

    /*
    Stat and sevr are always in the same place in each data structure.
    */

    odssReaddata.status = *(short *)args.dbr;
    odssReaddata.status = htons(odssReaddata.status);
    odssReaddata.severity = *(short *)(((char *)args.dbr) + sizeof(short));
    odssReaddata.severity = htons(odssReaddata.severity);
    type = ca_field_type(currCtl -> channel);
    odssReaddata.dataType = type;
    odssReaddata.dataType = htons(odssReaddata.dataType);

    /*
    Get a pointer to the value.
    */

    pValue = dbr_value_ptr(args.dbr, args.type);
    if (type == DBF_STRING)
	size = strlen(pValue) + 1;
    else
	size = dbr_size_n(type, args.count);
    odssReaddata.dataLength = htons(size);
    odssReaddata.readState = 0;
    msgType = htons(ODSS_READ_DATA);
    writeClient(msgType, (char *) &odssReaddata, sizeof(odssReaddata),
				pValue, size);
    currCtl -> readPending = 0;
    currCtl -> readTimer = 0;
}

/****************************************************************/
/*								*/
/*		WRITE EVENT CALLBACK HANDLER			*/
/*								*/
/*	To verify a write, we only determine that the write	*/
/*	made it ALL the way down to the IOC/RTU. This is done	*/
/*	by having a read transaction immendiately follow the	*/
/*	right and then catching the read event callback here.	*/
/*	We do NOT veryify that the data written is the data	*/
/*	that was sent for two reasons. First, the link layers	*/
/*	all guarantee that the data is accurately transmitted.	*/
/*	Second, there is nothing that prevents any other task	*/
/*	or CA user from writing over the same field immediately	*/
/*	after we have. Another good reason is that the ODSS	*/
/*	protocol has no definition for "Remote node is still	*/
/*	online and data came back, but it wasn't what you	*/
/*	sent" which is about as probable as me hitting the	*/
/*	lotto.							*/
/*								*/
/*	The same comments as for the read data callback		*/
/*	with respect to data arriving AFTER the timeout		*/
/*	apply here.						*/
/*								*/
/****************************************************************/

void writeHandler(args)
struct event_handler_args	args;
{
    ODSSCTL		*currCtl = (ODSSCTL *) args.usr;
    ODSS_WRITE_ACK_MSG	odssWriteAck;
    short		msgType;

    odssWriteAck.ODSShandle = currCtl -> handle;
    odssWriteAck.status = htons(0);
    msgType = htons(ODSS_WRITE_ACK);
    writeClient(msgType, (char *) &odssWriteAck, sizeof(odssWriteAck), NULL, 0);
    currCtl -> writeRdbkPending = 0;
    currCtl -> writeRdbkTimer = 0;
}
void	odssSignalHandler(int whichone)
{
time_t	end_time;
int i;
switch(whichone)
  {
  case SIGINT:
    closeClientChannels();
    close(clientFd);
    VERBOSE(V_INFO)(LOG_ALERT|LOG_LOCAL1,"Exiting due to ^C");
    exit(1);
  case SIGUSR1:
    if(++dbVerbose > V_DEBUG) dbVerbose = V_DEBUG;
    syslog(LOG_ALERT|LOG_LOCAL1,"Message level now at %s\n",msgLevel[dbVerbose]);
    break;
  case SIGUSR2:
    if(--dbVerbose < 0) dbVerbose = 0;
    syslog(LOG_ALERT|LOG_LOCAL1,"Message level now at %s\n",msgLevel[dbVerbose]);
    break;
  case SIGPIPE:
    VERBOSE(V_STATUS)(LOG_ALERT|LOG_LOCAL1,"Broken pipe on client FD <%d>\n",clientFd);
    close(clientFd);
    closeClientChannels();
    clientFd = -1;
    break;
  }
}
