/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/


/****************************************************************/
/*								*/
/*		SERIAL REDUNDANCY HANDSHAKE TASK		*/
/*								*/
/*	This is a task and a bunch of support routines to	*/
/*	communicate with a redundant partner.			*/
/*								*/
/*	Author: Gordon Uchenick					*/
/*								*/
/****************************************************************/

#include <vxWorks.h>
#include <taskLib.h>
#include <ioLib.h>
#include <tickLib.h>
#include <logLib.h>

/****** Structure to maintain status ******/

typedef struct
    {
    short	timeOut;	/* Established during attach */
    short	mustStrobe;	/* Strobe override */
    long	lastPulse;	/* Last time user pulsed */
    long	lastPartner;	/* Last time we got updated */
    char	localState;	/* 'W' = Slave Want,
				   'M' = Master,
				   'S' = Slave,
				   'T' = Timed out,
				   'U' = Unregistered */
    char	remoteState;	/* Same coding as localState */
    } TASK_STATUS;


/******	Tuning/sizing parameters ******/

#define MAX_SERIAL_TASKS	32
#define TASK_PRI		110
#define TASK_STACK		2048
#define SENTINEL		0x01


/****** Private Data ******/

static int		serialTicks;	/* Ticks between executions */
static int		serialFd;	/* FD of comm port */

static char		serialBuff[MAX_SERIAL_TASKS];
static TASK_STATUS	serialTable[MAX_SERIAL_TASKS];
static char		header = SENTINEL;

/****** Global Data ******/

int	serialRedundInfo = 0;	/* Disable messages from Serial Redund */


/****** Function Declarations ******/

void serialRedund(void);



/****************************************************************/
/*								*/
/*		INITIALIZE THE SERIAL HANDSHAKE TASK		*/
/*								*/
/****************************************************************/

int initSerialRedund(portName, ticks)
char		*portName;		/* Device name */
int		ticks;			/* Ticks between executions */
{
    int			i;

    /*
    Set the status of each task to undefined.
    */

    for (i = 0; i < MAX_SERIAL_TASKS; i++)
	{
	serialTable[i].localState = 'U';
	serialTable[i].remoteState = 'U';
	}

    /*
    Open the serial port
    */

    if ((serialFd = open(portName, O_RDWR, 0x666)) == ERROR)
	{
	logMsg("Can't open serial line for redundancy, %d\n", errno,0,0,0,0,0);
	return (ERROR);
	}

    /*
    Save the upate rate
    */

    serialTicks = ticks;

    /*
    Launch the redundant update task
    */

    if (taskSpawn("serialRedund", TASK_PRI, VX_FP_TASK, TASK_STACK, (FUNCPTR)serialRedund,
		0, 0, 0, 0, 0, 0, 0, 0, 0, 0) == ERROR)
	return (ERROR);
    return (!ERROR);
}


/****************************************************************/
/*								*/
/*		THE REDUNDANT UPDATE TASK			*/
/*								*/
/****************************************************************/

void serialRedund(void)
{
    int			i;
    int			currTime;
    int			now;
    int			bytesAvail;
    char		sentinel;

    for (;;)
	{

	/*
	Send our status to our partner.
	*/

	currTime = tickGet();
	for (i = 0; i < MAX_SERIAL_TASKS; i++)
	    {
	    if (serialTable[i].localState != 'U')
	    	if ((currTime - serialTable[i].lastPulse) > serialTable[i].timeOut)
		    serialTable[i].localState = 'T';
	    serialBuff[i] = serialTable[i].localState;
	    }

	write(serialFd, &header, sizeof(header));
	write(serialFd, serialBuff, sizeof(serialBuff));

	/*
	See if there is any data to work with. Find the sentinel
	*/

	sentinel = ~SENTINEL;
	do
	    {
	    ioctl(serialFd, FIONREAD, (int)&bytesAvail);
	    if (!bytesAvail)
		{
		taskDelay(1);
		continue;
		}
	    read(serialFd, &sentinel, sizeof(sentinel));
	    } while (((sentinel != SENTINEL) && ((tickGet() - currTime) < serialTicks)) ||
			(bytesAvail > (sizeof (serialBuff) + sizeof (sentinel))));

	if (sentinel == SENTINEL)
	    {

	    /*
	    Wait for enough data to be there, but not too long
	    */

	    do
		{
	    	ioctl(serialFd, FIONREAD, (int)&bytesAvail);
		if (bytesAvail >= sizeof(serialBuff))
		    break;
		taskDelay(1);
	    	} while ((tickGet() - currTime) < serialTicks);
	    if (serialRedundInfo)
		{
		if (bytesAvail > (sizeof (serialBuff) * 2))
		    {
		    logMsg ("Bytes Available = %d\n", bytesAvail,0,0,0,0,0);
		    }
		}
	    if (bytesAvail >= sizeof(serialBuff))
		{
		read(serialFd, serialBuff, sizeof(serialBuff));
		now = tickGet();
	        for (i = 0; i < MAX_SERIAL_TASKS; i++)
		    {
		    serialTable[i].lastPartner = now;
		    serialTable[i].remoteState = serialBuff[i];
		    }
		}
	    }

        now = tickGet();
	if ((now - currTime) < serialTicks)
	    taskDelay(serialTicks - now + currTime);
	}
}


/****************************************************************/
/*								*/
/*		UTILITY ROUTINES FOR DRIVERS			*/
/*								*/
/****************************************************************/


int attachSerialRedund(index, timeOut)
int	index;			/* Index into array, 0 <= index < MAX_SERIAL_TASKS */
int	timeOut;		/* Both for this task strobing and response from partner */
{
    if ((index < 0) || (index >= MAX_SERIAL_TASKS))
	return (ERROR);
    if (timeOut <= 0)
	return (ERROR);
    serialTable[index].timeOut = timeOut;
    return (!ERROR);
}


int pulseSerialRedund(index, localState)
int	index;			/* Index into array, same as attachSerialRedund */
int	localState;		/* 'M' = master, 'S' = slave, 'W' = slavewant */
{
    if ((index < 0) || (index >= MAX_SERIAL_TASKS))
	return (ERROR);
    if ((localState != 'M') && (localState != 'S') && (localState != 'W'))
	return (ERROR);
    serialTable[index].localState = localState;
    serialTable[index].lastPulse = tickGet();
    return (!ERROR);
}


int setStrobeSerialRedund(index)
int	index;
{
    if ((index < 0) || (index >= MAX_SERIAL_TASKS))
	return (ERROR);
    serialTable[index].mustStrobe = 1;
    return (!ERROR);
}

int clearStrobeSerialRedund(index)
int	index;
{
    if ((index < 0) || (index >= MAX_SERIAL_TASKS))
	return (ERROR);
    serialTable[index].mustStrobe = 0;
    return (!ERROR);
}

int remoteSerialRedund(index)
int	index;			/* Index into array, same as attachSerialRedund */
{
    if ((index < 0) || (index >= MAX_SERIAL_TASKS))
	return (ERROR);
    if (serialTable[index].remoteState != 'U')
	if ((tickGet() - serialTable[index].lastPartner) > (serialTicks << 1))
	    {
	    if (serialRedundInfo)
		{
		logMsg ("Partner timed out... last update %ld... current tickcount %ld\n",
			serialTable[index].lastPartner, tickGet (),0,0,0,0);
		}
	    return ('T');
	    }
    return (serialTable[index].remoteState);
}

int localSerialRedund(index)
int	index;
{
    if ((index < 0) || (index >= MAX_SERIAL_TASKS))
	return (ERROR);
    return (serialTable[index].localState);
}

int localStrobeSerialRedund(index)
int	index;
{
    if ((index < 0) || (index >= MAX_SERIAL_TASKS))
	return (ERROR);
    return (serialTable[index].mustStrobe);
}

int eraseSerialRedund(index)
int	index;
{
    if ((index < 0) || (index >= MAX_SERIAL_TASKS))
	return (ERROR);
    serialTable[index].remoteState = 'U';
    return (!ERROR);
}
