/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/


/*
  bootld.c - bootline/UDP server daemon
  1993.11.11,ktb
 *
  see header in bootlstab.template for explanations of the
  bootline resource file contents
 *
  -d    toggle debug mode flag
  -l[#] toggle logging mode flag (optional 'verbosity' level)
  -p#   set UDP port id
  -f<>  set bootline resource file
 *
  SIGINT	report STATUS (^C interactively)
  SIGHUP	refresh the bootline cache from the file
  SIGUSR1	toggle the DEBUG flag
  SIGUSR2	toggle the LOGGING flag
  SIGALRM	increment the logging level
  SIGVTALRM	reset the logging level to 1
  SIGTERM	EXIT gracefully
*/

#include <ctype.h>
#include <errno.h>
#include <fcntl.h>
#include <malloc.h>
#include <netdb.h>
#include <stdio.h>
#include <string.h>
#include <time.h>
#include <varargs.h>
#include <unistd.h>
#include <netinet/in.h>
#include <sys/file.h>
#include <sys/ioctl.h>
#include <sys/signal.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <sys/termios.h>
#include "bootline.h"

static	int	parse_switches (int, char **);
static	int	parse_bootlstab (void);
static	int	passiveUDP (char *);
static	int	svc_bootline (int);
static	int	errexit ();
static  void	leaving (void);

#define MAXVARS 10
#define MAXVARLEN 128

struct _container
{
  struct _container *next;
  bootrsp *brsp;
};

static int _debug_, _logtrans_ = 1, loglevel = 1, logmode;
static int detached, pid, lockfd;
static int linenum, numunits, nummsgs, numresps;
static struct _container *head;
static char *progpath;
static char *progname;
static char *varlist[MAXVARS];
static char *btabname = "bootlstab";
static char *lockfname = "/usr/spool/bootld/.lock";
static char *logfname = "/usr/spool/bootld/bootld.log";
static char *portid = BOOTLPORT;

static timestamp( ch )
char ch;
{
  time_t now; time( &now );
  printf( " %c %s", ch, ctime( &now ));
}

static currstate()
{
  printf( "%s/%s: (%u)%s\n", progpath, progname, pid, detached ? " daemonized" : "" );
  printf( "Transaction Logging: %s%s port(%s)\n", _logtrans_ ? "ON" : "OFF", _debug_ ? " (DEBUG MODE)" : "", portid );
  printf( "Local directory ($0): %s\n", varlist[0] );
  printf( "#Units cached: %u  #Msgs recvd: %u  #Bootlines sent: %u\n", numunits, nummsgs, numresps );
}

static closelog( reason )
char *reason;
{
  printf( reason ), timestamp( '@' );
  close(1), close(2), open( "/dev/null", O_RDWR ), dup(1);
}

static openlog()
{
  close(1), close(2);
  if( open( logfname, logmode, 0644 ) < 0 )
    errexit( "openlog" );
  dup(1);
}

static loglevel_INC()
{
  if( loglevel < 4 ) ++loglevel;
  printf( "sigALRM: logging verbosity = %d", loglevel ), timestamp( '@' );
}

static loglevel_RST()
{
  loglevel = 1;
  printf( "sigVTALRM: logging verbosity reset" ), timestamp( '@' );
}

static toggle_debug()
{
  _debug_ = ! _debug_;
  if( detached )
  {
    if( ! _logtrans_ )
    {
      if( ! _debug_ )
      {
        closelog( "sigUSR1 (toggle_debug)\nLOGGING & DEBUG OFF" );
        return;
      }
      else
        openlog();
    }
  }
  printf( "sigUSR1 (toggle_debug) %s", _debug_ ? "ON" : "OFF" ), timestamp( '@' );
}

static toggle_logtrans()
{
  _logtrans_ = ! _logtrans_;
  if( detached )
  {
    if( ! _debug_ )
    {
      if( ! _logtrans_ )
      {
        closelog( "sigUSR2 (toggle_logtrans)\nLOGGING & DEBUG OFF" );
        return;
      }
      else
        openlog();
    }
  }
  printf( "sigUSR2 (toggle_logtrans) %s", _logtrans_ ? "ON" : "OFF" ), timestamp( '@' );
}

static reload_bootlstab()
{
  if( _debug_ || _logtrans_ || ! detached )
    printf( "sigHUP (reload_bootlstab)" ), timestamp( '@' );
  if( _logtrans_ && loglevel > 2 )
    closelog( "" ), openlog(); /*truncate*/
  parse_bootlstab();
  if( _debug_ || (_logtrans_ && loglevel > 1) || ! detached )
    currstate();
}

static ctrlC()
{
  printf( "\b\bSTATUS:" ), timestamp( '\n' );
  currstate();
}

static term()
{
  if( detached )
  {
    if( ! (_debug_ || _logtrans_) )
      openlog();
  }
  leaving ();
  exit(0);
}

static void leaving(void)
{
  static char *byebye = ":exited\n";
  if( loglevel > 1 )
  {
    printf( "------------------------------------------------------------" );
    timestamp( '\n' );
    printf( "terminating...\n" );
    currstate();
    printf( "------------------------------------------------------------\n" );
  }
  else printf( "terminating" ), timestamp( '@' );
  if( detached )
    write( lockfd, byebye, strlen( byebye ));
}


main( int ac, char *av[] )
{
  fd_set readfds, activefds;
  int bootlinesock, fd = 1, numfds = getdtablesize();
  printf( "bootline server\n" );
  parse_switches( ac, av );
 /**/
  if( _debug_ )
    pid = getpid();
  else
  {
    register int i = fork();
    if( i )
    {
      if( i < 0 )
        errexit( "daemonize" );
      printf( "daemon pid:%u\n", i );
      leaving ();
      exit(0);
    }

#ifdef SOLARIS
    setpgrp();
#else
#  ifdef AXP
    setpgrp();
    setpgid(0,0);
#  else
    setpgrp( 0, pid);
#  endif AXP
#endif	SOLARIS

#ifdef	SOLARIS
    if( (lockfd = open( lockfname, O_RDWR|O_CREAT|O_TRUNC, 0644 )) < 0 || 
				((lockf( lockfd, F_TLOCK, 0 )) == -1) )
#else
    if( (lockfd = open( lockfname, O_RDWR|O_CREAT|O_TRUNC, 0644 )) < 0 || 
				((flock( lockfd, LOCK_EX|LOCK_NB )) == -1) )
#endif	SOLARIS
    {
      errexit( "lockfile" );
    }
    else
    {
      char buff[40];
      sprintf( buff, "bootld:%u\n", pid );
      write( lockfd, buff, strlen(buff) );
    }
    i = open( "/dev/tty", O_RDWR ), ioctl( i, TIOCNOTTY, 0 );
    for( i = 0; i < numfds; ++i )
    {
      if( i != lockfd )
        close( i );
    }
    i = open( "/dev/null", O_RDWR ), dup( i ), dup( i );
    detached = 1;
    if( _logtrans_ )
      openlog();
  }
  signal( SIGQUIT, SIG_IGN );
  signal( SIGTSTP, SIG_IGN );
  signal( SIGTERM,  term );
  signal( SIGINT,  ctrlC );
 /**/
  parse_bootlstab();
  signal( SIGHUP, reload_bootlstab );
  signal( SIGUSR1, toggle_debug );
  signal( SIGUSR2, toggle_logtrans );
  signal( SIGALRM, loglevel_INC );
  signal( SIGVTALRM, loglevel_RST );
  if( detached && (_debug_ || _logtrans_) )
    ctrlC();
  FD_ZERO(&activefds);
  bootlinesock = passiveUDP( portid );
  if( setsockopt( bootlinesock, SOL_SOCKET, SO_BROADCAST, 
					(const char *)&fd, sizeof(int) ) < 0 )
    errexit( "setsockopt(BROAD)" );
  if( setsockopt( bootlinesock, SOL_SOCKET, SO_REUSEADDR, 
					(const char *)&fd, sizeof(int) ) < 0 )
    errexit( "setsockopt(REUSE)" );
  FD_SET(bootlinesock, &activefds);
  if( _debug_ )
    printf( "active on BOOTLINE port (%s)...\n", portid );
 /**/
  for(;;)
  {
    if( detached )
      fflush( stdout ), fflush( stderr );
    bcopy( (char*) &activefds, (char*) &readfds, sizeof(fd_set) );
    if( select( numfds, &readfds, (fd_set*) 0, (fd_set*) 0, (struct timeval*) 0 ) < 0 )
    {
      if( errno == EINTR ) continue;
      errexit( "select" );
    }
   /**/
    if( FD_ISSET(bootlinesock, &readfds) )
    {
      if( svc_bootline( bootlinesock ) < 0 )
        errexit( "main(svc_bootline)" );
    }
  }
}


static int parse_switches( ac, av )
int ac;
char **av;
{
  static char wpath[132+1];
  register int i;
  if( **av != '/' )
  {
    getwd( wpath, sizeof(wpath) );
    while( **av == '.' )
    {
      if( *(*av += 2) == '/' )
        ++(*av), *strrchr( wpath, '/' ) = 0;
    }
    progpath = wpath, progname = *av;
  }
  else
  {
    progname = strrchr( *av, '/' ); *progname++ = 0;
    if( strlen( progpath = *av ) == 0 )
      strcpy( progpath = wpath, "/" );
  }
  for( i = 0; i < MAXVARS; ++i )
    varlist[i] = calloc( 1, MAXVARLEN+1 );
  getwd( varlist[0], MAXVARLEN );
 /**/
  while( --ac )
  {
    if( **(++av) == '-' )
    {
      ++(*av);
      switch( **av )
      {
       case 'd': _debug_ = ! _debug_;
        continue;
       case 'p': portid = *av + 1;
        continue;
       case 'f': btabname = *av + 1;
        continue;
       case 'l':
        _logtrans_ = *(++*av) ? 1 : ! _logtrans_;
        loglevel = _logtrans_ ? atoi(*av) : 1;
        if( _logtrans_ && loglevel <= 0 ) loglevel = 1;
        continue;
      }
    }
    printf( "** don't recognize: (%s)\n", *av );
    printf( "usage: %s [-d] [-l] [-p#] [-f<file>]\n" );
    printf( "  -d toggle debug, -l toggle logging\n" );
    printf( "  -p# specify UDP port number\n" );
    printf( "  -f<file> specify resource file\n" );
    leaving ();
    exit(1);
  }
  logmode = O_WRONLY | O_CREAT | O_APPEND | (_logtrans_ > 2 ? O_TRUNC : 0);
  if( ! _debug_ )
  {
    printf( "signals supported:\n" );
    printf( "  INT   STATUS report (by ^C interactively)\n" );
    printf( "  HUP   reload resource file contents\n" );
    printf( "  USR1  toggle DEBUG mode\n" );
    printf( "  USR2  toggle LOGGING mode\n" );
    printf( "  TERM  EXIT with style\n" );
    printf( "\n" );
  }
}


static char *dot_addr( ip )
int ip;
{
  static char buff[16];
  sprintf( buff, "%u.%u.%u.%u", (ip>>24)&0xff, (ip>>16)&0xff, (ip>>8)&0xff, ip&0xff );
  return buff;
}

static bootrsp *bootlookupname( req_name )
char *req_name;
{
  register struct _container *p;
  for( p = head; p; p = p-> next )
  {
    if( strcmp( p-> brsp-> target_name, req_name ) == 0 )
      return p-> brsp;
  }
  return 0;
}

static bootrsp *bootlookup( req_IPaddr )
unsigned long req_IPaddr;
{
  register struct _container *p;
  for( p = head; p; p = p-> next )
  {
    if( p-> brsp-> target_IPaddr == req_IPaddr )
      return p-> brsp;
  }
  return 0;
}

static char *bindvars( val )
char *val;
{
  static char tmp[1024+1];
  register char *sp, *dp = tmp;
  register int clen = strlen( strcpy( tmp, val ));
  while( (sp = strchr( val, '$' )) != 0 )
  {
    register int n = *(sp+1) - '0', vlen;
    if( n < 0 || n > 9 || (clen + (vlen = strlen( varlist[ n ] ))) >= sizeof(tmp) )
      return 0;
  /**/
    dp += (int) (sp - val);
/*ANSI dp += sprintf( dp, "%s", varlist[ n - '0' ] );
*/
    strcpy( dp, varlist[ n ] ), dp += vlen;
    strcpy( dp, sp += 2 ), val = sp, clen += vlen;
  }
  return tmp;
}


static showunit( p )
bootrsp *p;
{
  if( loglevel < 2 ) return;
  printf( "------------------------------------------------------------\n" );
  if( loglevel < 3 )
  {
    printf( "\te=%s:%08lx\n", dot_addr( ntohl(p-> target_IPaddr) ), p-> target_subnetmask );
    printf( "\ttn=\"%s\"\n", p-> target_name );
  }
  else
  {
    printf( "\t\"%s\"(%d)\n", p-> bootinterface, ntohs(p-> proc_num) );
    printf( "\t\"%s:%s\"\n", p-> boothost_name, p-> imagefile );
    printf( "\te=%s:%08lx\n", dot_addr( ntohl(p-> target_IPaddr) ), p-> target_subnetmask );
    printf( "\tb=%s:%08lx\n", dot_addr( ntohl(p-> bkplane_IPaddr) ), p-> bkplane_subnetmask );
    printf( "\th=%s\n", dot_addr( ntohl(p-> boothost_IPaddr) ));
    printf( "\tg=%s\n", dot_addr( ntohl(p-> gateway_IPaddr) ));
    printf( "\ttn=\"%s\"\n", p-> target_name );
    printf( "\tu=\"%s\"\n", p-> default_user );
    printf( "\tpw=\"%s\"\n", p-> default_pass );
    printf( "\ts=\"%s\"\n", p-> scriptfile );
    printf( "\tf=%04hx\n", p-> boot_flags );
    printf( "\to=\"%s\"\n", p-> other );
  }
  printf( "------------------------------------------------------------\n" );
}

static cnv_XX(val,p)char*val;u_short*p;{u_short n;sscanf(val,"%hi",&n);*p=htons(n);}
static cnv_IP(val,p)char*val;u_long*p;{*p=htonl(inet_addr(val));}
static cnv_SM(val,p)char*val;u_long*p;{u_long n;sscanf(val,"%lx",&n);*p=htonl(n);}
static cnv_IF(val,p)char*val,*p;{sprintf(p,"%0.*s",LEN_IFCNAME-1,val);}
static cnv_NN(val,p)char*val,*p;{sprintf(p,"%0.*s",LEN_NODENAME-1,val);}
static cnv_FP(val,p)char*val,*p;{sprintf(p,"%0.*s",LEN_FILEPATH-1,val);}
static cnv_UN(val,p)char*val,*p;{sprintf(p,"%0.*s",LEN_USERNAME-1,val);}
static cnv_PW(val,p)char*val,*p;{sprintf(p,"%0.*s",LEN_PASSWORD-1,val);}
static cnv_MD(val,p)char*val,*p;{sprintf(p,"%0.*s",LEN_MISCDATA-1,val);}

static lkup_tIP( p, df, nDB )
bootrsp *p, *df;
struct hostent *nDB;
{ if( nDB ) bcopy( nDB-> h_addr_list[0], (char*) &p-> target_IPaddr, nDB-> h_length ); }

static lkup_hst( p, df, nDB )
bootrsp *p, *df;
struct hostent *nDB;
{
  if( ! _debug_ )
    printf( "%5u: ", linenum );
  printf( "\ttrying \"%s\" for boothost name...", dot_addr( ntohl(p-> boothost_IPaddr) ));
  if( (nDB = gethostbyaddr( (const char *)&p-> boothost_IPaddr, 
						sizeof(u_long), AF_INET )) == 0 )
  {
    printf( "\ttrying \"%s\" for boothost name...", dot_addr( ntohl(df-> boothost_IPaddr) ));
    if( (nDB = gethostbyaddr( (const char *)&df-> boothost_IPaddr, 
						sizeof(u_long), AF_INET )) == 0 )
      { printf( "failed...\n" ); return; }
  }
  sprintf( p-> boothost_name, "%0.*s", LEN_NODENAME-1, nDB-> h_name );
  printf( "found (%s)\n", p-> boothost_name );
}

static lkup_hIP( p, df, nDB )
bootrsp *p, *df;
struct hostent *nDB;
{
  if( ! _debug_ )
    printf( "%5u: ", linenum );
  printf( "\ttrying \"%s\" for boothost IPaddr...", p-> boothost_name );
  if( (nDB = gethostbyname( p-> boothost_name )) == 0 )
  {
    printf( "\ttrying \"%s\" for boothost IPaddr...", df-> boothost_name );
    if( (nDB = gethostbyname( df-> boothost_name )) == 0 )
      { printf("failed...\n" ); return; }
  }
  bcopy( nDB-> h_addr_list[0], (char*) &p-> boothost_IPaddr, nDB-> h_length );
  printf( "found (%s)\n", dot_addr( ntohl(p-> boothost_IPaddr) ));
}


static parse_token( tkn, val, wrk, def, netDB )
char *tkn, *val;
bootrsp *wrk, *def;
struct hostent *netDB;
{
# define OFF(mem) ((int)&((bootrsp*)0)->mem)
# define LEN(mem) sizeof(((bootrsp*)0)->mem)
 /**/
  static struct _parsetable { char *op; int off, len, clrOK, defOK, (*cnv)(), (*lkup)(); } tkntbl[] =
  {
    { "ifc",    OFF(bootinterface[0]),   LEN(bootinterface),      0, 1, cnv_IF,        0 }, /* e.g. "ln"             */
    { "trgIP",  OFF(target_IPaddr),      LEN(target_IPaddr),      0, 0, cnv_IP, lkup_tIP }, /* e.g. "192.246.18.220" */
    { "trgSNM", OFF(target_subnetmask),  LEN(target_subnetmask),  1, 1, cnv_SM,        0 }, /* e.g. "ffffff00"       */
    { "bkpIP",  OFF(bkplane_IPaddr),     LEN(bkplane_IPaddr),     1, 0, cnv_IP,        0 }, /* e.g. "192.246.19.220" */
    { "bkpSNM", OFF(bkplane_subnetmask), LEN(bkplane_subnetmask), 1, 1, cnv_SM,        0 }, /* e.g. "ffff0000"       */
    { "hst",    OFF(boothost_name[0]),   LEN(boothost_name),      0, 1, cnv_NN, lkup_hst }, /* e.g. "bud"            */
    { "hstIP",  OFF(boothost_IPaddr),    LEN(boothost_IPaddr),    0, 1, cnv_IP, lkup_hIP }, /* e.g. "192.246.18.1"   */
    { "gtwyIP", OFF(gateway_IPaddr),     LEN(gateway_IPaddr),     1, 1, cnv_IP,        0 }, /* e.g. "192.246.17.9"   */
    { "os",     OFF(imagefile[0]),       LEN(imagefile),          0, 1, cnv_FP,        0 }, /* e.g. "/tis/vw510/config/msol/vxWorks" */
    { "scr",    OFF(scriptfile[0]),      LEN(scriptfile),         1, 1, cnv_FP,        0 }, /* e.g. "/epics/project/ioc1/start.cmd"  */
    { "usr",    OFF(default_user[0]),    LEN(default_user),       1, 1, cnv_UN,        0 }, /* e.g. "epics"          */
    { "pass",   OFF(default_pass[0]),    LEN(default_pass),       1, 1, cnv_PW,        0 }, /* e.g. "xyzzy"          */
    { "misc",   OFF(other[0]),           LEN(other),              1, 1, cnv_MD,        0 }, /* usually null          */
    { "proc",   OFF(proc_num),           LEN(proc_num),           1, 1, cnv_XX,        0 }, /* usually 0             */
    { "flgs",   OFF(boot_flags),         LEN(boot_flags),         1, 1, cnv_XX,        0 }, /* usually 0             */
    { 0,        0,                       0,                       0, 0,      0,        0 }
  };
 /**/
  register struct _parsetable *tbl;
  register char *oper;
  for( tbl = tkntbl; (oper = tbl-> op) != 0; ++tbl )
  {
    register int off = tbl-> off, len = tbl-> len;
    if( strcmp( tkn, oper ) == 0 )
    {
     /*clear field?*/
      if( strcmp( val, "-" ) == 0 )
      {
        if( tbl-> clrOK )
          bzero( (char*) wrk + off, len );
        else
        {
          if( ! _debug_ )
            printf( "%5u: ", linenum );
          printf( "** value required (%s)\n", oper );
        }
      }
      else
     /*use default?*/
      if( strcmp( val, "*" ) == 0 )
      {
        if( tbl-> defOK )
        {
          if( wrk != def )
            bcopy( (char*) def + off, (char*) wrk + off, len );
        }
        else
        {
          if( ! _debug_ )
            printf( "%5u: ", linenum );
          printf( "** default value disallowed (%s)\n", oper );
        }
      }
      else
     /*perform lookup?*/
      if( strcmp( val, "?" ) == 0 )
      {
        if( tbl-> lkup != 0 )
        {
          bzero( (char*) wrk + off, len );
          (*tbl-> lkup)( wrk, def, netDB );
        }
        else
        {
          if( ! _debug_ )
            printf( "%5u: ", linenum );
          printf( "** no lookup defined (%s)\n", oper );
          return;
        }
      }
      else
     /*user gave us something...*/
      {
        register char *tmp = bindvars( val );
        bzero( (char*) wrk + off, len );
        if( tmp == 0 )
        {
          if( ! _debug_ )
            printf( "%5u: ", linenum );
          printf( "** invalid var reference (%s)\n", oper );
          return;
        }
        (*tbl-> cnv)( tmp, (char*) wrk + off );
      }
      break;
    }
  }
}


static void parse_define_var( cp )
char *cp;
{
  register int n = *++cp - '0';
  if( n >= 1 && n <= 9 && *(cp + 1) == '=' )
  {
    register char *vp = bindvars( cp + 2 );
    if( vp != 0 && strlen( vp ) <= MAXVARLEN )
    {
      strcpy( varlist[ n ], vp );
      return;
    }
  }
  if( ! _debug_ )
    printf( "%5u: ", linenum );
  printf( "** improper var definition (%s)\n", cp - 1 );
}


static bootrsp *parse_bootlineentry( cp, def )
char *cp;
bootrsp *def;
{
  static bootrsp cur;
  register char *wp;
  register bootrsp *work = &cur;
  struct hostent *netinfo;
  char buff[1024+1];
  int bcount = strlen( cp );
 /**/
  while( isspace(cp[ --bcount ]) )
    /**/;
  cp[ bcount+1 ] = 0;
  while( isspace(*cp) )
    ++cp;
  if( (wp = strchr( cp, '#' )) != 0 )
    *wp = 0;
  if( strlen( cp ) == 0 )
    return 0;
  if( *cp == '$' )
  {
    parse_define_var( cp );
    return 0;
  }
 /**/
  cur = *def;
  if( strncmp( cp, "default", 7 ) == 0 )
  {
    work = def, cp += 7;
  }
  else
  {
    sscanf( cp, "%s%n", buff, &bcount ), cp += bcount;
    sprintf( cur.target_name, "%0.*s", LEN_NODENAME-1, buff );
    if( (netinfo = gethostbyname( cur.target_name )) != 0 )
      bcopy( netinfo-> h_addr_list[0], (char*) &cur.target_IPaddr, netinfo-> h_length );
  }
 /**/
  while( sscanf( cp, "%s%n", buff, &bcount ) > 0 && bcount )
  {
    if( (wp = strchr( buff, '=' )) == 0 )
    {
      if( ! _debug_ )
        printf( "%5u: ", linenum );
      printf( "** improper format: (%s)\n", buff );
    }
    else
    {
      *wp++ = 0;
      parse_token( buff, wp, work, def, netinfo );
    }
    cp += bcount;
  }
 /**/
  if( work == def )
  {
    if( _debug_ || (_logtrans_ && loglevel > 1) )
      printf( "default entry:\n" ), showunit( def );
    return 0;
  }
 /**/
  buff[0] = 0;
  if( strlen( cur.bootinterface ) == 0 ) strcat( buff, " ifc" );
  if( strlen( cur.target_name )   == 0 ) strcat( buff, " trg" );
  if(         cur.target_IPaddr   == 0 ) strcat( buff, " trgIP" );
  if( strlen( cur.boothost_name ) == 0 ) strcat( buff, " hst" );
  if(         cur.boothost_IPaddr == 0 ) strcat( buff, " hstIP" );
  if( strlen( cur.imagefile )     == 0 ) strcat( buff, " os" );
 /**/
  if( strlen( buff ) != 0 )
  {
    if( ! _debug_ )
      printf( "%5u: ", linenum );
    printf( "** required values missing:%s\n", buff );
    if( netinfo == 0 )
      printf( "**** possibly because unit (%s) was not in hosts file\n", cur.target_name );
    return 0;
  }
 /**/
  if( bootlookup( cur.target_IPaddr ) != 0 || bootlookupname( cur.target_name ) != 0 )
  {
    if( ! _debug_ )
          printf( "%5u: ", linenum );
    printf( "** duplicate or mismatch: (%s) (%s)\n", cur.target_name, dot_addr( ntohl(cur.target_IPaddr) ));
    return 0;
  }
 /**/
  return &cur;
}


static int parse_bootlstab()
{
  struct hostent *hostinfo;
  bootrsp defboot, *new;
  char buff[1024+1];
  FILE *fp;
 /**/
  linenum = numunits = 0;
  while( head != 0 )
  {
    register struct _container *tmp = head-> next;
    cfree( head-> brsp ), cfree( head );
    head = tmp;
  }
 /**/
  printf( "current directory: (%s)\n", varlist[0] );
  sprintf( buff, "./%s", btabname );
  if( (fp = fopen( buff, "r" )) == 0 )
  {
    sprintf( buff, "%s/%s", progpath, btabname );
    if( (fp = fopen( buff, "r" )) == 0 )
    {
      sprintf( buff, "/usr/etc/%s", btabname );
      if( (fp = fopen( buff, "r" )) == 0 )
      {
        sprintf( buff, "/etc/%s", btabname );
        if( (fp = fopen( buff, "r" )) == 0 )
          errexit( "parse_bootlstab" );
      }
    }
  }
  printf( "bootline resource file: (%s)\n", buff );
 /**/
  sethostent( 1 );
  bzero( (char*) &defboot, sizeof(bootrsp) );
  defboot.msgtype = htons(0xBABE);
  strcpy( defboot.bootinterface, "ln" );
  strcpy( defboot.target_name, "<default>" );
  gethostname( defboot.boothost_name, LEN_NODENAME );
  if( (hostinfo = gethostbyname( defboot.boothost_name )) != 0 )
    bcopy( hostinfo-> h_addr_list[0], (char*) &defboot.boothost_IPaddr, hostinfo-> h_length );
  defboot.target_subnetmask = htonl(0xffffff00);
 /**/
  while( ! feof( fp ) )
  {
    register int end = 0, OK;
    buff[0]=0;
    do {
      ++linenum;
      if( OK = (fgets( buff + end, sizeof(buff) - end, fp ) != 0) )
        buff[ end = strlen( buff ) - 1 ] = 0;
    } while( OK && buff[ --end ] == '\\' );
    if( ! OK ) break;
   /**/
    if( _debug_ )
      printf( "%5u: <%0.*s%c\n", linenum, detached ? end : 70, buff, detached ? '>' : (end <= 70 ? '>' : '|') );
    if( (new = parse_bootlineentry( buff, &defboot )) != 0 )
    {
      struct _container *p = (struct _container*) calloc( 1, sizeof(struct _container) );
      *(p-> brsp = (bootrsp*) calloc( 1, sizeof(bootrsp) )) = *new;
      p-> next = head, head = p;
      if( _debug_ || (_logtrans_ && loglevel > 1) )
        printf( "new entry:\n" ), showunit( new );
      ++numunits;
    }
  }
 /**/
  if( _debug_ || _logtrans_ || ! detached )
    printf( "%d line(s) processed, %d unit(s) loaded\n", linenum, numunits );
  fclose( fp );
  endhostent();
}


static int passiveUDP( port )
char *port;
{
  struct sockaddr_in sin;
  int sock;
 /**/
  bzero( (char*) &sin, sizeof(struct sockaddr_in) );
  sin.sin_family = AF_INET;
  sin.sin_addr.s_addr = htonl(INADDR_ANY);
  sin.sin_port = htons(atoi( port ));
 /**/
  if( (sock = socket( PF_INET, SOCK_DGRAM, 0 )) < 0 )
    errexit( "socket UDP" );
  if( bind( sock, (struct sockaddr*) &sin, sizeof(struct sockaddr_in) ) < 0 )
    errexit( "bind UDP" );
  return sock;
}

static int svc_bootline( fd )
int fd;
{
  struct sockaddr_in from;
  int alen = sizeof(struct sockaddr_in);
  bootreq req;
  bootrsp *rsp;
  ++nummsgs;
  if( recvfrom( fd, (char *)&req, sizeof(bootreq), 0, 
					(struct sockaddr*) &from, &alen ) < 0 )
    return -1;
  if( _debug_ )
  {
    timestamp( '\n' );
    printf( "recv: 0x%04x | from: %u @ %s\n", 
		ntohs(req.msgtype), ntohs(from.sin_port), 
		dot_addr( ntohl(from.sin_addr.s_addr) ));
  }
  if( req.msgtype != htons(0xDEAD) )
    return 0;
 /**/
  if( (rsp = bootlookup( from.sin_addr.s_addr )) != 0 )
  {
    if( sendto( fd, (const char *)rsp, sizeof(bootrsp), 0, 
					(struct sockaddr*) &from, alen ) < 0 )
      return -1;
    ++numresps;
    if( _debug_ || _logtrans_ )
    {
      printf( "resp# %u: %s(%s/%u)", 
		numresps, rsp-> target_name, dot_addr( ntohl(from.sin_addr.s_addr)), 
		ntohs(from.sin_port) );
      timestamp( '@' );
      if( _debug_ || (_logtrans_ && loglevel > 1) )
        showunit( rsp );
    }
  }
  return 0;
}

static int errexit( fmt, va_alist )
char *fmt;
va_dcl
{
  va_list parms;
  va_start( parms );
  fprintf( stderr, "\n** %s @ ", progname );
  vfprintf( stderr, fmt, parms );
  perror( " " );
  leaving ();
  exit(1);
}
