/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/


/* ar_schedule.c - ar_schedule handling.
**
** ar_schedules are specified in place of a numeric rate field
** in the ar_reduc file to define a set of successive rates that
** reduce the storage requirements for data past a certain age.
**
** The required information is maintained in a plain text file
** named /tis4000db/ar_schedule.dat with structure:
** <schedule_name>:
**	<age_value> <interval_value>
**	...
** <schedule_name>
**	...
**
** NOTES:
**	schedule names are currently restricted to 5 characters
**		(for compatibility with ar_reduc.dbf)
**	the first age value in each schedule must be 0, and
**		defines the rate value used while data is
**		collected by the trendProcessor.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <malloc.h>
#include "ar_schedule.h"

#ifndef TRUE
#define TRUE	(1==1)
#endif TRUE
#ifndef FALSE
#define FALSE	(1==0)
#endif FALSE


static char inpstring [132];

/* routines */

/* get array of schedule names */
char	**GetScheduleNames ()
{
    FILE *fp;
    int	ix;
    char *cp;
    char *sp;
    char **newPtr;
    char **retPtr = NULL;
    int nstrings = 0;

#ifdef WIN32
    fp = fopen ("\\tis4000db\\ar_schedule.def", "rt");
#else
    fp = fopen ("/tis4000db/ar_schedule.def", "rt");
#endif WIN32
    if (fp)
    {
        while (!feof (fp))
        {
            ix = 0;
            while (!feof (fp))
            {
                inpstring [ix] = fgetc(fp);
                if ((inpstring [ix] == 0x0d) || (inpstring [ix] == 0x0a) 
		  || (inpstring [ix] == -1))
		{
                    break;
                }
		ix++;
                if (ix == 80)
                    return NULL; /* bad string in file. */
            }
	    inpstring [ix] = 0;

	    /* process input line */
	    sp = inpstring;
	    while (*sp && (*sp == ' ' || *sp == 0x09))
		sp++;
	    if ((*sp == 0) || (*sp == '#'))
		continue; /* blank line or comment! */

	    /* if contains one or more ':' chars, it is a schedule name */
	    if ((cp = strchr (sp, ':')) != NULL)
	    {
		if ((nstrings % 20) == 0)
		{
		    /* expand for 20 more strings */
		    if ((newPtr = calloc (nstrings + 20, sizeof(char *))) == NULL)
		    {
			for (ix = 0; ix < nstrings; ix++)
			    free (retPtr[ix]);
			free (retPtr);
			return NULL;
		    }
		    else
		    {
			for (ix = 0; ix < nstrings; ix++)
			    newPtr[ix] = retPtr[ix];
			free (retPtr);
			retPtr = newPtr;
		    }
		}
		*cp = 0;
		if ((retPtr [nstrings] = malloc (strlen (sp) + 1)) == NULL)
		{
		    for (ix = 0; ix < nstrings; ix++)
			free (retPtr[ix]);
		    free (retPtr);
		    return NULL;
		}
		else
		{
		    strcpy (retPtr [nstrings], sp);
		    nstrings++;
		}
	    }
	}
        fclose (fp);
    } /* end of opened file code */
    return retPtr;
}

void FreeScheduleNames (char **namePtr)
{
    int ix;

    for (ix = 0; namePtr[ix]; ix++)
	free (namePtr[ix]);
    free (namePtr);
}

/* get initial reduction rate entry */
int	GetBasicRate (char *schedName)
{
    REDUC_SCHEDULE sched;

    if (GetSchedule (schedName, &sched) && sched.numschedule)
	return (sched.schedule[0].interval);
    else
	return 0;
}

/* get complete schedule data */
int	GetSchedule (char *schedName, REDUC_SCHEDULE *schedule)
{
    FILE *fp;
    int	ix;
    char *cp;
    char *sp;
    int	have_schedule = FALSE;

#ifdef WIN32
    fp = fopen ("\\tis4000db\\ar_schedule.def", "rt");
#else
    fp = fopen ("/tis4000db/ar_schedule.def", "rt");
#endif WIN32
    if (fp)
    {
        while (!feof (fp))
        {
            ix = 0;
            while (!feof (fp))
            {
                inpstring [ix] = fgetc(fp);
                if ((inpstring [ix] == 0x0d) || (inpstring [ix] == 0x0a) 
		  || (inpstring [ix] == -1))
		{
                    break;
                }
		ix++;
                if (ix == 80)
                    return have_schedule; /* bad string in file. */
            }
	    inpstring [ix] = 0;

	    /* process input line */
	    sp = inpstring;
	    while (*sp && (*sp == ' ' || *sp == 0x09))
		sp++;
	    if ((*sp == 0) || (*sp == '#'))
		continue; /* blank line or comment! */

	    /* if contains one or more ':' chars, it is a schedule name */
	    if ((cp = strchr (sp, ':')) != NULL)
	    {
		if (have_schedule)
		    return (TRUE);
		/* start of a set of reduc rates */
		*cp = 0;
		if (strcmp (schedName, sp) == 0)
		{
		    have_schedule = TRUE;
		    strncpy (schedule->name, schedName, MAX_NAME);
		    schedule->numschedule = 0;
		}
	    }
	    else if (have_schedule && ((cp = strchr (sp, ',')) != NULL)) /* must be an age, rate pair */
	    {
		schedule->schedule[schedule->numschedule].age = atoi (sp);
		schedule->schedule[schedule->numschedule].interval = atoi(cp+1);
		schedule->numschedule++;
	    }
	}
        fclose (fp);
    } /* end of opened file code */
    return have_schedule;
}

/* write a schedule */
int	PutSchedule (REDUC_SCHEDULE *schedule)
{
    /* copy file info until item found, or end of file */
    /* update with new info */
    return 0; /* unimplemented */
}

/* delete a schedule */
int	RemoveSchedule (char *schedName)
{
    /* copy file info until item found, or end of file */
    /* removee the info */
    return 0; /* unimplemented */
}

