/*
Copyright (c) 1992 Thomas A. Limoncelli, tom_limoncelli@warren.mentorg.com

Permission to use, copy, modify, distribute, and sell this software and
its documentation for any purpose is hereby granted without fee,
provided that (i) the above copyright notices and this permission
notice appear in all copies of the software and related documentation,
and (ii) the name of Mentor Graphics Corp or any company or
organization that I am associated with may not be used in any
advertising or publicity relating to the software without the specific,
prior written permission of that company or organization

THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,
EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY
WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.

IN NO EVENT SHALL THOMAS A. LIMONCELLI OR MENTOR GRAPHICS CORP BE
LIABLE FOR ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES
OF ANY KIND, OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA
OR PROFITS, WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON
ANY THEORY OF LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE
OR PERFORMANCE OF THIS SOFTWARE.
*/

/******************************************************************************
  tapixo.c


*******************************************************************************/

#include <stdio.h>
#include <fcntl.h>
#include <string.h>
#include <ctype.h>

/* ASCII constants */
#define STX (2)
#define EOT (4)
#define ACK (6)
#define LF (10)
#define CR (13)
#define NAK (21)
#define ESC (27)
#define RS (30)

#define MAX_PACKET	(10000)	/* we'll never get a packet this big */
#define MAXLINE	(1000)
#define MAX_ERROR  1000
/***************************************************************************
calculate the checksum of the packet.

Each checksum is computed by performing the simple arithmetic sum of
7-bit values of all characters preceding it in that block.  (This
means that the STX and ETB/ETX are included in the sum).  The checksum
is then the least significant 12 bits of this resulting sum.

The checksum is transmitted as three printable ASCII characters having
values between HEX 30 and HEX 3F (the characters 0123456789:;<=>?).
The most significant four bits of the sum are encoded as the 4 LSB of
the third chacter. 

SAMPLE Checksum Example:

STX   000 0010
1     011 0001
2     011 0010
3     011 0011
CR    000 1101
A     100 0001
B     100 0010
C     100 0011
CR    000 1101
ETX   000 0011
    10111 1011
    1   7   ;

So, the checksum is "17;"
****************************************************************************/ 

char *checksum(pk)
char *pk;
{
	static char check[10];
	int sum = 0;

	for (;*pk; pk++) sum += *pk;
	check[2] = '0' + (sum & 15); sum = sum >> 4;
	check[1] = '0' + (sum & 15); sum = sum >> 4;
	check[0] = '0' + (sum & 15);
	check[3] = 0;

/* fprintf(stderr, "CHECKSUM=:"); safeprint(check); fprintf(stderr,":\n");  */
	return check;
}

/******************************************************************************/
int modem_login(modem, error_message)
int modem;
char error_message[MAX_ERROR+1];
{

	char pack[MAX_PACKET];
	int mesnum, failcnt, len;
	char c;

	failcnt = 3;
	while (failcnt--) {

              /* FOR AUTOMATIC REMOTE ENTRY:

                 Remote sends ESC SST  (typically ESC PG1)
                 ESC signifies entry device intends to talk in automatic dump mode.

                 SS is a set of two alphanumeric characters signifying a type of
                 service to be accessed:

                 For a paging service where:
                 Field 1 = Pager ID 
                 Field 2 = Message
                 SS will be sent at "PG".

                 Where T is a single alphanumeric character relating to the type of
                 terminal or device attempting to send the message.  T = "1" is a
                 category of entry devices using the same protocol.  The PETs and IXO
                 devices are members of this category.  T = "7" or "8" or "9" are
                 reserved for wild card terminals or devices which may relate to a
                 specific users' system.

                 Remote sends 6-char alphanumeric password then CR.  (Password is
                 optional and is, in general, reserved for future services.  Password
                 may be interpreted as either a caller ID or a system entry key.
                 Length of password, when used, may be different in some systems.)
                */

		/* fprintf(stderr, "Sending ESC\n");  */
		write(modem, "\033", 1);
		send_data(modem, "PG1000000\r");


                /*Central then sends one of three sequences:
                  CR ACK CR         This means "logon accepted"
                  CR NAK CR         This means "requested again"
                  CR ESC EOT CR     This means "Forced disconnect"
		*/ 

	/* fprintf(stderr, "Waiting for acceptance (aren't we all)\n");  */
		get_packet(modem, pack);
		     
		len = strlen(pack);
		if ((len==1) && (*pack == ACK)) {
			/* fprintf(stderr, "login successful\n");  */
			break;
		}
		if ((len==1) && (*pack == NAK)) {
			/* fprintf(stderr, "retrying to log in.\n");  */
			continue;
		}
		if ((len==2) && (*pack == ESC) && (pack[1] == EOT)) {
			/* fprintf(stderr, "forced disconnect\n");  */
			sprintf(error_message,"Paging service disconnected.");
			hangup_modem(modem);
			return 0;
		}
                /* When an incorrect logon sequence beginning with ESC is received, the
                 paging central may respond with an ID= if it requires a
                 retransmission.

                 Central responds with a message sequence (lines of text, each one
                 followed by a CR).  (Typical messages are "Processing - Please Wait",
                 "4 pages sent" or "Bad Phone Line Call Again".
                */
		/* fprintf(stderr, "#UNKNOWNPROTO not the response we're looking for.  Disconnecting.\n"); */
		sprintf(error_message,"Paging service using unknown protocal.");
		hangup_modem(modem);
		return 0;
	}
	if (failcnt==-1) {
	   sprintf(error_message,"Unable to login to paging service.");
	   hangup_modem(modem);
	   return 0;
        }

/* fprintf(stderr, "waiting for message go ahead\n");  */
	failcnt=40;
	while (failcnt--) {
                /* Central may then send another message sequence (optional).

                   Central sends ESC [p CR
                   Message go ahead is sent when paging central is ready for new
                   information.  The "p" is lowercase.
                */
		get_packet(modem, pack);
		if (!strcmp(pack, "\033[p")) 
		    break;
	}
	if (failcnt==-1) {

	    sprintf(error_message,"Paging service not ready to accept transmission."); 
	    hangup_modem(modem);
	    return 0;
	 } 
    return 1;
}

/******************************************************************************/
int tapixo_protocal(modem, pin, message)
int    modem;
char   pin[MAX_PACKET];
char   message[MAXLINE+1];

{
	char pack[MAX_PACKET];	/* use same workspace */
	int mesnum, failcnt, len;
	char c;


	/* fprintf(stderr, "got message go-ahead\n");  */

		failcnt=100;
		while (failcnt--) {
                /* Remote sends STX field1 CR field2 CR field3 CR ... fieldn CR
                   EEE <CHECKSUM> CR
                   You can send as many fields as you wish.  PG terminals send the
                   pager-id in field1 and the message in field2.  (no other fields are
                   defined).  This entire message can be a max of 250 bytes.  A field may
                   be any length and where necessary may be continued in following
                   blocks.  A field always ends with a CR.  The CR field delimiter
                   suggests CR may not be used within a field.  A block always begins
                   with a STX and ends with a checksum followed by a CR.  The characters
                   preceding the checksum depends on what, if anything, is continued
                   beyond the block boundary.
                 */
			/* build the packet to be sent */
			pack[0] = STX;
			pack[1] = 0;
			strcat(pack, pin);
			strcat(pack, "\r");
			strcat(pack, message);
			strcat(pack, "\r\3");	/* CR then ETX */
			strcat(pack, checksum(pack));
			strcat(pack, "\r");
			send_data(modem, pack);
                        
			/* wait for response and deal */
			/* fprintf(stderr, "waiting for validation\n"); */
                        if (!get_packet(modem, pack) )
			     return 0;

			len = strlen(pack);
			if ((len==1)&&(*pack==ACK)) {
				/* fprintf(stderr, "#MESOK %d message xmitted fine\n", mesnum++);  */
				return 1;
			} else if ((len==1)&&(*pack==NAK)) {
				/* fprintf(stderr, "re-retrying message %d\n", mesnum); */
			} else if ((len==1)&&(*pack==RS)) {
				/* fprintf(stderr, "#MESREJECT %d message rejected\n", mesnum++); */
				return 0;
			} else if ((len==2)&&(*pack==ESC)&&(*pack==EOT)) {
				/* fprintf(stderr, "#FORDIS forced disconnect\n"); */
				return 0;
			}
		if (failcnt==-1) {
			/* fprintf(stderr, "#PROTERR couldn't send packets\n"); */
			return 0;
		}
	}
}

/************************************************************************************/
int modem_logout(modem)
int modem;
{

	char pack[MAX_PACKET];	
	int failcnt, len;

	/* fprintf(stderr, "#DONE we're done.  Logging out.\n"); */
	send_data(modem, "\4\r");	/* EOT then CR */

	failcnt=3;
	while (failcnt--) {
		/* fprintf(stderr, "waiting for system logout\n"); */
		get_packet(modem, pack);
		len=strlen(pack);
		if ((len==2) && (pack[0] == ESC) && (pack[1] == EOT)) {
			/* fprintf(stderr, "#BYE asked to hangup\n"); */
			hangup_modem(modem);
			return 1;
		}
		if ((len==1)&&(*pack==RS)) {
			/* fprintf(stderr, "#WRONGANY something went wrong. (useless msg)\n"); */
			hangup_modem(modem);
                        return 0;
		}
/*		if (!stricmp(line, "NO CARRIER")) break; */
	}
	/* fprintf(stderr, "#BYE we're leaving.\n"); */
	hangup_modem(modem);
	return 0;
}

