/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/


/****************************************************************/
/*								*/
/*		COMMON LEAD/FOLLOW MODULES			*/
/*								*/
/*	These modules are common to both ldfw1 and ldfw2	*/
/*	point types.						*/
/*								*/
/*	Author: Gordon Uchenick					*/
/*								*/
/****************************************************************/

#include <vxWorks.h>
#include <types.h>
#include <lstLib.h>
#include <tickLib.h>

#include "alarm.h"
#include "dbDefs.h"
#include "dbCommon.h"
#include "link.h"
#include "db_field_log.h"
#include "dbAddr.h"
#include "rec/pumpRecord.h"
#include "tankfarm.h"
#include "ldfw.h"

/******
	Configuration and Sizing Constants
******/

#define	MAX_PUMPS	20

/******
	Pump Status and Availability Definitions

	The pumpStatus word has a status in the lower order
	byte and two availability codes in the upper byte.
	The low order byte is the same as the defines for the
	pump state that are in <tankfarm.h>. These are the 
	availability definitions.
******/

#define CONFIGURED	0x100
#define	AVAILABLE_START	0x200
#define AVAILABLE_STOP	0x400
#define AVAILABLE	(AVAILABLE_START | AVAILABLE_STOP)
#define UNAVAILABLE	0

/******
	Pump selection mode definition (in psmd field).
******/

#define	RUNTIME		0
#define DESIGNATED	1

/****************************************************************/
/*								*/
/*	    SEE IF ALL CONFIGURED PUMPS ARE INITIALIZED		*/
/*								*/
/*	Return a 1 if they are, 0 if they aren't.		*/
/*								*/
/****************************************************************/

int ldfwCheckInit(pumpList, pumpStatus)
struct link		*pumpList;
short			*pumpStatus;
{
    int			status = 1;
    short		i;
    struct dbAddr	*pdbAddr;
    struct pumpRecord	*pumpRec;

    for (i = 0; i < MAX_PUMPS; i++)
	{
	if (pumpList[i].type == CONSTANT)
	    continue;				/* Not configured */
	if (pumpList[i].type != DB_LINK)
	    {
	    status = 0;
	    continue;				/* Configuration error */
	    }
	pumpStatus[i] = CONFIGURED;
	pdbAddr = (struct dbAddr *) (pumpList[i].value.db_link.pdbAddr);
	pumpRec = (struct pumpRecord *) (pdbAddr -> precord);
	if ((pumpRec -> init == 0) || (pumpRec -> ini2 == 0))
	    status = 0;
	}
    return (status);
}


/****************************************************************/
/*								*/
/*		DETERMINE THE NUMBER OF RUNNING PUMPS		*/
/*								*/
/****************************************************************/

void ldfwPumpsRunning(pumpList, lowerLimit, upperLimit, pumpStatus, pumpTimes,
		  runningPumps, stoppingPumps, reservePumps, lastChange, lastCommand)
struct link	*pumpList;
short		*lowerLimit;
short		*upperLimit;
short		*pumpStatus;
unsigned long	*pumpTimes;
short		*runningPumps;
short		*stoppingPumps;
short		reservePumps;
unsigned long	*lastChange;
unsigned long	*lastCommand;
{
    short		pumpsConfigured = 0;
    short		i;
    short		j;
    struct dbAddr	*pdbAddr;
    struct pumpRecord	*pumpRec;
    short		oldState;
    short		newAvail;

    /*
    Clear running and stopping counts.
    */

    *runningPumps = 0;
    *stoppingPumps = 0;

    /*
    Clear upper and lower limits.
    */

    *lowerLimit = 0;
    *upperLimit = 0;

    /*
    Traverse the pump list to see which ones are running and
    which ones aren't.
    */

    for (i = 0; i < MAX_PUMPS; i++)
	{
	if (pumpStatus[i] & CONFIGURED)
	    {
	    pumpsConfigured++;

	    /*
	    We'll just look directly into the pump record to get
	    its status, etc.
	    */

	    pdbAddr = (struct dbAddr *)(pumpList[i].value.db_link.pdbAddr);
	    pumpRec = (struct pumpRecord *)(pdbAddr -> precord);

	    /*
	    See if the pump is available to start or to stop.
	    */

	    newAvail = ldfwPumpAvailable(pumpRec);

	    /*
	    See if the pump is RUNNING or STARTING. A running pump has (val equal
	    to running) or (val equal to starting) or (val equal to fault and stat
	    equal to failed to stop). If any of these, bump the count.
	    */

	    if ((pumpRec -> val == PUMP_RUNNING) || (pumpRec -> val == PUMP_STARTING)
	    || ((pumpRec -> val == PUMP_FAULT) && (pumpRec -> stat == PUMP_STOP)))
		{
		(*runningPumps)++;

		/*
		If we are LOCAL, MANUAL, or LOCKED, then increment the lower limit
		*/

		if ((pumpRec -> swsv == RL_LOCAL) || (pumpRec -> mode == TF_MANUAL) ||
			(pumpRec -> mode == TF_LOCKED))
		    (*lowerLimit)++;

		/*
		Any pump that is running contributes to the upper limit.
		*/

		(*upperLimit)++;
		}

	    /*
	    See if the pump is STOPPING. If so, then bump that count.
	    */

	    else if (pumpRec -> val == PUMP_STOPPING)
		{
		(*stoppingPumps)++;

		/*
		Any pump that is stopping also contributes to the upper limit.
		*/

		(*upperLimit)++;
		}

	    else
		{

		/*
		Not running, starting, or stopping. If the pump can be started, then
		it contributes to the upper limit.
		*/

		if (newAvail & AVAILABLE_START)
		    (*upperLimit)++;
		}

	    /*
	    Snatch the pump status as of the last scan.  We'll use that 
	    to determine if we should update timers.
	    */

	    oldState = pumpStatus[i] & 0xff;


	    /*
	    See if the last command given to the pump just completed.
	    */

	    if (((oldState == PUMP_STOPPING) && (pumpRec -> val == PUMP_MIN_OFF))
	     || ((oldState == PUMP_STOPPING) && (pumpRec -> val == PUMP_STOPPED))
	     || ((oldState == PUMP_STARTING) && (pumpRec -> val == PUMP_RUNNING)))
	        {

	        /*
	        Yes, update the last pump change time if this command was later
	        than the previous pump change time.
	        */

	        if (pumpTimes[i] > *lastChange)
		    *lastChange = pumpTimes[i];
	        }

	    /*
	    See if the last command given to the pump just failed.
	    */

	    if (((oldState == PUMP_STARTING) && (pumpRec -> val == PUMP_MIN_OFF))
	     || ((oldState == PUMP_STARTING) && (pumpRec -> val == PUMP_STOPPED))
	     || ((oldState == PUMP_STARTING) && (pumpRec -> val == PUMP_FAULT))
	     || ((oldState == PUMP_STOPPING) && (pumpRec -> val == PUMP_FAULT))
	     || ((oldState == PUMP_STOPPING) && (pumpRec -> val == PUMP_RUNNING)))
		{
		unsigned long	latestCmd = 0;

		/*
		This command just failed. Set the last command time to the latest
		command in our list, but _not_ to the one that just failed.
		*/

		for (j = 0; j < MAX_PUMPS; j++)
		    {
		    if (j == i)
			continue;
		    if (latestCmd < pumpTimes[j])
			latestCmd = pumpTimes[j];
		    *lastCommand = latestCmd;
		    }
		}

	    /*
	    Check availability.
	    */

	    if (!newAvail)
		{

		/*
		Pump is not under our control. If a command was just issued (either by
		operator in manual or by someone at pump station in local, record the time.
		We record the time that we commanded all pumps that are under our control.
		*/

		if (((oldState == PUMP_STOPPED) && (pumpRec -> val == PUMP_STARTING))
		 || ((oldState == PUMP_MIN_OFF) && (pumpRec -> val == PUMP_STARTING))
		 || ((oldState == PUMP_RUNNING) && (pumpRec -> val == PUMP_STOPPING)))
		    {

		    /*
		    A new command was just issued. Remember it's time in case it succeeds.
		    */

		    *lastCommand = pumpTimes[i] = tickGet();
		    }
		}

	    /*
	    Set availability and status into the pump status array.
	    */

	    pumpStatus[i] = CONFIGURED | newAvail | ((short)(pumpRec -> val));
	    }
	}
    /*
    Adjust the upper limit, if necessary.
    */

    if ((pumpsConfigured - reservePumps) < *upperLimit)
        *upperLimit = pumpsConfigured - reservePumps;
}


/****************************************************************/
/*								*/
/*		DETERMINE IF A PUMP IS AVAILABLE		*/
/*								*/
/*	A pump is available if we think that we could control	*/
/*	it if we might ever want to. This does NOT NOT NOT	*/
/*	include checking for minimum off time.			*/
/*								*/
/*	Return bit masks (or together if necessary):		*/
/*		AVAILABLE_START if the pump can be started	*/
/*		AVAILABLE_STOP if the pump can be stopped	*/
/*								*/
/****************************************************************/

int ldfwPumpAvailable(pumpRec)
struct pumpRecord	*pumpRec;
{
    /*
    An available pump can't be tagged, off scan, or not fully
    initialized.
    */

    if (pumpRec -> sevr == VALID_ALARM)
	return (UNAVAILABLE);

    /*
    An available pump must be in AUTOMATIC.
    */

    if (pumpRec -> mode == TF_MANUAL)
	return (UNAVAILABLE);
    if (pumpRec -> mode == TF_LOCKED)
	return (UNAVAILABLE);

    /*
    An available pump must be in REMOTE.
    */

    if (pumpRec -> swsv == RL_LOCAL)
	return (UNAVAILABLE);

    /*
    An available pump must not be in PERMISSIVE alarm.
    */

    if (pumpRec -> stat == PERMISSIVE_ALM)
	return (UNAVAILABLE);

    /*
    An available pump must not have failed to start or stop.
    */

    if (pumpRec -> stat == PUMP_START)
	return (UNAVAILABLE);
    if (pumpRec -> stat == PUMP_STOP)
	return (UNAVAILABLE);

    /*
    An available pump must not be in EMERGENCY STOP.
    */

    if (pumpRec -> estp == 1)
	return (UNAVAILABLE);

    /*
    It also can't have an uncommanded change of state.
    */

    if (pumpRec -> stat == TF_UNCOMMANDED)
	return (UNAVAILABLE);

    /*
    If we get this far, then we can command the pump. If it is in
    pressure, temperature, or vibration alarm, it's OK to stop the
    pump, but not to start it.
    */

    if (pumpRec -> stat == PRESSURE)
	return (AVAILABLE_STOP);
    if (pumpRec -> stat == TEMPERATURE)
	return (AVAILABLE_STOP);
    if (pumpRec -> stat == VIBRATION)
	return (AVAILABLE_STOP);

    /*
    If none of the above, then the pump must be available to be either
    started or stopped.
    */

    return (AVAILABLE);
}


/****************************************************************/
/*								*/
/*			START A PUMP				*/
/*								*/
/*	Return a 1 if we started a pump, a zero if we 		*/
/*	couldn't.						*/
/*								*/
/****************************************************************/

int ldfwStartPump(lastChange, secsBetweenStarts, pumpList, pumpStatus, pumpTimes,
		offTimeSetpoint, selMode, lastDesigStart, desigSeq, lastCommand)
unsigned long		*lastChange;		/* Last pump change time */
float			secsBetweenStarts;	/* Seconds between starts */
struct link		*pumpList;		/* List of pumps */
short			*pumpStatus;		/* Pump status array */
unsigned long		*pumpTimes;		/* Pump command times */
float			offTimeSetpoint;	/* Off time setpoint */
short			selMode;		/* Pump selection mode */
short			*lastDesigStart;	/* Last designated pump started */
short			*desigSeq;		/* Designated pump start sequence */
unsigned long		*lastCommand;		/* Time of last command issued */
{
    short			numAvail;
    short			i;
    short			minRunTime;
    short			maxOffTime;
    short			origDesigSeq;
    short			nextSeq;
    unsigned long		offTime = 0;
    struct dbAddr		*pdbAddr;
    struct pumpRecord		*pumpRec;
    struct pumpRecord		*minPump = NULL;
    struct pumpRecord		*maxPump = NULL;
    unsigned long		lastTime;	/* Last change or command */

    /*
    Get later of last change time or last command time.
    */

    lastTime = *lastChange;
    if (lastTime < *lastCommand)
	lastTime = *lastCommand;

    /*
    See if it's too soon to start a pump.
    */

    if ((tickGet() - lastTime) < (secsBetweenStarts * 60.0))
	return (0);

    /*
    Check pump selection mode.
    */

    if (selMode == DESIGNATED)
	{

	/*
	Operator has designated a sequence. We will follow it as best we can.
	That means we will start pumps in the order given, ignoring those that
	are not available. If a pump is available but can't be started because
	it's in minimum off time, do nothing. If the operator wanted us to find
	a different pump, we wouldn't be in this mode.
	*/

	/*
	Try to find the next one in the sequence, wrapping around to 1 if necessary.
	If we get back to where we started, then we can't start a pump in this
	mode.
	*/

	origDesigSeq = *lastDesigStart;
	nextSeq = origDesigSeq + 1;
	while (nextSeq != origDesigSeq)
	    {
	    for (i = 0; i < MAX_PUMPS; i++)
	    	if (desigSeq[i] == nextSeq)
		    break;
	    if (i >= MAX_PUMPS)
		{
		if (nextSeq == 1)
		    return (0);		/* Missing pump in designated sequence! */
		else
		    {
		    nextSeq = 1;
		    continue;
		    }
		}
	    
	    /*
	    Found the next pump in the sequence. If it isn't available, skip it and
	    keep on going to the next one. This means that if a designated sequence
	    has been entered and one of the pumps fails that we just won't consider
	    that one.
	    */

	    if (!(pumpStatus[i] & CONFIGURED))
		{
		nextSeq++;
		continue;
		}
	    if (!(pumpStatus[i] & AVAILABLE_START))
		{
		nextSeq++;
		continue;
		}
	    if ((pumpStatus[i] & 0xff) == PUMP_MIN_OFF)
		return (0);		/* The pump we want is available, but we have to
					   wait. A later execution of this block will
					   actually get it started. */

	    if ((pumpStatus[i] & 0xff) == PUMP_STOPPED)
		{

		/*
		Gotcha! Start this pump!
		*/
/*
printf("Starting pump %d\n", i + 1);
date();
*/
		pdbAddr = (struct dbAddr *) (pumpList[i].value.db_link.pdbAddr);
		pumpRec = (struct pumpRecord *) (pdbAddr -> precord);
		pumpRec -> strs = 1;
		pumpStatus[i] &= 0xff00;
		pumpStatus[i] |= PUMP_RUNNING;
		*lastCommand = pumpTimes[i] = tickGet();
		*lastDesigStart = desigSeq[i];
		return (1);
		}

	    /*
	    If we get here, the pump we want is probably already running. Keep on
	    looking, in sequence order, to find another one.
	    */

	    nextSeq++;
	    continue;
	    }

	/*
	If we got here, then we can't start a pump in designated sequence mode.
	*/

	return (0);
	}


    /*
    We're in the runtime mode. We want to start the available pump
    that has the least run time, subject to the Off Time Setpoint.
    Traverse the Pump Status List to find the number of available
    pumps that have been stopped longer than the Off Time Setpoint.
    Note that a pump that is stopped but within its minimum off
    time has a unique status so we don't have to make two checks.
    While we are making this loop, keep track of which one has
    the minimum run time.
    */

    minRunTime = -1;
    numAvail = 0;
    for (i = 0; i < MAX_PUMPS; i++)
	{
	if (!(pumpStatus[i] & CONFIGURED))
	    continue;

	/*
	Only condsider pumps that are stopped. Pumps that are
	inside minimum off time can't be started.
	*/

	if ((pumpStatus[i] & 0xff) != PUMP_STOPPED)
	    continue;

	/*
	Only consider pumps that can be started and have been off
	longer than off setpoint time.
	*/

	if (!(pumpStatus[i] & AVAILABLE_START))
	    continue;
	if ((tickGet() - pumpTimes[i]) < (offTimeSetpoint * 60.0))
	    continue;

	/*
	This pump has been off longer than off time setpoint. Remember its 
	index and record addr if it is now the one with the lowest accumulated 
	run time.
	*/

	pdbAddr = (struct dbAddr *) (pumpList[i].value.db_link.pdbAddr);
	pumpRec = (struct pumpRecord *)(pdbAddr -> precord);
	if (minRunTime < 0)
	    {
	    minRunTime = i;
	    minPump = pumpRec;
	    }
	else if (pumpRec -> acrt < minPump -> acrt)
	    {
	    minRunTime = i;
	    minPump = pumpRec;
	    }
	numAvail++;
	}

    if (numAvail)
	{

	/*
	At least one pump was off longer than the off time setpoint. 
	Start the one with the least run time.
	*/
/*
printf("Starting pump %d\n", minRunTime + 1);
date();
*/
        minPump -> strs = 1;	/* Start the pump */
        pumpStatus[minRunTime] &= 0xff00;
        pumpStatus[minRunTime] |= PUMP_STARTING;
        *lastCommand = pumpTimes[minRunTime] = tickGet();
        return (1);
        }

    else
	{

	/*
	No pumps were off longer than the Off Time Setpoint. See if there
	is still one we can start. If there is more than one pump that can 
	be started, start the one that has been off the longest.
	*/

	maxOffTime = -1;

    	for (i = 0; i < MAX_PUMPS; i++)
	    {
	    if (!(pumpStatus[i] & CONFIGURED))
	        continue;
	    if ((pumpStatus[i] & 0xff) != PUMP_STOPPED)
	        continue;
	    if (!(pumpStatus[i] & AVAILABLE_START))
		continue;

	    /*
	    This pump is available. Remember its index and record addr
	    if it is the one that's been off the longest.
	    */

	    pdbAddr = (struct dbAddr *) (pumpList[i].value.db_link.pdbAddr);
	    pumpRec = (struct pumpRecord *)(pdbAddr -> precord);
	    if (maxOffTime < 0)
	        {
	        maxOffTime = i;
	        maxPump = pumpRec;
	        offTime = tickGet() - pumpTimes[i];
	        }
	    else if ((tickGet() - pumpTimes[i]) > offTime)
	        {
	        maxOffTime = i;
	        maxPump = pumpRec;
	        offTime = tickGet() - pumpTimes[i];
	        }
	    }


	/*
	If we found a pump that can be started, go ahead and do it.
	*/

	if (maxOffTime >= 0)
	    {
/*
printf("Starting pump %d\n", maxPump + 1);
date();
*/
	    maxPump -> strs = 1;	/* Start the pump */
	    pumpStatus[maxOffTime] &= 0xff00;
	    pumpStatus[maxOffTime] |= PUMP_STARTING;
	    *lastCommand = pumpTimes[maxOffTime] = tickGet();
	    return (1);
	    }
	}

    /*
    If we get here, we can't start a pump.
    */

    return (0);
}


/****************************************************************/
/*								*/
/*			STOP A PUMP				*/
/*								*/
/*	Return a 1 if we stopped a pump, a zero if we		*/
/*	couldn't.						*/
/*								*/
/****************************************************************/

int ldfwStopPump(lastChange, secsBetweenStops, pumpList, pumpStatus, pumpTimes,
		selMode, manSeq, lastCommand)
unsigned long		*lastChange;
float			secsBetweenStops;
struct link		*pumpList;
short			*pumpStatus;
unsigned long		*pumpTimes;
short			selMode;
short			*manSeq;
unsigned long		*lastCommand;
{
    short			i;
    short			maxRunTime;
    struct dbAddr		*pdbAddr;
    struct pumpRecord		*pumpRec;
    double			runTime = 0;
    unsigned long		timeSinceStart = 0;
    unsigned long		lastTime;


    /*
    Get later of last change time or last command time.
    */

    lastTime = *lastChange;
    if (lastTime < *lastCommand)
	lastTime = *lastCommand;

    /*
    See if it's too soon to stop a pump.
    */

    if ((tickGet() - lastTime) < (secsBetweenStops * 60.0))
	return (0);

    /*
    Check mode.
    */

    maxRunTime = -1;
    if (selMode == RUNTIME)
	{

	/*
	Runtime mode. Find the pump that is available and has the
	greatest run time.
	*/

	for (i = 0; i < MAX_PUMPS; i++)
	    {
	    pdbAddr = (struct dbAddr *) (pumpList[i].value.db_link.pdbAddr);
	    pumpRec = (struct pumpRecord *) (pdbAddr -> precord);
	    if (!(pumpStatus[i] & CONFIGURED))
		continue;
	    if ((pumpStatus[i] & 0xff) != PUMP_RUNNING)
		continue;
	    if (!(pumpStatus[i] & AVAILABLE_STOP))
		continue;
	    
	    /*
	    This pump is available. Remember it's index and record
	    address if it is now the one with the greatest accumulated
	    run time.
	    */

	    if (maxRunTime < 0)
		{
		maxRunTime = i;
		runTime = pumpRec -> acrt;
		}
	    else if (pumpRec -> acrt > runTime)
		{
		maxRunTime = i;
		runTime = pumpRec -> acrt;
		}
	    }
	}

    else
	{

	/*
	Designated Mode. Turn off the available pump that has been running the longest.
	Only look at the pumps that are in the manual sequence.
	*/

	for (i = 0; i < MAX_PUMPS; i++)
	    {
	    if (!manSeq[i])
		continue;
	    if (!(pumpStatus[i] & CONFIGURED))
		continue;
	    if ((pumpStatus[i] & 0xff) != PUMP_RUNNING)
		continue;
	    if (!(pumpStatus[i] & AVAILABLE_STOP))
		continue;
	    if (maxRunTime < 0)
		{
		maxRunTime = i;
		timeSinceStart = tickGet() - pumpTimes[i];
		}
	    else if ((tickGet() - pumpTimes[i]) > timeSinceStart)
		{
		maxRunTime = i;
		timeSinceStart = tickGet() - pumpTimes[i];
		}
	    }
	}
    
    /*
    If we found a pump to stop, do it.
    */

    if (maxRunTime >= 0)
    	{
/*
printf("Stopping pump %d\n", maxRunTime + 1);
date();
*/
	pdbAddr = (struct dbAddr *) (pumpList[maxRunTime].value.db_link.pdbAddr);
	pumpRec = (struct pumpRecord *) (pdbAddr -> precord);
    	pumpRec -> stps = 1;	/* Stop the pump */
    	pumpStatus[maxRunTime] &= 0xff00;
    	pumpStatus[maxRunTime] |= PUMP_STOPPING;
    	*lastCommand = pumpTimes[maxRunTime] = tickGet();
    	return (1);
    	}

    /*
    If we get here, we couldn't stop a pump.
    */

    return (0);
}
