/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/


/****************************************************************/
/*								*/
/*			LDFW2 RECORD PROCESSING			*/
/*								*/
/*	Record support for the LDFW2 block type. This block	*/
/*	controls the start, stop, and sequence from one to	*/
/*	twenty pumps. Timing parameters control behavior	*/
/*	in runtime selection mode. The operator can also	*/
/*	enter a designated sequence. This block is basically	*/
/*	the same as LDFW1, but also adds speed adjustment for	*/
/*	variable speed pumps.					*/
/*								*/
/*	Author: Gordon Uchenick					*/
/*								*/
/****************************************************************/

#include <vxWorks.h>
#include <inetLib.h>
#include <types.h>
#include <lstLib.h>
#include <string.h>

#include "alarm.h"
#include "dbDefs.h"
#include "dbCommon.h"
#include "link.h"
#include "db_field_log.h"
#include "dbAccess.h"
#include "dbRecType.h"
#include "recSup.h"
#include "special.h"
#include "rec/ldfw1Record.h"
#include "rec/ldfw2Record.h"
#include "rec/pidRecord.h"
#include "rec/pumpRecord.h"
#include "redflds.h"
#include "redque.h"
#include "restable.h"
#include "tankfarm.h"
#include "db_addr.h"
#include "dbEvent.h"
#include "recGbl.h"
#include "ldfw.h"
#include "ldfw1redfldidx.h"
#include "ldfw2redfldidx.h"


/******
	Definitions for the RSET
******/

#define report NULL
#define initialize NULL
static long init_record();
static long process();
static long special();
#define get_value NULL
#define cvt_dbaddr NULL
#define get_array_info NULL
#define put_array_info NULL
#define get_units NULL
static long get_precision();
#define get_enum_str NULL
#define get_enum_strs NULL
#define put_enum_str NULL
static long get_graphic_double();
static long get_control_double();
#define get_alarm_double NULL

struct rset ldfw2RSET = {

    RSETNUMBER,
    report,
    initialize,
    init_record,
    process,
    special,
    get_value,
    cvt_dbaddr,
    get_array_info,
    put_array_info,
    get_units,
    get_precision,
    get_enum_str,
    get_enum_strs,
    put_enum_str,
    get_graphic_double,
    get_control_double,
    get_alarm_double
    };
    
/****************************************************************/
/*								*/
/*			EXTERNAL DATA				*/
/*								*/
/****************************************************************/

/* defined as global in ../db/iocInit.c */
extern long *process_cnts;
/* defined as global in ../db/procstat.c */
extern long *process_stats;
extern unsigned short DBC_CRCRecord();


/****************************************************************/
/*								*/
/*			PRIVATE DATA				*/
/*								*/
/****************************************************************/

static unsigned short type;
static unsigned short ldfw1_type;
static struct fldDes **mylist = 0;
static struct fldDes **ldfw1_mylist = 0;

static char	*namelist [] = {
		"STAT",
		"SEVR",
		"VAL",
		"LALM",
		"INI2",
		"LATM",
		"MLST",
		"PSPD",
		"LPST",
		"NPRC",
		0
		};
static long alarm ();
static void monitor();

#define CONFIGURED	0x100
#define MAX_PUMPS	20


/****************************************************************/
/*								*/
/*			INITIALIZE A RECORD			*/
/*								*/
/****************************************************************/

static long init_record (pldfw2)
struct ldfw2Record	*pldfw2;
{
    struct dbAddr	*pdbAddr;
    struct dbCommon	*recHdr;
    struct ldfw1Record	*pldfw1;
    short		master;
    short		i;
    char		*pstr;
    char		redbuf[512];

    if (!(mylist))

	{
	/* find record type for ldfw2 */

	for (type = 1, i = 0; type < 100; type++)
	    {
	    pstr = GET_PRECTYPE(type);
	    if (strncmp(pstr, "ldfw2", 5) == 0)
		{
		i = 1;
		break;
		}
	    }
	
	if (i == 0)
	    {
	    recGblRecordError(S_db_noRecType, (struct dbCommon *)pldfw2, "ldfw2:init_record");
	    return (S_db_noRecType);
	    }

	/* find the type for the ldfw1 block */

	for (ldfw1_type = 1, i = 0; ldfw1_type < 100; ldfw1_type++)
	    {
	    pstr = GET_PRECTYPE(ldfw1_type);
	    if (strncmp(pstr, "ldfw1", 5) == 0)
		{
		i = 1;
		break;
		}
	    }
	
	if (i == 0)
	    {
	    recGblRecordError(S_db_noRecType, (struct dbCommon *)pldfw2, "ldfw2:init_record,ldfw1");
	    return (S_db_noRecType);
	    }
	
    	/*
    	Initialize redundancy stuff here.
    	*/

	mylist = redFldAlloc (type, NUMLDFW2FLDS);

	if (!mylist)
	    {
		recGblRecordError (S_rec_Alloc, (struct dbCommon *)pldfw2,
				   "ldfw2:init_record, mylist failed");
		return (S_rec_Alloc);
	    }

	for (i = 0; i < NUMLDFW2FLDS; i++)
	    redFldAdd (type, i, namelist[i]);
	}

    /* if no ldfw1 block has init'd to build its redundancy list
       then this ldfw2 block can not init
    */
    if (!(ldfw1_mylist))
	{
	    if (papRedFldList[ldfw1_type] == 0)
	    {
		recGblRecordError (S_db_notFound, (struct dbCommon *)pldfw2, 
				   "ldfw2:init_record,ldwf1_mylist");
		return (S_db_notFound);
	    }
	    else
	    {
		ldfw1_mylist = papRedFldList[ldfw1_type];
	    }
	}

    /*
    Initialization stuff typical to all point types.
    */

    pldfw2 -> init = TRUE;
    pldfw2 -> nsta = NO_ALARM;
    pldfw2 -> nsev = NO_ALARM;

    if (pldfw2 -> latm.secPastEpoch == 0)
	tsLocalTime(&pldfw2 -> latm);
    if (pldfw2 -> lptm.secPastEpoch == 0)
	tsLocalTime(&pldfw2 -> lptm);


    /*
    Locate the LDFW1 record that contains our parameters.
    */

    if (pldfw2 -> ldfw.type != DB_LINK)
    {
	recGblRecordError (S_db_badField, (struct dbCommon *)pldfw2, 
				"ldfw2:init_record,no ldfw1");
	return (S_db_badField);
    }
    pdbAddr = (struct dbAddr *) (pldfw2 -> ldfw.value.db_link.pdbAddr);
    pldfw1 = (struct ldfw1Record *) (pdbAddr -> precord);

    /*
    See if all the pumps we need to work with are fully initialized.
    */

    resGetCaAnswer(pldfw2 -> radr, &master);
    if (!ldfwCheckInit(&(pldfw1 -> pm01), &(pldfw1 -> ps01)))
	{
	if (master)
	    {
	    pldfw2 -> nsta = NO_INIT_ALARM;
	    pldfw2 -> nsev = VALID_ALARM;
	    monitor(pldfw2, 0, 0, 0, 0, pldfw2 -> lpst, pldfw2 -> pspd, pldfw2 -> nprc);
	    return (0);
	    }
	}

    /*
    See if the suspend mode input has fully initialized.
    */


    if (pldfw1 -> susp.type != CONSTANT)
	{
	if (pldfw1 -> susp.type != DB_LINK)
	    {
	    if (master)
		{
		pldfw2 -> nsta = NO_INIT_ALARM;
		pldfw2 -> nsev = VALID_ALARM;
		monitor(pldfw2, 0, 0, 0, 0, pldfw2 -> lpst, pldfw2 -> pspd, pldfw2 -> nprc);
	    	return (0);
		}
	    }
	pdbAddr = (struct dbAddr *) (pldfw1 -> susp.value.db_link.pdbAddr);
	recHdr = (struct dbCommon *) (pdbAddr -> precord);
	if ((recHdr -> init == 0) || (recHdr -> ini2 == 0))
	    {
	    if (master)
		{
		pldfw2 -> nsta = NO_INIT_ALARM;
		pldfw2 -> nsev = VALID_ALARM;
		monitor(pldfw2, 0, 0, 0, 0, pldfw2 -> lpst, pldfw2 -> pspd, pldfw2 -> nprc);
	    	return (0);
		}
	    }
	}

    /*
    If we are on the master machine, set our mode to MANUAL.
    */

    if (master)
	{
    	pldfw2 -> mode = 1;		/* Manual mode */
	db_post_events((struct dbCommon *)pldfw2, (caddr_t)&(pldfw2 -> mode), DBE_VALUE);
    	pldfw2 -> ini2 = TRUE;
	pldfw2 -> rcrc = DBC_CRCRecord(pldfw2, type);

	/*
	Force redundant update to get master disk copy in sync now!
	*/

	redQInit ((struct dbCommon *)pldfw2, redbuf, sizeof (redbuf), type);

	redQField (redbuf, mylist [LDFW2_INI2]);

	redQSend (redbuf);
	}

    /*
    We've initialized -- await first processing pass to get rest going.
    */

    return (1);
}


/****************************************************************/
/*								*/
/*			PROCESS A RECORD			*/
/*								*/
/****************************************************************/

static long process(paddr)
struct dbAddr		*paddr;
{
    struct ldfw2Record		*pldfw2 = (struct ldfw2Record *) (paddr -> precord);
    struct ldfw1Record		*pldfw1;
    struct dbAddr		*pdbAddr;
    float			rawSpeed;
    float			adjust;
    float			*afList;
    long			options = 0;
    long			nRequest = 1;
    short			*pStat;
    short			lowerLimit;
    short			upperLimit;
    short			runningPumps;
    short			stoppingPumps;
    short			pumpsRequested;
    short			lastDesigStart;
    short			status;
    short			runConfirm;
    short			i;
    short			susm;
    unsigned short		stat = pldfw2 -> stat;
    unsigned short		sevr = pldfw2 -> sevr;
    short			val  = pldfw2 -> val;
    short			lalm = pldfw2 -> lalm;
    TS_STAMP			latm = pldfw2 -> latm;
    float			mlst = pldfw2 -> mlst;
    float			pspd = pldfw2 -> pspd;
    short			lpst = pldfw2 -> lpst;
    short			nprc = pldfw2 -> nprc;
    unsigned long		lpct;
    unsigned long		lpcm;
    unsigned long		pc01;
    unsigned long		pc02;
    unsigned long		pc03;
    unsigned long		pc04;
    unsigned long		pc05;
    unsigned long		pc06;
    unsigned long		pc07;
    unsigned long		pc08;
    unsigned long		pc09;
    unsigned long		pc10;
    unsigned long		pc11;
    unsigned long		pc12;
    unsigned long		pc13;
    unsigned long		pc14;
    unsigned long		pc15;
    unsigned long		pc16;
    unsigned long		pc17;
    unsigned long		pc18;
    unsigned long		pc19;
    unsigned long		pc20;
    char			redbuf[512];


    /*
    Do nothing if we didn't init.
    */

    if ((pldfw2 -> init == 0) || (pldfw2 -> ini2 == 0))
	return (0);

    /*
    Locate the LDFW1 record that contains our parameters.
    */

    pdbAddr = (struct dbAddr *) (pldfw2 -> ldfw.value.db_link.pdbAddr);
    pldfw1 = (struct ldfw1Record *) (pdbAddr -> precord);

    /* remember for redundancy */
    lpct = pldfw1 -> lpct;
    lpcm = pldfw1 -> lpcm;
    pc01 = pldfw1 -> pc01;
    pc02 = pldfw1 -> pc02;
    pc03 = pldfw1 -> pc03;
    pc04 = pldfw1 -> pc04;
    pc05 = pldfw1 -> pc05;
    pc06 = pldfw1 -> pc06;
    pc07 = pldfw1 -> pc07;
    pc08 = pldfw1 -> pc08;
    pc09 = pldfw1 -> pc09;
    pc10 = pldfw1 -> pc10;
    pc11 = pldfw1 -> pc11;
    pc12 = pldfw1 -> pc12;
    pc13 = pldfw1 -> pc13;
    pc14 = pldfw1 -> pc14;
    pc15 = pldfw1 -> pc15;
    pc16 = pldfw1 -> pc16;
    pc17 = pldfw1 -> pc17;
    pc18 = pldfw1 -> pc18;
    pc19 = pldfw1 -> pc19;
    pc20 = pldfw1 -> pc20;

    /*
    Get last process time.
    */

    tsLocalTime(&(pldfw2 -> lptm));

    /*
    Init new running/stopping pump counts to previous values.
    */

    runningPumps = pldfw2 -> val;
    stoppingPumps = pldfw2 -> nstp;
    runningPumps -= stoppingPumps;
    lastDesigStart = pldfw2 -> lpst;
    upperLimit = pldfw2 -> ulim;
    lowerLimit = pldfw2 -> llim;
    rawSpeed = pldfw2 -> pspd;
    runConfirm = pldfw2 -> nprc;

    /*
    Copy current limits so we can detect change later.
    */

    lowerLimit = pldfw2 -> llim;
    upperLimit = pldfw2 -> ulim;

    /*
    Check for tagged or off scan.
    */

    if (pldfw2 -> tagd)
	{
	pldfw2 -> nsta = TAGGED_ALARM;
	pldfw2 -> nsev = VALID_ALARM;
	}
    else if (pldfw2 -> oscn == OFF_SCAN)
	{
	pldfw2 -> nsta = SCAN_ALARM;
	pldfw2 -> nsev = VALID_ALARM;
	}
    else
	{

	pldfw2 -> nsta = NO_ALARM;
	pldfw2 -> nsev = NO_ALARM;

	/*
	This record must process. Get the number of pumps that are
	currently running. This will also update all our timers for
	pumps in case we are in manual or suspended mode.
	*/

	ldfwPumpsRunning(&(pldfw1 -> pm01), &lowerLimit, &upperLimit, &(pldfw1 -> ps01),
		&(pldfw1 -> pc01), &runningPumps, &stoppingPumps, pldfw1 -> nrpm,
		&(pldfw1 -> lpct), &(pldfw1 -> lpcm));

	/*
	The count of running pumps returned by ldfwPumpsRunning() includes pumps
	that are STARTING. We don't want to include them for selecting the adjustment
	factor.
	*/

	runConfirm = runningPumps + stoppingPumps;
	pStat = &(pldfw1 -> ps01);
	for (i = 0; i < MAX_PUMPS; i++)
	    {
	    if (!(pStat[i] & CONFIGURED))
		continue;
	    if ((pStat[i] & 0xff) == PUMP_STARTING)
		runConfirm--;
	    }

	afList = &(pldfw2 -> af01);
	adjust = 1.0;
	if (runConfirm > pldfw2 -> nprc)
	    adjust = afList[runConfirm - 1];
	else if (runConfirm < pldfw2 -> nprc)
	    adjust = 1.0 / afList[pldfw2 -> nprc];
	
	/*
	Fetch the controller output. Add it to the previous
	value and then clamp it.
	*/

	status = dbGetLink(&pldfw2 -> scnt.value.db_link, 
			   (struct dbCommon *)pldfw2, 
			   DBR_FLOAT, (caddr_t) &rawSpeed, 
			   &options, &nRequest);

        rawSpeed += pldfw2 -> pspd;
        if (rawSpeed > 100.0)
	    rawSpeed = 100.0;
        else if (rawSpeed < 0.0)
	    rawSpeed = 0.0;

	/*
	Adjust it for the number of pumps that are running. Then,
	clamp it again.
	*/

	rawSpeed *= adjust;
	if (rawSpeed > 100.0)
	    rawSpeed = 100.0;
	else if (rawSpeed < 0.0)
	    rawSpeed = 0.0;

	/*
	If defined, get the suspended mode input.
	*/

	if (pldfw1 -> susp.type == DB_LINK)
	    {

	    status = dbGetLink(&pldfw1 -> susp.value.db_link, 
			       (struct dbCommon *)pldfw1, 
			       DBR_SHORT, (caddr_t) &susm, 
			       &options, &nRequest);
	    }

	/*
	Continue processing only if we are not suspended and also are in automatic.
	*/

	if ((susm != pldfw1 -> susv) && (pldfw2 -> mode == 0))
	    {

	    /*
	    Compare the requested speed to the start/stop limits.
	    */

	    pumpsRequested = pldfw2 -> val;	/* Assume no change */
	    if (rawSpeed > pldfw2 -> stsp)
		pumpsRequested++;
	    else if (pumpsRequested && (rawSpeed < pldfw2 -> spsp))
		pumpsRequested--;

	    if (pumpsRequested > upperLimit)
		{
		pumpsRequested = upperLimit;
		if (pldfw1 -> usvr > pldfw2 -> nsev)
		    {
		    pldfw2 -> nsev = pldfw1 -> usvr;
		    pldfw2 -> nsta = HIGH_ALARM;
		    }
		}
	    if (pumpsRequested < lowerLimit)
		{
		pumpsRequested = lowerLimit;
		if (pldfw1 -> lsvr > pldfw2 -> nsev)
		    {
		    pldfw2 -> nsev = pldfw1 -> lsvr;
		    pldfw2 -> nsta = LOW_ALARM;
		    }
		}

	    /*
	    Compare the number of pumps that are running to the number that are
	    supposed to be running. If they are different, then we have to do
	    something about it.
	    */

	    if (pumpsRequested > (runningPumps + stoppingPumps))
		{

		/*
		Start another pump.
		*/

		if (ldfwStartPump(&(pldfw1 -> lpct), pldfw1 -> strs, &(pldfw1 -> pm01),
			&(pldfw1 -> ps01), &(pldfw1 -> pc01), pldfw1 -> ospt,
			pldfw2 -> psmd, &lastDesigStart, &(pldfw1 -> ds01), &(pldfw1 -> lpcm)))
		    runningPumps++;
		}

	   else if (pumpsRequested < runningPumps)
		{

		/*
		Stop a pump.
		*/

		if (ldfwStopPump(&(pldfw1 -> lpct), pldfw1 -> stps, &(pldfw1 -> pm01), 
			&(pldfw1 -> ps01), &(pldfw1 -> pc01), pldfw2 -> psmd, 
			&(pldfw1 -> ds01), &(pldfw1 -> lpcm)))
		    stoppingPumps++;
		}
	    }
	alarm (pldfw2);
	}

    /*
    Call the monitor routine to let the rest of the world know what's going on.
    */

    monitor(pldfw2, lowerLimit, upperLimit, runningPumps, stoppingPumps, lastDesigStart,
	rawSpeed, runConfirm, susm);

    /*
    Handle redundancy here
    */

    redQInit ((struct dbCommon *)pldfw2, redbuf, sizeof (redbuf), type);

    if (stat != pldfw2 -> stat)
	redQField (redbuf, mylist [LDFW2_STAT]);

    if (sevr != pldfw2 -> sevr)
	redQField (redbuf, mylist [LDFW2_SEVR]);

    if (val != pldfw2 -> val)
	redQField (redbuf, mylist [LDFW2_VAL]);

    if (lalm != pldfw2 -> lalm)
	redQField (redbuf, mylist [LDFW2_LALM]);

    if (TsCmpStampsNE (&latm, &pldfw2 -> latm) )
	redQField (redbuf, mylist [LDFW2_LATM]);

    if (mlst != pldfw2 -> mlst)
	redQField (redbuf, mylist [LDFW2_MLST]);

    if (pspd != pldfw2 -> pspd)
	redQField (redbuf, mylist [LDFW2_PSPD]);

    if (lpst != pldfw2 -> lpst)
	redQField (redbuf, mylist [LDFW2_LPST]);

    if (nprc != pldfw2 -> nprc)
	redQField (redbuf, mylist [LDFW2_NPRC]);

    redQSend (redbuf);

    /* now handle redunancy for the referenced ldfw1 block */

    redQInit ((struct dbCommon *)pldfw1, redbuf, sizeof (redbuf), ldfw1_type);

    if (lpct != pldfw1 -> lpct)
	redQField (redbuf, ldfw1_mylist [LPCT]);

    if (lpcm != pldfw1 -> lpcm)
	redQField (redbuf, ldfw1_mylist [LPCM]);

    if (pc01 != pldfw1 -> pc01)
	redQField (redbuf, ldfw1_mylist [PC01]);

    if (pc02 != pldfw1 -> pc02)
	redQField (redbuf, ldfw1_mylist [PC02]);

    if (pc03 != pldfw1 -> pc03)
	redQField (redbuf, ldfw1_mylist [PC03]);

    if (pc04 != pldfw1 -> pc04)
	redQField (redbuf, ldfw1_mylist [PC04]);

    if (pc05 != pldfw1 -> pc05)
	redQField (redbuf, ldfw1_mylist [PC05]);

    if (pc06 != pldfw1 -> pc06)
	redQField (redbuf, ldfw1_mylist [PC06]);

    if (pc07 != pldfw1 -> pc07)
	redQField (redbuf, ldfw1_mylist [PC07]);

    if (pc08 != pldfw1 -> pc08)
	redQField (redbuf, ldfw1_mylist [PC08]);

    if (pc09 != pldfw1 -> pc09)
	redQField (redbuf, ldfw1_mylist [PC09]);

    if (pc10 != pldfw1 -> pc10)
	redQField (redbuf, ldfw1_mylist [PC10]);

    if (pc11 != pldfw1 -> pc11)
	redQField (redbuf, ldfw1_mylist [PC11]);

    if (pc12 != pldfw1 -> pc12)
	redQField (redbuf, ldfw1_mylist [PC12]);

    if (pc13 != pldfw1 -> pc13)
	redQField (redbuf, ldfw1_mylist [PC13]);

    if (pc14 != pldfw1 -> pc14)
	redQField (redbuf, ldfw1_mylist [PC14]);

    if (pc15 != pldfw1 -> pc15)
	redQField (redbuf, ldfw1_mylist [PC15]);

    if (pc16 != pldfw1 -> pc16)
	redQField (redbuf, ldfw1_mylist [PC16]);

    if (pc17 != pldfw1 -> pc17)
	redQField (redbuf, ldfw1_mylist [PC17]);

    if (pc18 != pldfw1 -> pc18)
	redQField (redbuf, ldfw1_mylist [PC18]);

    if (pc19 != pldfw1 -> pc19)
	redQField (redbuf, ldfw1_mylist [PC19]);

    if (pc20 != pldfw1 -> pc20)
	redQField (redbuf, ldfw1_mylist [PC20]);

    redQSend (redbuf);


    return (0);
}


/****************************************************************/
/*								*/
/*			SPECIAL PROCESSING			*/
/*								*/
/****************************************************************/

static long special()
{
/* None, yet.*/
    return (0);
}


/****************************************************************/
/*								*/
/*		CHECK FOR AND REPORT ALARM CONDITIONS		*/
/*								*/
/****************************************************************/

static long alarm (pldfw2)
    struct ldfw1Record	*pldfw2;
{
    int	adis;

    adis = AlarmDisableLink ((void *)pldfw2);
    if ( (pldfw2 -> aenb == ALARM_DISABLED) || (adis != 0) )
    {
	pldfw2 -> nsta = INHIBIT_ALARM;
	pldfw2 -> nsev = NO_ALARM;
    }
    return (0);
}


/****************************************************************/
/*								*/
/*		POST MONITOR EVENTS AS NECESSARY		*/
/*								*/
/****************************************************************/

static void monitor (pldfw2, lowerLimit, upperLimit, runningPumps, stoppingPumps, 
		lastDesigStart, rawSpeed, runConfirm, susm)
struct ldfw2Record	*pldfw2;
short			lowerLimit;
short			upperLimit;
short			runningPumps;
short			stoppingPumps;
short			lastDesigStart;
float			rawSpeed;
short			runConfirm;
{
    unsigned short	eventMask = 0;
    short		stat, sevr, nsta, nsev;
    int			set_ack = 1;
    float		delta;

    /*
    Check lower limit.
    */

    if (lowerLimit != pldfw2 -> llim)
	{
	pldfw2 -> llim = lowerLimit;
	db_post_events((struct dbCommon *)pldfw2, (caddr_t)&(pldfw2 -> llim), DBE_VALUE);
	}

    /*
    Check upper limit.
    */

    if (upperLimit != pldfw2 -> ulim)
	{
	pldfw2 -> ulim = upperLimit;
	db_post_events((struct dbCommon *)pldfw2, (caddr_t)&(pldfw2 -> ulim), DBE_VALUE);
	}

    /*
    Check last designated start index.
    */

    if (lastDesigStart != pldfw2 -> lpst)
	{
	pldfw2 -> lpst = lastDesigStart;
	db_post_events((struct dbCommon *)pldfw2, (caddr_t)&(pldfw2 -> lpst), DBE_VALUE);
	pldfw2 -> rcrc = DBC_CRCRecord(pldfw2, type);
	}

    /*
    Check for stopping pumps.
    */

    if (stoppingPumps != pldfw2 -> nstp)
	{
	pldfw2 -> nstp = stoppingPumps;
	db_post_events((struct dbCommon *)pldfw2, (caddr_t)&(pldfw2 -> nstp), DBE_VALUE);
	}

    /*
    Check for a change in pumps with run confirm.
    */

    if (runConfirm != pldfw2 -> nprc)
	{
	pldfw2 -> nprc = runConfirm;
	db_post_events((struct dbCommon *)pldfw2, (caddr_t)&(pldfw2 -> nprc), DBE_VALUE);
	}

    /*
    Check for a change in suspend mode input value.
    */

    if (susm != pldfw2 -> susm)
	{
	pldfw2 -> susm = susm;
	db_post_events((struct dbCommon *)pldfw2, (caddr_t)&(pldfw2 -> susm), DBE_VALUE);
	}

    /*
    Check for reportable change in pump speed.
    */

    pldfw2 -> pspd = rawSpeed;
    delta = pldfw2 -> pspd - pldfw2 -> mlst;
    if (delta < 0.0)
	delta = -delta;
    if (delta > pldfw2 -> mdel)
	{
	db_post_events((struct dbCommon *)pldfw2, (caddr_t)&(pldfw2 -> pspd), DBE_VALUE);
	pldfw2 -> mlst = pldfw2 -> pspd;
	}
    /*
    Check for change is status or severity, remembering if either
    one of them changed.
    */
    stat = pldfw2 -> stat;
    sevr = pldfw2 -> sevr;
    nsta = pldfw2 -> nsta;
    nsev = pldfw2 -> nsev;

    /* set current stat and sevr */
    pldfw2 -> stat = nsta;
    pldfw2 -> sevr = nsev;
    pldfw2 -> nsta = 0;
    pldfw2 -> nsev = 0;

    /* if we are coming out of VALID_ALARM and our new stat/sevr are
       the same as when we went into VALID_ALARM, do not set ack
    */
    if ( (nsev != VALID_ALARM) && (sevr == VALID_ALARM) &&
	 (nsta == pldfw2 -> psta) && (nsev == pldfw2 -> psev) )
    {
	set_ack = 0;
    }
    /* if we are going into VALID_ALARM, and our previous stat was not
       VALID_ALARM, remember previous non VALID_ALARM stat/sevr
       and we do not want to set ACK,
    */
    if ( (nsev == VALID_ALARM) && (sevr != VALID_ALARM) )
    {
	set_ack = 0;
	pldfw2 -> psta = stat;
	pldfw2 -> psev = sevr;
	db_post_events ((struct dbCommon *)pldfw2, (caddr_t)&pldfw2 -> psta, DBE_VALUE);
	db_post_events ((struct dbCommon *)pldfw2, (caddr_t)&pldfw2 -> psev, DBE_VALUE);
    }

    /* alarm condition changed this scan */

    if ((stat != nsta) || (sevr != nsev)) 
    {
	/* post events for alarm condition change*/

	eventMask = DBE_ALARM;

	/* post stat and nsev fields */

	db_post_events ((struct dbCommon *)pldfw2, (caddr_t)&pldfw2 -> stat, DBE_VALUE);
	db_post_events ((struct dbCommon *)pldfw2, (caddr_t)&pldfw2 -> sevr, DBE_VALUE);
	if ( (pldfw2 -> sevr != NO_ALARM) && (set_ack) &&
	     (pldfw2 -> sevr != EVENT_ALARM) )
	{
	    if (pldfw2 -> ack == 0) 
	    {
		pldfw2 -> ack = 1;           /* new unacknowledged alarm */
		db_post_events ((struct dbCommon *)pldfw2, (caddr_t)&pldfw2 -> ack, DBE_VALUE);
	    }
	}
    }
    /*
    Now look for a change in value.
    */

    if ((runningPumps + stoppingPumps) != pldfw2 -> val)
	{
	pldfw2 -> val = runningPumps + stoppingPumps;
	eventMask |= DBE_VALUE;
	if (pldfw2 -> tenb)
	    eventMask |= DBE_LOG;
	}
    if (eventMask & DBE_ALARM)
	{
	pldfw2 -> lalm = pldfw2 -> val;
	tsLocalTime(&(pldfw2 -> latm));
	}
    if (eventMask)
	db_post_events((struct dbCommon *)pldfw2, (caddr_t)&(pldfw2 -> val), eventMask);
}

/****************************************************************/
/*								*/
/*		DATABASE ACCESS UTILITY ROUTINES		*/
/*								*/
/****************************************************************/

static long get_precision(paddr, precision)
struct dbAddr	*paddr;
long		*precision;
{
    struct ldfw2Record *pldfw2 = (struct ldfw2Record *)(paddr -> precord);

    /*
    Only the pump speed has a precision.
    */

    if ( ((struct fldDes *)(paddr -> pfldDes)) -> offset ==
		(int) &(((struct ldfw2Record *) 0) -> pspd) )
	*precision = pldfw2 -> prec;
    else
    	*precision = 0;

    return (0);
}

static long get_graphic_double(paddr, pgd)
struct dbAddr		*paddr;
struct dbr_grDouble	*pgd;
{
    short		offset;

    /*
    Display range for the adjustment factors is 0.00 -> 9.99.
    Display range for everything else is 0 -> 100.0
    */

    offset = ((struct fldDes *)(paddr -> pfldDes)) -> offset;
    if ( (offset >= (int)&(((struct ldfw2Record *)0) -> af01))
    &&   (offset <= (int)&(((struct ldfw2Record *)0) -> af20)) )
	pgd -> upper_disp_limit = 9.99;
    else
	pgd -> upper_disp_limit = 100.0;
    pgd -> lower_disp_limit = 0;
    return (0);
}

static long get_control_double(paddr, pcd)
struct dbAddr		*paddr;
struct dbr_ctrlDouble	*pcd;
{
    short		offset;

    /*
    Control range for the adjustment factors is 0.00 -> 9.99.
    Control range for everything else is 0 -> 100.0
    */

    offset = ((struct fldDes *)(paddr -> pfldDes)) -> offset;
    if ( (offset >= (int)&(((struct ldfw2Record *)0) -> af01))
    &&   (offset <= (int)&(((struct ldfw2Record *)0) -> af20)) )
	pcd -> upper_ctrl_limit = 9.99;
    else
	pcd -> upper_ctrl_limit = 100.0;
    pcd -> lower_ctrl_limit = 0;
    return (0);
}
