
/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/

/************************************************************************/
/*                                                                      */
/*            IOC RESOURCE MANAGER TASK                                 */
/*                                                                      */
/*    This task exchanges status with the IOCs sharing responsibility   */
/*    for this IOC's resources.                                         */
/*                                                                      */
/*    Summary of execution:                                             */
/*    1. Initialization - Start the resource manager task. This task    */
/*        then creates and binds a listen socket to accept              */
/*        connections to partner IOCs. It then publishes the port       */
/*        value to the dbserver task.                                   */
/*    2. Periodic execution - Scan the resource table for resource      */
/*        status changes; notify the channel access users when          */
/*        the resources enter slave status.                             */
/*        Scan resource table for disconnected partners, connect        */
/*        if possible.                                                  */
/*        Build resource report messages for each connected             */
/*        partner IOC. Send the messages, mark failed connections.      */
/*    3. Event execution - Accept a new connection if the resource is   */
/*        disconnected.                                                 */
/*        Process received resource reports into the resource           */
/*        table's partner status fields.                                */
/*                                                                      */
/************************************************************************/

/************************* MOD LOG ***************************************
01/21/92    AGI    410-581-0422    irani@tate.com
        Original entry.
*************************************************************************/

/************************************************************************/
/*                                                                      */
/*            INCLUDE FILES                                             */
/*                                                                      */
/************************************************************************/

#include <vxWorks.h>
#include <sys/socket.h>
#include <in.h>
#include <errno.h>
#include <systime.h>
#include <bootLib.h>
#include <logLib.h>
#include <netLib.h>
#include <sockLib.h>
#include <iosLib.h>
#include <string.h>
#include <selectLib.h>
#include <hostLib.h>
#include <inetLib.h>
#include <tickLib.h>
#include <errnoLib.h>
#include <stdio.h>

#include "inet_ntoa.h"
#include "restable.h"
#include "resseq.h"
#include "dbDefs.h"
#include "dbCommon.h"
#include "task_params.h"        /* For getting task priorities */
#include "dbRecords.h"
#include "redmsgs.h"

/************************************************************************/
/*                                                                      */
/*            GLOBAL DATA WE DEFINE                                     */
/*                                                                      */
/************************************************************************/


/************************************************************************/
/*                                                                      */
/*            GLOBAL DATA WE REFERENCE                                  */
/*                                                                      */
/************************************************************************/
struct in_addr ourInetAddr;	/* local box IP address */

/************************************************************************/
/*                                                                      */
/*            TUNING PARAMETERS                                         */
/*                                                                      */
/************************************************************************/
#define	TICKSPERSEC	60	/* number of ticks per second on vxWorks */
#define	TICKSPERMIN	(60*TICKSPERSEC)

#define MC68000 1
#if CPU != MC68000
#define MAX_RESFDS      10      /* max number of partners we can handle */
#else
#define MAX_RESFDS	1
#endif

#define MAX_REPORTS     10      /* max number of reports per report msg */
#define MSGTIMEOUT      (10*TICKSPERSEC)	/* twenty seconds for now  */
#define RES_WAKEUPRATE	(6)		/* every tenth second */
#define RES_EXCHRATE	(2*TICKSPERSEC) /* status message exchange rate */
#define RES_REGRATE	(5*TICKSPERSEC) /* retry prog registration after 5 sec */
#define MAXREGS		5	/* max number of registrations/loop */

/************************************************************************/
/*                                                                      */
/*            OUR PRIVATE DATA                                          */
/*                                                                      */
/************************************************************************/

static int              rmfd;        /* fd for listen socket            */
static fd_set           rmReadFds;   /* fd mask for select()            */
static struct timeval   selectPoll = {1, 0}; /* select delay limit      */
static int		msgtimeout;
static int		res_exchrate;
static int		res_wakeuprate;
static int		res_regrate;
static int		maxregs;

static int		selectcalls;
static int		msgsent;
static int		msgrecv;
static int		regsent;
static int		regrecv;
static int		rptsent;
static int		rptrecv;

char			rmErrBuff [300];

#define	MAX_TIMEOUT	(1<<30) /* like 2 years */

#define	MAX_PROGID	256	/* maximum length of program ID string */

#define RESRPTMSG	0x8000	     /* resource report message */
#define	SEQREGMSG	0x8001	     /* seq registry message identifier */
#define	SEQRPTMSG	0x8002	     /* seq report message identifier */

typedef struct {
    unsigned short	msgtype;     /* resource message type (high bit
    					set to distinguish from reports) */
    unsigned short	msglen;	     /* total size of message */
    unsigned short	seq_token;   /* sequence program ID */
    unsigned short	seq_ptoken;  /* partner's sequence program ID */
    char		resource [16]; /* resource the program is based on */
    char		progid [MAX_PROGID]; /* room for prog name and beginning
					 macro table entries */
} RES_SEQREGMSG; 

#define MAX_SEQRPT	8192	     /* maximum state report size */

typedef struct {
    unsigned short	msgtype;     /* resource message type (high bit
					set to distinguish from reports) */
    unsigned short	msglen;	     /* total size of message */
    unsigned short	seq_token;   /* sequence program ID */
    unsigned short	seq_ptoken;  /* partner's sequence program ID */
    unsigned short	reportlen;   /* length of state report */
    char		report[MAX_SEQRPT]; /* the state report */
} RES_SEQRPTMSG;

/* these two structures transfer the status of the resources between systems */
typedef struct {
    char                name [16];   /* resource name                   */
    unsigned short      state;       /* partner's local state           */
} RES_REPORT;

typedef struct {
    unsigned short	msgtype;     /* resource message type (high bit
    					set to distinguish from reports) */
    unsigned short	msglen;	     /* total size of message */
    unsigned short      repCnt;     /* number of reports in message (<0x8000 !!) */
    RES_REPORT          report[MAX_REPORTS]; /* state reports          */
} RES_MESSAGE;

/* This structure provides a maximum size for message data */
typedef union {
    RES_MESSAGE		 resmsg;     /* resource report message */
    RES_SEQREGMSG	 resseqreg;  /* sequence registration message */
    RES_SEQRPTMSG	 resseqrpt;  /* sequence state report message */
} RES_MSGUNION;

/* This is a reserved area for message information */
static RES_MSGUNION      msg;        /* communication message area      */
#define REPMSGLEN	 (sizeof(RES_MESSAGE))
#define MAXMSGLEN	 (sizeof(RES_MSGUNION))

#define	NOCONNECT	0	/* no partner sequence prog */
#define NEWCONNECT	1	/* new connection to partner sequence program */
#define CONNECT		2	/* exchanging states with partner sequence program */

/* this structure hold data about each sequence program */
typedef struct {
    void		*next;       /* next sequence program */
    unsigned short	seq_token;   /* sequence program ID */
    unsigned short	seq_ptoken;  /* partner's sequence program ID */
    unsigned short	linkstatus;  /* status of state sharing */
    long	        lastmsg;     /* last activity time	*/
    void		*stateSem;   /* mutex for program data */
    short		seq_replen;  /* length of sequence report data */
    char		*seq_report; /* sequence report data */
    int			need_txmit;  /* outbound data needs transmission */
    short		tx_replen;   /* length of outbound report data */
    RES_SEQRPTMSG	*tx_report;  /* outbound report data */
    void		(*usrFunc)();/* user callback */
    long		usrArg;      /* user argument */
    char		resource [16]; /* resource the program is based on */
    char		progid [MAX_PROGID]; /* room for prog name and beginning
					 macro table entries */
} RES_SEQPROG; 

void			*seqSem;     /* mutex for program list */
static unsigned short	next_token;  /* program ID token value seed */
static RES_SEQPROG	*seqprograms;/* pointer to list of sequence programs */

typedef struct {
    struct in_addr       Inet;       /* connected Inet address          */
    int                  fd;         /* connection fd                   */
    long	         lastmsg;    /* last message report rx time	*/
    int			 recvlen;    /* number of bytes in recvbuf	*/
    char		 recvbuf[MAXMSGLEN]; /* recvbuf for connection */
} RES_CONNECTION;

static RES_CONNECTION    resFds [MAX_RESFDS]; /* current connections       */

int RESMGRDEBUG;

/************************************************************************/
/*                                                                      */
/*                Function prototypes                                   */
/*                                                                      */
/************************************************************************/
void acceptNewPartner (int rmfd);
void readPartnerMsg (RES_CONNECTION *conn);
void sendSeqStates(void);
int sendSeqRegistration (RES_SEQPROG *seqp, RES_TABLE *rp);
void makeResConnections (void);
void connectPartner (RES_TABLE *rp);
void sendResReports (void);
int rmConnDsp (void);
void resMgrTask(void);
/************************************************************************/
/*                                                                      */
/*                INITIALIZE                                            */
/*                                                                      */
/************************************************************************/

int resMgrInit(void)
{
    RESMGRDEBUG = 0; /* start with debug off */
    msgtimeout = MSGTIMEOUT;
    res_wakeuprate = RES_WAKEUPRATE;
    res_exchrate = RES_EXCHRATE;
    res_regrate = RES_REGRATE;
    maxregs = MAXREGS;

    /*
    Spawn the resMgr task. For now, use the same
    priority as the periodic scan task.
    */

    if (taskSpawn(RESMGR_NAME, RESMGR_PRI, VX_FP_TASK, RESMGR_STACK,
                   (FUNCPTR)resMgrTask,0,0,0,0,0,0,0,0,0,0) == ERROR) {
        logMsg("Can't spawn resMgrTask.\n",0,0,0,0,0,0);
        return (-1);
    }

    /*
    Initialization complete
    */
    return(0);
}

/****************************************************************/
/*                                                              */
/*            THE RESOURCE MANAGER TASK                         */
/*                                                              */
/*    This is the resource manager task. It expects to run      */
/*    in conjunction with the periodic scan task, so they       */
/*    are set up to run at the same priority, at least          */
/*    for the time being.                                       */
/*                                                              */
/****************************************************************/

void resMgrTask(void)
{
    long            nextTime;
    long            nextWakeup;
    long            delay;
    struct sockaddr_in serverAddr;    /* server's address   */
    short           serveport;
    FAST int        status;
    FAST int        i;
    RES_TABLE       *rp;
    RES_SEQPROG     *seqp;
    fd_set          tempfds;
    int		    addrSize;

    if (rmfd != 0 && rmfd != ERROR)
        if ((status = close(rmfd)) == ERROR)
            logMsg("Unable to close open resMgr socket\n",0,0,0,0,0,0);

    /* Open a stream socket. */
    if ((rmfd = socket(AF_INET, SOCK_STREAM, 0)) == ERROR) {
        logMsg("Socket creation error\n",0,0,0,0,0,0);
        taskSuspend(0);
    }
    
    /* Zero the sock_addr structure */
    bfill((char *)&serverAddr, sizeof(serverAddr), 0);
    serverAddr.sin_family = AF_INET;
    serverAddr.sin_port = htons(0);

    /* get server port address */
    if (bind(rmfd, (struct sockaddr *)&serverAddr, sizeof(serverAddr)) == ERROR) {
        logMsg("Bind error\n",0,0,0,0,0,0);
        close(rmfd);
	rmfd = 0;
        taskSuspend(0);
    }
    addrSize = sizeof(serverAddr);
    if (getsockname (rmfd, (struct sockaddr *)&serverAddr, &addrSize)) {
        logMsg("ResMgr:getsockname error\n",0,0,0,0,0,0);
        close(rmfd);
	rmfd = 0;
        taskSuspend(0);
    }
    serveport = htons (serverAddr.sin_port);

    /* mutex sem to protect sequence list structure */
    seqSem = semBCreate(SEM_Q_FIFO, SEM_FULL);

    /* clear fd mask, listen and accept new connections */
    FD_ZERO (&rmReadFds);
    FD_SET (rmfd, &rmReadFds); /* select on incoming connections */
    if (listen(rmfd, 0) == ERROR) {
        logMsg("Listen error\n",0,0,0,0,0,0);
        close(rmfd);
	rmfd = 0;
        taskSuspend(0);
    }
    nextTime = tickGet () + res_exchrate;
    nextWakeup = tickGet () + res_wakeuprate;

    while (TRUE) {
	/* reset select delay time for next report */
        delay = nextWakeup - tickGet ();
        if (delay < 0)
	{
	    logMsg ("fallen behind by %d, adjusting nextWakeup to compensate\n", -delay,0,0,0,0,0);
            delay = 0;
	    nextWakeup = tickGet () + res_wakeuprate;
	}
        selectPoll.tv_sec = delay / 60;
        selectPoll.tv_usec = (delay % 60) * ((1000000 / 60) + 1);
	tempfds = rmReadFds;
	selectcalls++;
	status = select (sizeof(tempfds)*8, &tempfds,
                 NULL, NULL, &selectPoll);
        if (status > 0) { /* One of the targets has activity */
            if (FD_ISSET (rmfd, &tempfds)) {
                if (RESMGRDEBUG > 2)
		    logMsg("ResMgr select activity on listen port\n",0,0,0,0,0,0);
                acceptNewPartner (rmfd);
                status--;
            }
            /* check the resource table fds for incoming signals */
	    for (i = 0; i < MAX_RESFDS && status; i++) {
                if (resFds[i].Inet.s_addr && FD_ISSET (resFds[i].fd,
                                 &tempfds)) {
	            if (RESMGRDEBUG > 1)
			logMsg("ResMgr select activity on %d\n", resFds[i].fd,0,0,0,0,0);
                    readPartnerMsg (&resFds[i]);
                    status--;
                }
            }
	    if (status) /* We must have a surprise fd set in the mask */
	    {
		sprintf (rmErrBuff, "Unexpected select bit set. Read mask %8.8lx%8.8lx%8.8lx, Result mask %8.8lx%8.8lx%8.8lx\n",
				*(long *)&rmReadFds, *(long *)(((char *)&rmReadFds)+4), *(long *)(((char *)&rmReadFds)+8),
				*(long *)&tempfds, *(long *)(((char *)&tempfds)+4), *(long *)(((char *)&tempfds)+8));
		/* rebuild the fd mask */
		FD_ZERO (&rmReadFds);
		FD_SET (rmfd, &rmReadFds); /* select on incoming connections */
		for (i = 0; i < MAX_RESFDS; i++) {
		    if (resFds[i].fd)
			FD_SET (resFds[i].fd, &rmReadFds);
		}
		sprintf (&rmErrBuff[strlen(rmErrBuff)], "       Reconstructed mask %8.8lx%8.8lx%8.8lx\n",
				*(long *)&rmReadFds, *(long *)(((char *)&rmReadFds)+4), *(long *)(((char *)&rmReadFds)+8));
		logMsg(rmErrBuff,0,0,0,0,0,0);
	    }
        }
        else if (status < 0) { /* error processing */
            logMsg("error on select statement\n",0,0,0,0,0,0);
        }

	/* process any state transfers */
 	sendSeqStates ();
	nextWakeup = tickGet () + res_wakeuprate;


	/* timed event processing */
        if (tickGet() >= nextTime) {
            if (RESMGRDEBUG > 3)
		logMsg("ResMgr timed scans\n",0,0,0,0,0,0);
            nextTime += res_exchrate; /* long for debugging */

            /* timed functions - CA deletion */
            /* wipe out channels on change to Slave status !!! */
            for (rp = resTableHead; rp; rp = rp->link) {
                /* if the driver is active,
		** tell dbServer the listen port for resMgr */
                if (rp->localState) {
		    if (rp->driverPort == 0) {
                        resRegisterDriver(rp, serveport, 0, 5);
                    }
                    else {
			if (rp->partnerConnected) {
                            /* clean up orphan connections */
                            for (i = 0; i < MAX_RESFDS; i++) {
                                if (rp->partnerFd == resFds[i].fd) {
                                    break;
                                }
                            }
                            if (i == MAX_RESFDS) {
                                rp->partnerFd = 0;
                	        rp->partnerPort = 0;
                                rp->partnerConnected = FALSE;
				rp->partnerState = RES_ISOLATED;
				/* scan seq programs for dependance on the new connection */
				semTake(seqSem, MAX_TIMEOUT );
				for (seqp = seqprograms; seqp; seqp = seqp->next) {
				    if (strcmp (rp->resourceName, seqp->resource) == 0) {
					seqp->lastmsg = tickGet ();
					seqp->linkstatus = NOCONNECT;
					seqp->seq_ptoken = 0;
				    }
                		}
				semGive (seqSem);
			    }
                        }
                        if (rp->localState & RES_CHANGED) {
			    semTake(seqSem, MAX_TIMEOUT );
			    for (seqp = seqprograms; seqp; seqp = seqp->next) {
				if (strcmp (rp->resourceName, seqp->resource) == 0) {
				    /* found a matching resource, wake up */
				    (*seqp->usrFunc)(seqp->usrArg);
				}
			    }
			    semGive (seqSem);
                            rp->localState &= ~RES_CHANGED;
                        }
                    }
		}
            }

	    /* wipe out TCP on timeout of message in!! */
	    for (i = 0; i < MAX_RESFDS; i++) {
		if (resFds[i].fd
		  && ((tickGet() - resFds[i].lastmsg) > msgtimeout)) {
		    /* close the connection */
		    for (rp = resTableHead; rp; rp = rp->link) {
	                if (rp->partnerInet.s_addr == resFds[i].Inet.s_addr) {
			    if (rp->partnerFd == 0) {
			        logMsg("to, inconsistent Fd values",0,0,0,0,0,0);
	                    }
                	    else {
                	        rp->partnerFd = 0;
                	        rp->partnerPort = 0;
			        rp->partnerConnected = FALSE;
				rp->partnerState = RES_ISOLATED;
				/* scan seq programs for dependance on the new connection */
				semTake(seqSem, MAX_TIMEOUT );
				for (seqp = seqprograms; seqp; seqp = seqp->next) {
				    if (strcmp (rp->resourceName, seqp->resource) == 0) {
					seqp->lastmsg = tickGet ();
					seqp->linkstatus = NOCONNECT;
					seqp->seq_ptoken = 0;
				    }
                		}
				semGive (seqSem);
			    }
	                }
        	    }
                    logMsg ("closing socket %d, updates timed out\n",
                               resFds[i].fd,0,0,0,0,0);
		    /* close socket */
                    close(resFds[i].fd);
                    FD_CLR (resFds[i].fd, &rmReadFds);
		    resFds[i].Inet.s_addr = 0;
		    resFds[i].fd = 0;
		    resFds[i].lastmsg = 0;
		}
            }

	    /* make connections */
	    makeResConnections ();

            /* send resource reports */
            sendResReports ();

        }
    }
}

void acceptNewPartner (int rmfd)
{
    int     tempfd;
    struct  sockaddr_in newaddress;
    RES_TABLE *rp = 0;
    RES_SEQPROG *seqp;
    int	    i;
    int     blankIx;
    int     addrlen = sizeof(newaddress);
    int	    regsSent;

    if ((tempfd = accept(rmfd, (struct sockaddr *)&newaddress, &addrlen)) != ERROR) {
        /* scan the fd table for the connecting IP */
        blankIx = MAX_RESFDS;
	for (i = 0; i < MAX_RESFDS; i++) {
	    if (resFds[i].Inet.s_addr == newaddress.sin_addr.s_addr) {
                /* already got him, close new connection & go home */
		if (rp->partnerFd) {
		    close (tempfd);
                    return;
		}
	    }
	    else if ((resFds[i].Inet.s_addr == 0) && (blankIx == MAX_RESFDS)) {
		blankIx = i;
            }
	}
	if (i == MAX_RESFDS) {
            if (blankIx < MAX_RESFDS) {
		resFds[blankIx].Inet = newaddress.sin_addr;
		resFds[blankIx].fd = tempfd;
		resFds[blankIx].lastmsg = tickGet ();
                resFds[blankIx].recvlen = 0;
                FD_SET (tempfd, &rmReadFds);
		logMsg("accepted new partner\n",0,0,0,0,0,0);
                rmConnDsp ();
            }
	    else /* no more room in table */
	    {
  	        logMsg("accept out of resource fd map space\n",0,0,0,0,0,0);
		taskSuspend(0);
            }
        }

	for (rp = resTableHead; rp; rp = rp->link) {
            if (rp->partnerInet.s_addr == newaddress.sin_addr.s_addr) {
		if (rp->partnerFd) {
		    logMsg("accept, inconsistent Fd values\n",0,0,0,0,0,0);
                }
                rp->partnerFd = tempfd;
                rp->partnerConnected = TRUE;
		rp->partnerState = 0;
		/* scan seq programs for dependance on the new connection */
		semTake(seqSem, MAX_TIMEOUT );
		for (regsSent = 0, seqp = seqprograms; seqp; seqp = seqp->next) {
		    if (strcmp (rp->resourceName, seqp->resource) == 0) {
                    	if (seqp->seq_ptoken) {
			    seqp->lastmsg = tickGet ();
			    seqp->linkstatus = NEWCONNECT;
			}
			else if ((tickGet () - seqp->lastmsg) > res_regrate) {
			    if (regsSent < maxregs)
				regsSent += sendSeqRegistration (seqp, rp);
                        }
		    }
                }
		semGive (seqSem);
	    }
        }
    }
    else {
        logMsg("Accept error\n",0,0,0,0,0,0);
        taskSuspend(0);
    }
}

void readPartnerMsg (RES_CONNECTION *conn)
{
    RES_TABLE *rp;
    RES_SEQPROG *seqp;
    int    i;
    int    status;
    RES_REPORT *report;
    int	   msglen;

    /* handle partial reads!  Read all they have into end of buffer, use all we can. */
    status = recv (conn->fd, &conn->recvbuf[conn->recvlen], sizeof(conn->recvbuf) - conn->recvlen, 0);

    /* if we have a illegal message, force a failure */
    if  ((status != -1) && ((status + conn->recvlen) >= 4)
      && ((((RES_MESSAGE *)conn->recvbuf)->msglen >= MAXMSGLEN) || (((RES_MESSAGE *)conn->recvbuf)->msglen) < 4)) {
	logMsg ("illegal message length %d, closing connection.\n", ((RES_MESSAGE *)conn->recvbuf)->msglen,0,0,0,0,0 );
	if (RESMGRDEBUG)
	    rmConnDsp ();
	status = -1;
    }

    /* if we got data, process the resource reports in the message */
    if (status > 0) {
	conn->recvlen += status;
	while ((conn->recvlen > 4) && (conn->recvlen >= (msglen = ((RES_MESSAGE *)conn->recvbuf)->msglen))) {
	    switch (((RES_MESSAGE *)conn->recvbuf)->msgtype) {
		case RESRPTMSG:
		    msgrecv++;
		    if (RESMGRDEBUG > 1)
			logMsg("Update %x, length %d\n", ((RES_MESSAGE *)conn->recvbuf)->repCnt, ((RES_MESSAGE *)conn->recvbuf)->msglen,0,0,0,0);
		    for (i = 0; i < MAX_RESFDS; i++) {
        		if (resFds[i].fd == conn->fd) {
			    resFds[i].lastmsg = tickGet ();
	                    break;
	        	}
		    }
	            for (report = &((RES_MESSAGE *)conn->recvbuf)->report[0];
		      ((RES_MESSAGE *)conn->recvbuf)->repCnt;
		      ((RES_MESSAGE *)conn->recvbuf)->repCnt--, report++) {
	        	if ((resGetByName (report->name, &rp) == RES_OK)
        		  && (rp->partnerFd)) {
                	    rp->partnerState = report->state;
	        	}
        		else {
	                    logMsg("recv, resource or fd missing\n",0,0,0,0,0,0);
		            /* should get partner to refind his IP addr */
        		}
		    }
		    break;
 		case SEQREGMSG:
		    regrecv++;
		    if (RESMGRDEBUG)
			logMsg("Reg %x, length %d\n", ((RES_SEQREGMSG *)conn->recvbuf)->seq_token, ((RES_MESSAGE *)conn->recvbuf)->msglen,0,0,0,0);
		    semTake(seqSem, MAX_TIMEOUT );
		    /* search for current registration, update if necessary */
		    for (seqp = seqprograms; seqp; seqp = seqp->next) {
			if ((strcmp (((RES_SEQREGMSG *)conn->recvbuf)->resource, seqp->resource) == 0)
			  && (strcmp (((RES_SEQREGMSG *)conn->recvbuf)->progid, seqp->progid) == 0)) {
			    semTake (seqp->stateSem, MAX_TIMEOUT);
			    seqp->seq_ptoken = ((RES_SEQREGMSG *)conn->recvbuf)->seq_token;
			    seqp->linkstatus = NEWCONNECT;
			    semGive (seqp->stateSem);
			    (*seqp->usrFunc) (seqp->usrArg);
			    seqp->lastmsg = tickGet ();
		            if ((((RES_SEQREGMSG *)conn->recvbuf)->seq_ptoken != seqp->seq_token)
		              && (resGetByName (seqp->resource, &rp) == RES_OK))
 				sendSeqRegistration (seqp, rp);
                            break;
			}
		    }
		    semGive (seqSem);
		    break;
		case SEQRPTMSG:
		    rptrecv++;
		    if (RESMGRDEBUG)
			logMsg("State for program %x, length %d\n",
				((RES_SEQRPTMSG *)conn->recvbuf)->seq_ptoken, ((RES_MESSAGE *)conn->recvbuf)->msglen,0,0,0,0);
		    /* search for current registration, update if necessary */
		    semTake(seqSem, MAX_TIMEOUT );
		    for (seqp = seqprograms; seqp; seqp = seqp->next) {
			if (((RES_SEQRPTMSG *)conn->recvbuf)->seq_ptoken == seqp->seq_token) {
			    if (seqp->seq_report &&
			      (seqp->seq_replen < ((RES_SEQRPTMSG *)conn->recvbuf)->reportlen)) {
				free (seqp->seq_report);
				seqp->seq_report = 0;
			    }
			    if (seqp->seq_report == 0)
				seqp->seq_report = malloc (((RES_SEQRPTMSG *)conn->recvbuf)->reportlen);
			    if (seqp->seq_report != 0) {
				semTake (seqp->stateSem, MAX_TIMEOUT);
				memcpy (seqp->seq_report, ((RES_SEQRPTMSG *)conn->recvbuf)->report,
				        ((RES_SEQRPTMSG *)conn->recvbuf)->reportlen);
				seqp->seq_replen = ((RES_SEQRPTMSG *)conn->recvbuf)->reportlen;
				seqp->lastmsg = tickGet ();
				semGive (seqp->stateSem);
				(*seqp->usrFunc) (seqp->usrArg);
			    }
			    break;
			}
        	    }
		    semGive (seqSem);
		    break;
		default:
		    /* unknown message type, dump all collected data */
		    logMsg ("Recv, illegal message received\n",0,0,0,0,0,0);
	            /* close the fd struct */ 
		    for (rp = resTableHead; rp; rp = rp->link) {
	                if (rp->partnerInet.s_addr == conn->Inet.s_addr) {
			    if (rp->partnerFd == 0) {
			        logMsg("recv, inconsistent Fd values\n",0,0,0,0,0,0);
	                    }
	                    else {
	                        rp->partnerFd = 0;
                		rp->partnerPort = 0;
			        rp->partnerConnected = FALSE;
			        rp->partnerState = RES_ISOLATED;
			        /* scan seq programs for dependance on the new connection */
				semTake(seqSem, MAX_TIMEOUT );
			        for (seqp = seqprograms; seqp; seqp = seqp->next) {
				    if (strcmp (rp->resourceName, seqp->resource) == 0) {
					seqp->lastmsg = tickGet ();
					seqp->linkstatus = NOCONNECT;
					seqp->seq_ptoken = 0;
	               	            }
			        }
				semGive (seqSem);
	                    }
	                }
		    }
		    conn->Inet.s_addr = 0;
		    conn->recvlen = 0;
		    msglen = 0;
		    conn->lastmsg = 0;
		    /* close socket */
		    close (conn->fd);
	            FD_CLR (conn->fd, &rmReadFds);
		    conn->fd = 0;
		    break;
	    }
	    /* adjust leftover bytes */
	    if ((conn->recvlen - msglen) > 0)
	    {
		if (RESMGRDEBUG)
		    logMsg ("Shifting %d remaining message bytes,\n", conn->recvlen - msglen,0,0,0,0,0);
	         memcpy (conn->recvbuf, &conn->recvbuf[msglen],
	                 conn->recvlen - msglen);
	    }
	    conn->recvlen -= msglen;
 	}
    }
    else if (status <= 0) {
	if (status < 0)
	{
            /* handle TCP errors, close the connection!! */
	    int	anerrno;
	    anerrno = errnoGet();
	    if((anerrno!=ECONNABORTED && anerrno!=ECONNRESET && anerrno!=ETIMEDOUT)) {
		logMsg("recv, client disconnect\n",0,0,0,0,0,0);
	    }
	}
	else
	{
	    logMsg("recv, no data after select\n",0,0,0,0,0,0);
	}
        /* close the fd struct */ 
	for (rp = resTableHead; rp; rp = rp->link) {
            if (rp->partnerInet.s_addr == conn->Inet.s_addr) {
		if (rp->partnerFd == 0) {
		    logMsg("recv, inconsistent Fd values\n",0,0,0,0,0,0);
                }
                else {
                    rp->partnerFd = 0;
                    rp->partnerPort = 0;
		    rp->partnerConnected = FALSE;
		    rp->partnerState = RES_ISOLATED;
		    /* scan seq programs for dependance on the new connection */
		    semTake(seqSem, MAX_TIMEOUT );
		    for (seqp = seqprograms; seqp; seqp = seqp->next) {
			if (strcmp (rp->resourceName, seqp->resource) == 0) {
			    seqp->lastmsg = tickGet ();
			    seqp->linkstatus = NOCONNECT;
			    seqp->seq_ptoken = 0;
		        }
               	    }
		    semGive(seqSem);
		}
            }
        }
	conn->Inet.s_addr = 0;
	conn->lastmsg = 0;
	conn->recvlen = 0;
	/* close socket */
	close (conn->fd);
        FD_CLR (conn->fd, &rmReadFds);
	conn->fd = 0;
    }
}

void makeResConnections (void)
{
    RES_TABLE *rp;
    RES_SEQPROG *seqp;
    int    i;
    int	   regsSent;

    for (rp = resTableHead; rp; rp = rp->link) {
        if (rp->partnerFd == 0 && rp->partnerInet.s_addr) {
            /* first check for another connection to the guy */
            for (i = 0; i < MAX_RESFDS; i++) {
                if (resFds[i].Inet.s_addr == rp->partnerInet.s_addr) {
                    rp->partnerFd = resFds[i].fd;
                    rp->partnerConnected = TRUE;
		    rp->partnerState = 0;
		    /* scan seq programs for dependance on the new connection */
		    semTake(seqSem, MAX_TIMEOUT );
		    for (regsSent = 0, seqp = seqprograms; seqp; seqp = seqp->next) {
		        if (strcmp (rp->resourceName, seqp->resource) == 0) {
			    if (seqp->seq_ptoken) {
			 	seqp->lastmsg = tickGet ();
			        seqp->linkstatus = NEWCONNECT;
			    }
			    else if ((tickGet () - seqp->lastmsg) > res_regrate) {
				if (regsSent < maxregs)
				    regsSent += sendSeqRegistration (seqp, rp);
                            }
                        }
		    }
		    semGive(seqSem);
		    break;
                }
            }
            /* if the fd is still 0, and he has a lower IP, try to connect him */
            if ((rp->partnerInet.s_addr) && (rp->partnerInet.s_addr < ourInetAddr.s_addr)
	      && (rp->partnerPort) && (rp->partnerFd == 0)) {
                connectPartner (rp);
            }
        }
	else {
	    /* scan seq programs for unregistered partners */
	    semTake(seqSem, MAX_TIMEOUT );
	    for (regsSent = 0, seqp = seqprograms; seqp; seqp = seqp->next) {
	        if (strcmp (rp->resourceName, seqp->resource) == 0) {
		    if (rp->partnerFd && (seqp->seq_ptoken == 0)
		      && ((tickGet () - seqp->lastmsg) > res_regrate)) {
			if (regsSent < maxregs)
			    regsSent += sendSeqRegistration (seqp, rp);
                    }
                }
	    }
	    semGive(seqSem);
	}
    }
}

void sendSeqStates(void)
{
    RES_TABLE *rp;
    RES_SEQPROG *seqp;
    RES_SEQPROG *seqp2;
    int    i;
    int	   status;

    /* scan seq programs for state transfers */
    semTake(seqSem, MAX_TIMEOUT );
    for (seqp = seqprograms; seqp; seqp = seqp->next) {
	if ( (resGetByName (seqp->resource, &rp) == RES_OK)
          && (seqp->need_txmit && rp->partnerFd)) {
   	    if ((status = send (rp->partnerFd, (char *)seqp->tx_report, seqp->tx_replen, 0)) != seqp->tx_replen) {
		if (status <= 0) {
		    /* handle TCP errors, close the connection!! */
		    int	anerrno;
		    anerrno = errnoGet();
		    if((anerrno!=ECONNABORTED && anerrno!=ECONNRESET && anerrno!=ETIMEDOUT)) {
		        logMsg("send, client disconnect\n",0,0,0,0,0,0);
		    }
	            /* find the connection to the guy */
 		    for (i = 0; i < MAX_RESFDS; i++) {
	                if (rp->partnerInet.s_addr == resFds[i].Inet.s_addr) {
			    if (rp->partnerFd == 0) {
			        logMsg("send, inconsistent Fd values\n",0,0,0,0,0,0);
	                    }
			    else {
				rp->partnerFd = 0;
				rp->partnerPort = 0;
				rp->partnerConnected = FALSE;
				rp->partnerState = RES_ISOLATED;
				/* scan seq programs for dependance on the connection */
				for (seqp2 = seqprograms; seqp2; seqp2 = seqp2->next) {
				    if (strcmp (rp->resourceName, seqp2->resource) == 0) {
				 	seqp2->lastmsg = tickGet ();
					seqp2->linkstatus = NOCONNECT;
					seqp2->seq_ptoken = 0;
				    }
	        		}
				/* close socket */
				close(resFds[i].fd);
				FD_CLR (resFds[i].fd, &rmReadFds);
				resFds[i].Inet.s_addr = 0;
				resFds[i].fd = 0;
				resFds[i].lastmsg = 0;
			    }
			    break;
	                }
 		    }
		}
		else {
		    /* short write - ??? */
		    logMsg("short send attempt!!",0,0,0,0,0,0);
		}
	    }
	    else {
	        seqp->lastmsg = tickGet ();
		seqp->need_txmit = 0;
		rptsent++;

	    }
	}
    }
    semGive(seqSem);
}

void sendResReports (void)
{
    int    i;
    int    status;
    RES_TABLE *rp;
    RES_SEQPROG *seqp;
    RES_REPORT *report;

    for (i = 0; i < MAX_RESFDS; i++) {
        if (resFds[i].Inet.s_addr) {
	    msg.resmsg.msgtype = RESRPTMSG;
	    msg.resmsg.repCnt = 0;
	    report = &msg.resmsg.report[0];
            for (rp = resTableHead; rp; rp = rp->link) {
                if (resFds[i].Inet.s_addr == rp->partnerInet.s_addr) {
                    strncpy (report->name,
                        rp->resourceName,
                        sizeof(report->name));
                    report->state = rp->localState;
                    report++;
                    msg.resmsg.repCnt++;
                }
            }
	    msg.resmsg.msglen = sizeof(RES_MESSAGE) - sizeof(msg.resmsg.report)
				+ (msg.resmsg.repCnt * sizeof(RES_REPORT));
	    /* send the flipping message!! */
	    if ((status = send (resFds[i].fd, (char *)&msg, msg.resmsg.msglen, 0)) != msg.resmsg.msglen) {
            	if (status <= 0) {
        	    /* handle TCP errors, close the connection!! */
		    int	anerrno;
		    anerrno = errnoGet();
		    if((anerrno!=ECONNABORTED && anerrno!=ECONNRESET && anerrno!=ETIMEDOUT)) {
		        logMsg("send, client disconnect\n",0,0,0,0,0,0);
		    }
		    for (rp = resTableHead; rp; rp = rp->link) {
	                if (rp->partnerInet.s_addr == resFds[i].Inet.s_addr) {
			    if (rp->partnerFd == 0) {
			        logMsg("send, inconsistent Fd values\n",0,0,0,0,0,0);
	                    }
                	    else {
                	        rp->partnerFd = 0;
                	        rp->partnerPort = 0;
			        rp->partnerConnected = FALSE;
				rp->partnerState = RES_ISOLATED;
				/* scan seq programs for dependance on the new connection */
				semTake(seqSem, MAX_TIMEOUT );
				for (seqp = seqprograms; seqp; seqp = seqp->next) {
				    if (strcmp (rp->resourceName, seqp->resource) == 0) {
				 	seqp->lastmsg = tickGet ();
					seqp->linkstatus = NOCONNECT;
					seqp->seq_ptoken = 0;
				    }
                		}
				semGive(seqSem);
			    }
	                }
        	    }
		    /* close socket */
                    close(resFds[i].fd);
                    FD_CLR (resFds[i].fd, &rmReadFds);
		    resFds[i].Inet.s_addr = 0;
		    resFds[i].fd = 0;
		    resFds[i].lastmsg = 0;
		}
		else {
		    /* short write - ??? */
		    logMsg("short send attempt!!",0,0,0,0,0,0);
                }
	    }
	    else {
		msgsent++;
	    }
	}
    }
}

void connectPartner (RES_TABLE *rp)
{
    RES_SEQPROG *seqp;
    int    tempfd;
    int    status;
    struct sockaddr_in newaddress;
    int    i;
    char    true = TRUE;
    int	   regsSent;

    if((tempfd = socket (AF_INET, SOCK_STREAM, 0)) == ERROR){
        logMsg("connect socket creation error\n",0,0,0,0,0,0);
        taskSuspend (0);
    }

    status = setsockopt(tempfd, SOL_SOCKET, SO_KEEPALIVE, &true, sizeof(true));
    if(status<0){
        logMsg("connect socket %d couldn't set keepalive error %d\n", tempfd, errno, 0, 0, 0, 0);
    }

    bfill((char *)&newaddress, sizeof newaddress, 0);
    newaddress.sin_family = AF_INET;
    newaddress.sin_addr = rp->partnerInet;
    newaddress.sin_port = htons(rp->partnerPort);
    if (connect(tempfd, (struct sockaddr *)&newaddress, sizeof(newaddress))) {
        /* connection error - delete temporary socket, try later */
        close (tempfd);
    }
    else {
        /* install fd in rp */
        for (i = 0; i < MAX_RESFDS; i++) {
            if (resFds[i].Inet.s_addr == 0) {
                resFds[i].Inet = rp->partnerInet;
                resFds[i].fd = tempfd;
                FD_SET (resFds[i].fd, &rmReadFds);
		resFds[i].lastmsg = tickGet ();
		rp->partnerFd = tempfd;
		rp->partnerConnected = TRUE;
		rp->partnerState = 0;
		/* scan seq programs for dependance on the new connection */
		semTake(seqSem, MAX_TIMEOUT );
		for (regsSent = 0, seqp = seqprograms; seqp; seqp = seqp->next) {
		    if (strcmp (rp->resourceName, seqp->resource) == 0) {
			if (seqp->seq_ptoken) {
			    seqp->lastmsg = tickGet ();
			    seqp->linkstatus = NEWCONNECT;
			}
			else if ((tickGet () - seqp->lastmsg) > res_regrate) {
                            if (regsSent < maxregs)
				regsSent += sendSeqRegistration (seqp, rp);
                        }
		    }
       		}
		semGive(seqSem);
		break;
            }
        }
        if (i == MAX_RESFDS) {
            logMsg("out of resource fd map space\n",0,0,0,0,0,0);
            taskSuspend(0);
        }
        logMsg ("connected to partner\n",0,0,0,0,0,0);
        rmConnDsp ();
    }
}

/* sequence program support */
int resSeqInit (progName, resName, macroString, func, arg)
char	*progName;
char	*resName;
char	*macroString;
void    (*func)();
long    arg;
{
    RES_SEQPROG *newprog;	/* new program pointer */
    RES_SEQPROG *curprog;	/* last program pointer */

    semTake(seqSem, MAX_TIMEOUT );

    for (curprog = (RES_SEQPROG *)&seqprograms; curprog->next; curprog = curprog->next) {
	/* ensure that the program is not already registered */
	if ( (strncmp (progName, ((RES_SEQPROG *)curprog->next)->progid, strlen(progName)) == 0)
	  && (strcmp (resName, ((RES_SEQPROG *)curprog->next)->resource) == 0)
	  && (strcmp (macroString, &((RES_SEQPROG *)curprog->next)->progid[strlen(progName)]) == 0))
	{
	    /* already registered, return an error */
 	logMsg ( "New Seq prog failed %s %-16.16s %s\n",
		   (int)progName,
		   (int)resName,
                   (int)macroString,0,0,0);
	    
            return 0;
	}
    }

    if ((strlen (progName) + strlen (macroString)) >= (MAX_PROGID - 2))
    {
	logMsg ("Program Id + MacroString exceeds buffer capacity\n",0,0,0,0,0,0);
	semGive (seqSem);
	taskSuspend (taskIdSelf ());
    }

    newprog = malloc ((sizeof(RES_SEQPROG)-MAX_PROGID)
			+strlen(progName)+strlen(macroString)+2);
    if (newprog) {
	newprog->next = 0;
	newprog->seq_token = ++next_token;
	newprog->lastmsg = tickGet ();
        newprog->linkstatus = NOCONNECT;
	newprog->stateSem = semBCreate(SEM_Q_FIFO, SEM_FULL);
	newprog->seq_ptoken = 0;
	newprog->seq_replen = 0;
	newprog->seq_report = 0;
	newprog->need_txmit = 0;
	newprog->tx_replen = 0;
	newprog->tx_report = 0;
	newprog->usrFunc = func;
	newprog->usrArg = arg;
	strcpy (newprog->resource, resName);
	strcpy (newprog->progid, progName);
	strcpy (&newprog->progid[strlen(progName)], macroString);
	curprog->next = newprog;
	if (RESMGRDEBUG)
 	    logMsg ( "New Seq prog %4.4d %-16.16s %s\n",
			   newprog->seq_token,
			   (int)newprog->resource,
                	   (int)newprog->progid,0,0,0);
    }
    else {
	/* malloc error, return an error */
	semGive(seqSem);
	taskSuspend (taskIdSelf ());
	return 0;
    }
    semGive(seqSem);
    return newprog->seq_token;
}

int resSeqDelete (progName, macroString)
char	*progName;
char	*macroString;
{
    RES_SEQPROG *curprog;	/* last program pointer */
    RES_SEQPROG *delprog;	/* program pointer to delete */

    semTake(seqSem, MAX_TIMEOUT );

    for (curprog = (RES_SEQPROG *)&seqprograms; curprog->next; curprog = curprog->next) {
	/* find program if registered */
	if ( (strncmp (progName, ((RES_SEQPROG *)curprog->next)->progid, strlen(progName)) == 0)
	  && (strcmp (macroString, &((RES_SEQPROG *)curprog->next)->progid[strlen(progName)]) == 0))
	{
	    /* Inform the operator */
 	    logMsg ( "Removing %s %s\n",
		   (int)progName,
                   (int)macroString,0,0,0,0);
	    /* remove from list */
	    delprog = curprog->next;
	    curprog->next = delprog->next;
	    if (delprog->seq_report)
		free (delprog->seq_report);
	    if (delprog->tx_report)
		free (delprog->tx_report);
	    free (delprog);
	    semGive(seqSem);
            return 1;
	}
    }
    semGive(seqSem);
    return 0;
}

int resSeqStatus (seq_token, activity)
unsigned short	seq_token;
int	*activity;
{
    RES_SEQPROG *curprog;	/* last program pointer */
    RES_TABLE *rp;		/* the resource entry */
    unsigned short answerState;	/* the state of the resource */

    semTake(seqSem, MAX_TIMEOUT );

    if (RESMGRDEBUG > 2)
	logMsg ("Checking status.\n",0,0,0,0,0,0);
    for (curprog = seqprograms; curprog; curprog = curprog->next) {
	/* find the program in the list */
	if (curprog->seq_token == seq_token) {
	    /* found him, check resource and state report */
	    if ( (resGetByName (curprog->resource, &rp) == RES_OK)
	      && (resGetCaAnswer (rp, &answerState) == RES_OK)) {
		if (answerState == RES_MSTR) {
		    if (curprog->linkstatus == NEWCONNECT) {
			*activity = SENDSTATEDATA;
		    }
                    else {
			*activity = NOACTION;
                    }
		}
                else {
		    if (curprog->seq_replen) {
		        *activity = READNEWSTATE;
		    }
		    else {
			*activity = NOACTION;
                    }
                }
	        semGive(seqSem);
                return (answerState);
	    }
	}
    }
    /* Error - not found, hold this guy in 'slave' with no action */
    *activity = NOACTION;
    semGive(seqSem);
    return (RES_SLV);

}

int resSeqSendState (unsigned short seq_token, char *stateBuf, unsigned short stateLen)
{
    RES_SEQPROG *curprog;	/* last program pointer */
    RES_TABLE *rp;		/* the resource entry */
    unsigned short answerState;	/* the state of the resource */
    unsigned short msgLen;	/* length of state report message */

    if (stateLen >= MAX_SEQRPT) {
	logMsg ("Requested state transfer length of %d exceeds system maximum of %d\n", stateLen, MAX_SEQRPT,0,0,0,0);
	return (RES_SLV); /* return failed */
    }

    semTake(seqSem, MAX_TIMEOUT );

    for (curprog = seqprograms; curprog; curprog = curprog->next) {
	/* find the program in the list */
	if (curprog->seq_token == seq_token) {
	    /* found him, check resource and state report */
	    if ( (resGetByName (curprog->resource, &rp) == RES_OK)
	      && (resGetCaAnswer (rp, &answerState) == RES_OK)) {
	        if ((rp->partnerFd)
	          && (curprog->seq_ptoken)) {
		    if (answerState == RES_MSTR) {
		    	msgLen = sizeof (RES_SEQRPTMSG) - MAX_SEQRPT + stateLen;
			if (curprog->tx_report &&
			  (curprog->tx_replen < msgLen)) {
			    free (curprog->tx_report);
			    curprog->tx_report = 0;
			}
			if (curprog->tx_report == 0)
			    curprog->tx_report = (RES_SEQRPTMSG *) malloc (msgLen);
			if (curprog->tx_report) {
			    semTake(curprog->stateSem, MAX_TIMEOUT );
			    curprog->tx_report->msgtype = SEQRPTMSG;
			    curprog->tx_report->msglen = msgLen;
			    curprog->tx_report->seq_token = curprog->seq_token;
			    curprog->tx_report->seq_ptoken = curprog->seq_ptoken;
			    curprog->tx_report->reportlen = stateLen;
			    memcpy (curprog->tx_report->report, stateBuf, stateLen);
			    curprog->tx_replen = msgLen;
			    curprog->need_txmit = 1;
			    semGive(curprog->stateSem);
			    if (RESMGRDEBUG)
				logMsg("ResMgr sending sequence state, length=%d, fd=%d\n", msgLen, rp->partnerFd,0,0,0,0);
			}
			curprog->linkstatus = CONNECT;
		    }
		}
		semGive(seqSem);
		return (answerState);
	    }
	}
    }
    /* Error - not found, ignore the operation */
    semGive(seqSem);
    return (RES_SLV);
}

int resSeqGetState (seq_token, stateBuf, stateLen)
unsigned short	seq_token;
char		*stateBuf;
unsigned short	stateLen;
{
    RES_SEQPROG *curprog;	/* last program pointer */
    RES_TABLE *rp;		/* the resource entry */
    unsigned short answerState;	/* the state of the resource */

    semTake(seqSem, MAX_TIMEOUT );

    for (curprog = seqprograms; curprog; curprog = curprog->next) {
	/* find the program in the list */
	if (curprog->seq_token == seq_token) {
	    /* found him, check resource and state report */
	    if ( (resGetByName (curprog->resource, &rp) == RES_OK)
	      && (resGetCaAnswer (rp, &answerState) == RES_OK)) {
		/* return report to user, clear report variables
		** User must free data. */
		semTake(curprog->stateSem, MAX_TIMEOUT );
		if (answerState != RES_MSTR) {
		    if (RESMGRDEBUG)
			logMsg("ResMgr get new state seq_token %d, length %d\n", seq_token, stateLen,0,0,0,0);
		    memcpy (stateBuf, curprog->seq_report, stateLen);
                }
	        curprog->seq_replen = 0;
		semGive(curprog->stateSem);
		semGive(seqSem);
		return (answerState);
	    }
	}
    }
    /* Error - not found, return slave */
    semGive(seqSem);
    return (RES_SLV);
}

int sendSeqRegistration (RES_SEQPROG *seqp, RES_TABLE *rp)
{
    unsigned short msgLen;	/* length of state report message */
    RES_SEQREGMSG *buffer;	/* allocated state report message */
    int		status;		/* outcome of TCP send op */
    int		retStatus = 0;	/* begin returning 0 */

    if (rp->partnerFd == 0)
	return retStatus; /* can't do anything without a link */

    semTake(seqp->stateSem, MAX_TIMEOUT );
    msgLen = sizeof (RES_SEQREGMSG) - MAX_PROGID + strlen(seqp->progid)+1;
    buffer = (RES_SEQREGMSG *) malloc (msgLen);
    if (buffer) {
	buffer->msgtype = SEQREGMSG;
	buffer->msglen = msgLen;
	buffer->seq_token = seqp->seq_token;
	buffer->seq_ptoken = seqp->seq_ptoken;
        memcpy (buffer->resource, seqp->resource, 16);
	strcpy (buffer->progid, seqp->progid);
	if (RESMGRDEBUG)
	    logMsg("ResMgr sending registration, length=%d, fd=%d\n", msgLen, rp->partnerFd,0,0,0,0);
	if ((status = send (rp->partnerFd, (char *)buffer, msgLen, 0)) != msgLen) {
    	    if (status <= 0) {
		/* Display TCP errors, but let main resMgr logic fix us up! */
		int	anerrno;
		anerrno = errnoGet();
		if((anerrno!=ECONNABORTED && anerrno!=ECONNRESET && anerrno!=ETIMEDOUT)) {
        	    logMsg("registration send, client disconnecting\n",0,0,0,0,0,0);
		}
	    }
	    else {
		/* short write - ??? */
		logMsg("short send attempt!!",0,0,0,0,0,0);
            }
	}
	else {
	    retStatus = 1; /*return success indication */
	    seqp->lastmsg = tickGet ();
	    regsent++;
	}
	free (buffer);
    }
    semGive(seqp->stateSem);
    return 1;
}

int rmConnDsp (void)
{
    int i;

    logMsg("connection display\n",0,0,0,0,0,0);
    for (i = 0; i < MAX_RESFDS; i++) {
        if (resFds[i].Inet.s_addr) {
	    logMsg("->IP: %s\tFD:%d\tLM:%ld\n",
		     (int)INET_NTOA(resFds[i].Inet),
		     resFds[i].fd, resFds[i].lastmsg,0,0,0);
        }
    }

    return (0);
}
int rmTableDump (void)
{
    int i;
    RES_SEQPROG *curprog;	/* program pointer */

    logMsg("connection display\n",0,0,0,0,0,0);
    for (i = 0; i < MAX_RESFDS; i++) {
        if (resFds[i].Inet.s_addr) {
	    logMsg("->IP: %s\tFD:%d\tLM:%ld\n",
		     (int)INET_NTOA(resFds[i].Inet),
		     resFds[i].fd, resFds[i].lastmsg,0,0,0);
        }
    }
    if (seqprograms == 0)
	return (0);
    logMsg("        Redundant sequence programs\n",0,0,0,0,0,0);
    semTake(seqSem, MAX_TIMEOUT );
    for (curprog = seqprograms; curprog; curprog = curprog->next) {
	logMsg ( "%4.4d/%4.4d LM:%ld %4.4s %-16.16s %s\n",
		   curprog->seq_token,
		   curprog->seq_ptoken,
		   curprog->lastmsg,
		   (int)((curprog->linkstatus==NOCONNECT)?"NOCN":((curprog->linkstatus==NEWCONNECT)?"NEWC":"CONN")),
		   (int)curprog->resource,
                   (int)curprog->progid);
    }
    semGive(seqSem);
    return (0);
}

int rmSetMsgTO (int time)
{
	msgtimeout = time;
	return 0;
}

int rmSetWakeupRate (int time)
{
	res_wakeuprate = time;
	return 0;
}

int rmSetMsgRate (int time)
{
	res_exchrate = time;
	return 0;
}

int rmSetMaxRegs (int cnt)
{
	maxregs = cnt;
	return 0;
}

int rmSetRegRate (int cnt)
{
	res_regrate = cnt;
	return 0;
}


int rmPrivate(char *resName)
{
    RES_TABLE	*rp;

    if (resGetByName (resName, &rp) == RES_OK) {
	resDeclarePrivate(rp);
	return (0);
    }
    else {
	logMsg ("could not find resource %s\n", (int)resName,0,0,0,0,0);
	return (-1);
    }
}

int rmMaster (char *resName)
{
    RES_TABLE    *rp;

    if (resGetByName (resName, &rp) == RES_OK) {
        rp->localState = RES_CHANGED|RES_MSTR;
        logMsg ("%s changed to Master\n", (int)resName,0,0,0,0,0);
	resSetCaAnswer (rp);
        return (0);
    }
    else {
        logMsg ("could not find resource %s\n", (int)resName,0,0,0,0,0);
	return(-1);
    }	
}

int rmSlave (char *resName)
{
    RES_TABLE    *rp;

    if (resGetByName (resName, &rp) == RES_OK) {
        rp->localState = RES_CHANGED|RES_SLV;
        logMsg ("%s changed to slave\n", (int)resName,0,0,0,0,0);
	resClearCaAnswer (rp);
        return (0);
    }
    else {
        logMsg ("could not find resource %s\n", (int)resName,0,0,0,0,0);
	return(-1);
    }	
}

int rmNewState (char *resName,int val)
{
    RES_TABLE    *rp;

    if (resGetByName (resName, &rp) == RES_OK) {
        rp->localState = val|RES_CHANGED;
        logMsg ("%s changed to %d\n", (int)resName, val,0,0,0,0);
	if (((val & ~RES_CHANGED) == RES_SLV)
	  || ((val & ~RES_CHANGED) == RES_SLVWANTCA)) {
	    resClearCaAnswer (rp);
	}
	else if ((val & ~RES_CHANGED) == RES_MSTR) {
	    resSetCaAnswer (rp);
	}
        return (0);
    }
    else {
        logMsg ("could not find resource %s\n", (int)resName,0,0,0,0,0);
	return(-1);
    }	
}

void rmShowCounts (void)
{
	logMsg ("ResMgr Counts: select calls: %d\nMsgs in: %d, out: %d\nRegister in: %d",
			selectcalls, msgrecv, msgsent, regrecv,0,0);
	logMsg ("ResMgr Counts(cont): out: %d\nState in: %d, out %d\n",
			regsent, rptrecv, rptsent,0,0,0);
}

void rmClearCounts (void)
{
	selectcalls = msgrecv = msgsent = regrecv = regsent = rptrecv = rptsent = 0;
}
