/****
*
*	File name :		l1_8250.c
*	Author :		Steve Johnson
*	Creation date :	December 1, 1985
*
*	Copyright (c) 1985,1999 Sigma Six Corporation
*
*	Provide the following entry points
*
*		pc8250_init
*		pc8250_connect
*		pc8250_publish
*		pc8250_accept
*		pc8250_write
*		pc8250_poll
*		pc8250_read
*		pc8250_clear
*		pc8250_shutdown
*		pc8250_exit
*
*	Physical I/O layer for DOS Intel 8250 type devices
*
****/

#include <sscommon.h>
#include <sstrace.h>
#include <ssnames.h>
#include <ssqueues.h>
#include <sslevel1.h>
#include <ssl18250.h>
#include <dos.h>
#include <conio.h>
#include <time.h>

#define LEVEL1_GROUP		"LEVEL1_8250"

/*******
*
*	8250 port assignments
*
*******/

#define I8259ICW1		0x20
#define I8259EOI		0x20
#define I8259OCW1		0x21
#define I8259OCW2		0x20
#define MaskIRQ3		0xF7
#define MaskIRQ4		0xEF
#define VectorIRQ3		0x0B
#define VectorIRQ4		0x0C
#define VectorCtrlBrk	0x1B
#define VectorCtrlC		0x23
#define	Com1Channel		0x3F0
#define Com2Channel		0x2F0

#define TXBuffer		0x8
#define	RXBuffer		0x8
#define	DLatchLSB		0x8
#define DLatchMSB		0x9
#define IntEnable		0x9
#define	IntID			0xA
#define LineControl		0xB
#define ModemControl	0xC
#define	LineStatus		0xD
#define ModemStatus		0xE


/*******
*
*	8250 mask bits and bytes for status and control masking
*
*******/

#define	Baud38400		0x03
#define	Baud19200		0x06
#define	Baud9600		0x0C
#define	Baud4800		0x18
#define	Baud2400		0x30
#define	Baud1800		0x40
#define	Baud1200		0x60

#define	D8Bits			0x03
#define	D7Bits			0x02
#define D6Bits			0x01
#define	D5Bits			0x00
#define	OneStopBit		0x00
#define	TwoStopBits		0x04
#define	NoParity		0x00
#define	OddParity		0x08
#define	EvenParity		0x18
#define MarkParity		0x28
#define SpaceParity		0x38
#define SetBreak		0x40
#define	AccessDivisor	0x80

#define	DisableInts		0x00
#define	EnableInts		0x01

#define	DataAvailable	0x01
#define DataErrors		0x1E
#define	DeltaCTS		0x01
#define	ClearToSend		0x10
#define	TxHoldEmpty		0x20
#define	TxShiftEmpty	0x40
#define ModemOUT2		0x08
#define ModemDTR		0x01
#define ModemInit		(ModemDTR + ModemOUT2)
#define ModemRTS		0x02
#define	TransmitClear	(TxHoldEmpty | TxShiftEmpty)

typedef struct {
	char		*comm_port_string;
	int			comm_port_8250;
} COMM_PORTS;
static COMM_PORTS port_table [] = {
	{ "COM1",	Com1Channel },
	{ "COM2",	Com2Channel }
};
#define NUM_PORTS	(sizeof (port_table) / sizeof (*port_table))

typedef struct {
	long			baud_rate_long;
	unsigned int	baud_rate_8250;
} BAUD_RATES;
static BAUD_RATES baud_rate_table [] = {
	{38400L,	Baud38400},
	{19200L,	Baud19200},
	{9600L,		Baud9600},
	{4800L,		Baud4800},
	{2400L,		Baud2400},
	{1800L,		Baud1800},
	{1200L,		Baud1200}
};
#define NUM_BAUDS	(sizeof (baud_rate_table) / sizeof (*baud_rate_table))

typedef struct {
	char		*parity_string;
	int		parity_8250;
} PARITIES;
static PARITIES parity_table [] = {
	{ "EVEN",	EvenParity },
	{ "ODD",	OddParity },
	{ "NONE",	NoParity },
	{ "SPACE",	SpaceParity },
	{ "MARK",	MarkParity }
};
#define NUM_PARITIES	(sizeof (parity_table) / sizeof (*parity_table))

typedef struct {
	int	data_bits_int;
	int	data_bits_8250;
} DATA_BITS;
static DATA_BITS bit_table [] = {
	{ 5,	D5Bits },
	{ 6,	D6Bits },
	{ 7,	D7Bits },
	{ 8,	D8Bits }
};
#define NUM_BITS	(sizeof (bit_table) / sizeof (*bit_table))

typedef struct {
	int	stop_bits_int;
	int	stop_bits_8250;
} STOP_BITS;
static STOP_BITS stop_table [] = {
	{ 1,	OneStopBit },
	{ 2,	TwoStopBits }
};
#define NUM_STOP_BITS	(sizeof (stop_table) / sizeof (*stop_table))

#define DEFAULT_WRITE_TIME_OUT	5
#define DEFAULT_FRAME_SIZE	2048

/****
*
*	Device context definitions
*
****/

typedef struct {
	QUEUE_DEF	link;
	int			old_value8259;
	size_t		frame_size;
	int			write_time_out;
	int			comm_port;
	char		*async_buffer;
	int			async_input;
	int			async_output;
	int			async_end;
	int			async_count;
	int			async_error;
	char		name [MAX_STRING];
} DEVICE_DEF;

static DEVICE_DEF	*int_device;
static QUEUE_DEF	root_device;

static void INTERRUPT	int_8250 (void);
static DEVICE_DEF		*add_device (char *);
static void				delete_device (DEVICE_DEF *);
static void				(INTERRUPT *old_irq_int) ();
static int				get_char (DEVICE_DEF *, UCHAR *);
static int				send_char (DEVICE_DEF *, UCHAR);
static void				get_configuration (char *, DEVICE_DEF *);
static int				get_comm_port (char *);
static int				get_baud_rate (char *);
static int				get_parity (char *);
static int				get_data_bits (char *);
static int				get_stop_bits (char *);

/****
*
****/

LEVEL1_STATUS pc8250_init (void)
{
	TRACE ("pc8250_init, entry point");
	queue_init (&root_device);
	return LEVEL1_NORMAL;
}

/****
*
****/

LEVEL1_STATUS pc8250_connect (LEVEL1_HANDLE *user_handle, char *name)
{
	int				baud_rate_div;
	int				parity;
	int				data_bits;
	int				stop_bits;
	DEVICE_DEF		*device;
	char			port_name [MAX_STRING];

	TRACE ("pc8250_connect, entry point");
	if (strlen (name) >= MAX_STRING) {
		ERROR_TRACE_S ("pc8250_connect, invalid device name = (%s)", name);
		return LEVEL1_BAD_NAME;
	}
	if (!ini_get_str (name, "COM_PORT", port_name)) {
		ERROR_TRACE_S ("pc8250_connect, no COM_PORT = (%s)", name);
		return LEVEL1_BAD_NAME;
	}
	device = add_device (name);
	if (!device) {
		ERROR_TRACE ("pc8250_connect, add_device failed");
		return LEVEL1_NO_MEMORY;
	}

	get_configuration (name, device);
	baud_rate_div = get_baud_rate (name);
	parity = get_parity (name);
	data_bits = get_data_bits (name);
	stop_bits = get_stop_bits (name);

	OUTPORT (device -> comm_port + IntEnable, DisableInts);
	OUTPORT (device -> comm_port + LineControl, AccessDivisor);
	OUTPORT (device -> comm_port + DLatchLSB, baud_rate_div);
	OUTPORT (device -> comm_port + DLatchMSB, 0);
	OUTPORT (device -> comm_port + LineControl
		   , parity + data_bits + stop_bits);
	INPORT (device -> comm_port + RXBuffer);
	INPORT (device -> comm_port + LineStatus);

	device -> async_count = 0;
	device -> async_input = 0;
	device -> async_output = 0;
	device -> async_end = device -> frame_size;
	device -> async_error = 0;
	TRACE_I ("pc8250_connect, malloc frame size = %d",device ->frame_size + 1);
	device -> async_buffer = (char *) malloc (device -> frame_size + 1);
	if (!device -> async_buffer) {
		ERROR_TRACE ("pc8250_connect, malloc failed");
		ERROR_TRACE_I ("pc8250_connect, errno = %d", errno);
		ERROR_TRACE_I ("pc8250_connect, size = %d", device -> frame_size + 1);
		delete_device (device);
		return LEVEL1_NO_MEMORY;
	}

	int_device = device;
	TRACE ("pc8250_connect, before disable interrupts");
	DISABLE ();
	device -> old_value8259 = INPORT (I8259OCW1) & 0xFF;
	if (device -> comm_port == Com1Channel) {
		TRACE ("pc8250_connect, com1 on IRQ4");
		old_irq_int = GET_VECTOR (VectorIRQ4);
		SET_VECTOR (VectorIRQ4, int_8250);
		OUTPORT (I8259OCW1, device -> old_value8259 & MaskIRQ4);
	} else {
		TRACE ("pc8250_connect, com2 on IRQ3");
		old_irq_int = GET_VECTOR (VectorIRQ3);
		SET_VECTOR (VectorIRQ3, int_8250);
		OUTPORT (I8259OCW1, device -> old_value8259 & MaskIRQ3);
	}
	OUTPORT (I8259OCW2, I8259EOI);
	INPORT (device -> comm_port + IntID);
	OUTPORT (device -> comm_port + IntEnable, EnableInts);
	ENABLE ();
	TRACE ("pc8250_connect, after enable interrupts");

	/****
	*
	*	set the modem control DTR status, should be tied
	*	to CTS on the receiver, to indicate we're ready
	****/
	OUTPORT (device -> comm_port + ModemControl, ModemInit + ModemRTS);

	*user_handle = (LEVEL1_HANDLE) device;
	TRACE ("pc8250_connect, successful");
	return LEVEL1_NORMAL;
}

/****
*
****/

LEVEL1_STATUS pc8250_publish (LEVEL1_HANDLE *user_handle, char *name)
{
	ERROR_TRACE_S ("pc8250_publish, not supported (%s)", name);
	return LEVEL1_NOT_IMPLEMENTED;
}

/****
*
****/

LEVEL1_STATUS pc8250_accept (LEVEL1_HANDLE	*client_handle
						   , LEVEL1_HANDLE	server_handle)
{
	ERROR_TRACE ("pc8250_accept, not supported");
	return LEVEL1_NOT_IMPLEMENTED;
}

/****
*
****/

LEVEL1_STATUS pc8250_write (LEVEL1_HANDLE	user_handle
						  , size_t			buffer_size
						  , void			*buffer)
{
	int			status;
	UCHAR		*char_buffer;
	DEVICE_DEF	*device;

	TRACE_B ("pc8250_write", buffer_size, buffer);
	if (!user_handle) {
		ERROR_TRACE ("pc8250_write, null handle");
		return LEVEL1_BAD_HANDLE;
	}
	device = (DEVICE_DEF *) user_handle;
	if (buffer_size > device -> frame_size) {
		ERROR_TRACE_I ("pc8250_write, buffer_size = %d", buffer_size);
		ERROR_TRACE_I ("pc8250_write, frame_size = %d", device -> frame_size);
		return LEVEL1_OVERFLOW;
	}
	char_buffer = (UCHAR *) buffer;
	printf ("\nOutbound: Sending %d bytes to com1...", buffer_size); fflush (stdout);
	while (buffer_size) {
		status = send_char (device, *char_buffer);
		if (LEVEL1_NORMAL != status) {
			ERROR_TRACE_I ("pc8250_write, status=%d", status);
			return status;
		}
		++char_buffer;
		--buffer_size;
	}
	printf ("finished.\n");
	return LEVEL1_NORMAL;
}

/****
*
****/

LEVEL1_STATUS pc8250_poll (LEVEL1_HANDLE user_handle, size_t *bytes_ready)
{
	DEVICE_DEF	*device;

	if (!user_handle) {
		ERROR_TRACE ("pc8250_poll, NULL handle");
		*bytes_ready = 0;
		return LEVEL1_BAD_HANDLE;
	}
	device = (DEVICE_DEF *) user_handle;
	if (device -> async_input == device -> async_output) {
		*bytes_ready = 0;
		return LEVEL1_NO_DATA;
	}
	*bytes_ready = device -> async_count;
	return LEVEL1_NORMAL;
}

/****
*
****/

LEVEL1_STATUS pc8250_read (LEVEL1_HANDLE	user_handle
						 , size_t			buffer_size
						 , void				*buffer
						 , size_t			*return_length)
{
	int			status;
	size_t		count;
	UCHAR		*char_buffer;
	DEVICE_DEF	*device;

	device = (DEVICE_DEF *) user_handle;
	if (!user_handle) {
		ERROR_TRACE ("pc8250_read, null handle");
		return LEVEL1_BAD_HANDLE;
	}
	count = 0;
	char_buffer = (UCHAR *) buffer;
	while (count < buffer_size) {
		status = get_char (device, char_buffer);
		if (LEVEL1_NO_DATA == status) {
			*return_length = count;
			if (count) {
				return LEVEL1_NORMAL;
			} else {
				TRACE ("pc8250_read, no data");
				return LEVEL1_NO_DATA;
			}
		} else if (LEVEL1_NORMAL != status) {
			TRACE_I ("pc8250_read, get_char = %d", status);
			return status;
		}
		++count;
		++char_buffer;
	}
	*return_length = count;
	return LEVEL1_NORMAL;
}

/****
*
****/

LEVEL1_STATUS pc8250_clear (LEVEL1_HANDLE user_handle)
{
	DEVICE_DEF	*device;

	TRACE ("pc8250_clear, entry point");
	if (!user_handle) {
		ERROR_TRACE ("pc8250_clear, NULL handle");
		return LEVEL1_BAD_HANDLE;
	}
	device = (DEVICE_DEF *) user_handle;
	DISABLE ();
	device -> async_count = 0;
	device -> async_input = 0;
	device -> async_output = 0;
	device -> async_end = device -> frame_size;
	device -> async_error = 0;
	ENABLE ();
	return LEVEL1_NORMAL;
}

/****
*
****/

LEVEL1_STATUS pc8250_shutdown (LEVEL1_HANDLE user_handle)
{
	DEVICE_DEF	*device;

	TRACE ("pc8250_shutdown, entry point");
	device = (DEVICE_DEF *) user_handle;
	DISABLE ();
	INPORT (device -> comm_port + IntID);
	OUTPORT (device -> comm_port + IntEnable, DisableInts);
	OUTPORT (I8259OCW1, device -> old_value8259);
	if (device -> comm_port == Com1Channel) {
		SET_VECTOR (VectorIRQ4, old_irq_int);
	} else {
		SET_VECTOR (VectorIRQ3, old_irq_int);
	}
	ENABLE ();
	delete_device (device);
	TRACE ("pc8250_shutdown, exit point");
	return LEVEL1_NORMAL;
}

/****
*
****/

LEVEL1_STATUS pc8250_exit (void)
{
	DEVICE_DEF		*device;

	TRACE ("pc8250_exit, entry point");
	while (queue_next (&root_device)) {
		TRACE ("pc8250_exit, shutdown next handle");
		device = (DEVICE_DEF *) queue_next (&root_device);
		pc8250_shutdown ((LEVEL1_HANDLE) device);
	}
	return LEVEL1_NORMAL;
}

/****
*
****/

static DEVICE_DEF *add_device (char *device_name)
{
	DEVICE_DEF	*device;

	device = (DEVICE_DEF *) malloc (sizeof (DEVICE_DEF));
	if (!device) {
		ERROR_TRACE_I ("add_device, malloc size = %d", sizeof (DEVICE_DEF));
		return NULL;
	}
	strcpy (device -> name, device_name);
	queue_append (&root_device, &device -> link);
	return device;
}

/****
*
****/

static void delete_device (DEVICE_DEF *device)
{
	TRACE ("delete_device, entry point");
	queue_remove (&root_device, &device -> link);
	free (device);
}

/*******
*
*		sends the given char to the 8250
*		returns false if the char can't go out
*
*******/

static int send_char (DEVICE_DEF *device, UCHAR chartosend)
{
	time_t		this_time;
	time_t		start_time;
	time_t		end_time;

	time (&start_time);
	end_time = start_time + device -> write_time_out;
	while ((INPORT (device -> comm_port + LineStatus) & TxHoldEmpty)
	!=     TxHoldEmpty) {
		time (&this_time);
		if (end_time < this_time) {
			TRACE_I ("send_char, timeout after %d", device -> write_time_out);
			return LEVEL1_TIMEOUT;
		}
	}
	OUTPORT (device -> comm_port + TXBuffer, chartosend & 0x0FF);

	return LEVEL1_NORMAL;
}

/*******
*	
*	reads the 8250 port status
*	does not wait for one if there isn't one available
*	returns the character in the char pointed to by chartoreturn
*
*******/

static int get_char (DEVICE_DEF *device, UCHAR *inchar)
{
	if (device -> async_error) {
		ERROR_TRACE_I ("get_char, async_error = 0x%02X", device -> async_error);
		device -> async_error = 0;
		return LEVEL1_NO_DATA;
	}

	if (device -> async_input != device -> async_output) {
		DISABLE ();
		*inchar = device -> async_buffer [device -> async_output];
		--device -> async_count;
		if (device -> async_output < device -> async_end) {
			++device -> async_output;
		} else {
			device -> async_output = 0;
		}
		ENABLE ();

		if (device -> async_error) {
			ERROR_TRACE_I ("get_char, async_error = 0x%02X",device->async_error);
			device -> async_error = 0;
			return LEVEL1_NO_DATA;
		} else {
			return LEVEL1_NORMAL;
		}
	} else {
		return LEVEL1_NO_DATA;
	}
}

/****
*
*	interrupt handler for 8250 through the 8259
*
****/

void INTERRUPT int_8250 (void)
{
	int	status;

	status = INPORT (int_device -> comm_port + LineStatus);
	if (status & DataAvailable) {
		if (status & 0x0E) {
		 	int_device -> async_error = status;
		}
		int_device -> async_buffer [int_device -> async_input] 
		= (char) (INPORT (int_device -> comm_port + RXBuffer) & 0x0FF);
		++int_device -> async_count;
		if (int_device -> async_input < int_device -> async_end) {
			++int_device -> async_input;
			if (int_device -> async_input == int_device -> async_output) {
			 	int_device -> async_error = -1;
			}
		} else {
			int_device -> async_input = 0;
		}
	}
	OUTPORT (I8259ICW1, I8259EOI);
}

static void get_configuration (char *name, DEVICE_DEF *device)
{
	TRACE ("get_configuration, entry point");
	device -> write_time_out = ini_get_int (name
										  , "WRITE_TIMEOUT"
										  , DEFAULT_WRITE_TIME_OUT);
	device -> frame_size = ini_get_int (name
									  , "FRAME_SIZE"
									  , DEFAULT_FRAME_SIZE);
	TRACE_I ("get_config, frame_size = %d", device -> frame_size);
	device -> comm_port = get_comm_port (name);
}

static int get_comm_port (char *name)
{
	int		count;
	char	comm_port_string [MAX_STRING];

	if (!ini_get_str (name, "COM_PORT", comm_port_string)) {
		strcpy (comm_port_string, "COM1");
	}
	for (count = 0; count < NUM_PORTS; ++count) {
		if (!strcmp (comm_port_string, port_table [count] .comm_port_string)) {
			TRACE_I ("get_comm_port, port = %x"
					, port_table [count] . comm_port_8250);
			return port_table [count] . comm_port_8250;
		}
	}
	ERROR_TRACE_S ("get_comm_port, invalid port = %s", comm_port_string);
	return Com1Channel;
}

static int get_baud_rate (char *name)
{
	int				count;
	long			default_baud_rate;
	long			local_baud_rate;

	default_baud_rate = ini_get_long (LEVEL1_GROUP, "BAUD_RATE", 4800L);
	local_baud_rate = ini_get_long (name, "BAUD_RATE", default_baud_rate);
	for (count = 0; count < NUM_BAUDS; ++count) {
		if (local_baud_rate == baud_rate_table [count] . baud_rate_long) {
			TRACE_I ("get_baud_rate, rate divisor = %x"
					, baud_rate_table [count] . baud_rate_8250);
			return baud_rate_table [count] . baud_rate_8250;
		}
	}
	ERROR_TRACE_L ("get_baud_rate, invalid rate = %d", local_baud_rate);
	return Baud9600;
}

static int get_parity (char *name)
{
	int		count;
	char	parity_string [MAX_STRING];

	if (!ini_get_str (name, "PARITY", parity_string)) {
		strcpy (parity_string, "NONE");
	}
	for (count = 0; count < NUM_PARITIES; ++count) {
		if (!strcmp (parity_string,parity_table[count].parity_string)){
			TRACE_I ("get_parity, parity = %x"
					, parity_table [count] . parity_8250);
			return parity_table [count] . parity_8250;
		}
	}
	ERROR_TRACE_S ("get_parity, invalid parity = %s", parity_string);
	return NoParity;
}

static int get_data_bits (char *name)
{
	int	count;
	int	data_bits_int;

	data_bits_int = ini_get_int (name, "DATA_BITS", 8);
	for (count = 0; count < NUM_BITS; ++count) {
		if (data_bits_int == bit_table [count] . data_bits_int) {
			TRACE_I ("get_data_bits, bits = %x"
					, bit_table [count] . data_bits_8250);
			return bit_table [count] . data_bits_8250;
		}
	}
	ERROR_TRACE_I ("get_data_bits, invalid dbits = %d", data_bits_int);
	return D8Bits;
}

static int get_stop_bits (char *name)
{
	int	count;
	int	stop_bits_int;

	stop_bits_int = ini_get_int (name, "STOP_BITS", 1);
	for (count = 0; count < NUM_STOP_BITS; ++count) {
		if (stop_bits_int == stop_table [count] . stop_bits_int) {
			TRACE_I ("get_stop_bits, bits = %x"
					, stop_table [count] . stop_bits_8250);
			return stop_table [count] . stop_bits_8250;
		}
	}
	ERROR_TRACE_I ("get_stop_bits, invalid sbits = %d", stop_bits_int);
	return OneStopBit;
}

