/****
*
*	File name :		l1_ssl.c
*	Author :		Steve Johnson
*	Creation date :	December 1, 1985
*
*	Copyright (c) 1985,1999 Sigma Six Corporation
*
*	Provide the following entry points
*
*		l1ssl_init
*		l1ssl_connect
*		l1ssl_publish
*		l1ssl_accept
*		l1ssl_write
*		l1ssl_poll
*		l1ssl_read
*		l1ssl_clear
*		l1ssl_shutdown
*		l1ssl_exit
*
*	Physical I/O layer for using ssl/IP to a telnet session
*
****/

#include <sscommon.h>
#include <sstrace.h>
#include <ssnames.h>
#include <ssqueues.h>
#include <sslevel1.h>
#include <ssl1ssl.h>
#include <sstcpip.h>
#include <sstcplib.h>
#include <ssaudit.h>
#include <sslshared.h>
#include <iocPriv.h>

char *last_ssl_name;

typedef enum {
	CLASS_PUBLISHED = 1,
	CLASS_ACCEPTED	= 2,
	CLASS_CONNECTED = 3,
} TCP_CLASS;

#define MAX_BUFFER_SIZE		8192
#define ESCAPE_CHAR			0xFF
#define LEVEL1_GROUP		"LEVEL1_SSL"

/****
*
*	Device context definitions
*
****/

typedef struct {
	QUEUE_DEF	link;
	int			socket_descr;
	int			ssl_descr;
	SSL			*ssl_handle;
	UCHAR		*read_buffer;
	size_t		read_index;
	int			escaped;
	size_t		read_buffer_size;
	int			shutdown_ready;
	TCP_CLASS	tcp_class;
	short		port;
	UCHAR		last_char;
	char		ssl_cn [MAX_STRING];
	char		ssl_dn [MAX_STRING];
	char		name [MAX_STRING];
	char		address [MAX_STRING];
} DEVICE_DEF;

static QUEUE_DEF	root_device;

static DEVICE_DEF	*add_device (char *, TCP_CLASS);
static void			delete_device (DEVICE_DEF *);
static int			read_socket (DEVICE_DEF *);
static int			set_options (int);
static void			close_ssl (DEVICE_DEF *);
static int			insert_data (DEVICE_DEF *, size_t, UCHAR *);
static void			escape_data (size_t, UCHAR *, UCHAR *, size_t *);

/****
*
****/

LEVEL1_STATUS l1ssl_init (void)
{
	TRACE ("l1ssl_init, entry point");
	queue_init (&root_device);
	last_ssl_name = NULL;
	return LEVEL1_NORMAL;
}

/****
*
****/

LEVEL1_STATUS l1ssl_connect (LEVEL1_HANDLE *user_handle, char *name)
{
	ERROR_TRACE ("l1ssl_connect, not implemented");
	return LEVEL1_NOT_IMPLEMENTED;
}

/****
*
****/

LEVEL1_STATUS l1ssl_publish (LEVEL1_HANDLE *user_handle, char *name)
{
	int				status;
	int				backlog;
	TCP_ADDR_IN		addr;
	short			port;
	DEVICE_DEF		*device;

	TRACE ("l1ssl_publish, entry point");
	TRACE_S ("l1ssl_publish, name = %s", name);
	device = add_device (name, CLASS_PUBLISHED);
	if (!device) {
		ERROR_TRACE ("l1ssl_publish, add_device failed");
		*user_handle = (LEVEL1_HANDLE) NULL;
		return LEVEL1_NO_MEMORY;
	}
	port = (short) ini_get_int (name, "TCP_PORT", 0);
	if (port < 1024) {
		ERROR_TRACE_S ("l1ssl_publish, no port for name = %s", name);
		delete_device (device);
		*user_handle = (LEVEL1_HANDLE) NULL;
		return LEVEL1_BAD_PORT;
	}
	device -> socket_descr = socket (AF_INET, SOCK_STREAM, 0);
	if (device -> socket_descr == -1) {
		ERROR_TRACE_I ("l1ssl_publish, socket errno = %d", GetLastError ());
		delete_device (device);
		*user_handle = (LEVEL1_HANDLE) NULL;
		return LEVEL1_NO_MEMORY;
	}
	if (!set_options (device -> socket_descr)) {
		ERROR_TRACE ("l1ssl_publish, set_options failed");
		shutdown (device -> socket_descr, 2);
		SOCKET_CLOSE (device -> socket_descr);
		delete_device (device);
		*user_handle = (LEVEL1_HANDLE) NULL;
		return LEVEL1_NO_MEMORY;
	}
	addr . sin_family = AF_INET;
	addr . sin_port = htons (port);
	FILL_ADDRESS (addr . sin_addr, INADDR_ANY);
	memset (addr . sin_zero, '\0', sizeof (addr . sin_zero));
	status = bind (device -> socket_descr
				 , (TCP_ADDR *) &addr
				 , sizeof (addr));
	if (status == -1) {
		ERROR_TRACE_I ("l1ssl_publish, bind errno = %d", errno);
		shutdown (device -> socket_descr, 2);
		SOCKET_CLOSE (device -> socket_descr);
		delete_device (device);
		*user_handle = (LEVEL1_HANDLE) NULL;
		return LEVEL1_IO_ERROR;
	}

	backlog = ini_get_int (name, "TCP_BACKLOG", 2);
	status = listen (device -> socket_descr, backlog);
	if (status == -1) {
		ERROR_TRACE_I ("l1ssl_publish, listen errno = %d", errno);
		shutdown (device -> socket_descr, 2);
		SOCKET_CLOSE (device -> socket_descr);
		delete_device (device);
		*user_handle = (LEVEL1_HANDLE) NULL;
		return LEVEL1_IO_ERROR;
	}
	if (!localtcp_add (device -> socket_descr)) {
		ERROR_TRACE ("l1ssl_publish, localtcp_add failed");
		shutdown (device -> socket_descr, 2);
		SOCKET_CLOSE (device -> socket_descr);
		delete_device (device);
		*user_handle = (LEVEL1_HANDLE) NULL;
		return LEVEL1_NO_MEMORY;
	}
	device -> ssl_handle = (SSL *) NULL;
	*user_handle = (LEVEL1_HANDLE) device;
	TRACE ("l1ssl_publish, successful");
	return LEVEL1_NORMAL;
}

/****
*
****/

LEVEL1_STATUS l1ssl_accept (LEVEL1_HANDLE	*client_handle
						  , LEVEL1_HANDLE	server_handle)
{
	int				status;
	long			blocking;
	DEVICE_DEF		*client;
	DEVICE_DEF		*server;
	char			*host_name;

	TRACE ("l1ssl_accept, entry point");
	if (!server_handle) {
		ERROR_TRACE ("tcpip_write, NULL handle");
		*client_handle = (LEVEL1_HANDLE) NULL;
		return LEVEL1_BAD_HANDLE;
	}
	server = (DEVICE_DEF *) server_handle;
	client = add_device (server -> name, CLASS_ACCEPTED);
	if (!client) {
		ERROR_TRACE ("l1ssl_accept, add_device failed");
		*client_handle = (LEVEL1_HANDLE) NULL;
		return LEVEL1_NO_MEMORY;
	}

    status = do_accept (server -> socket_descr
					  , &client -> socket_descr
					  , &host_name);
    if (!status) {
		ERROR_TRACE_I ("l1ssl_accept, do_accept errno = %d", GetLastError ());
		delete_device (client);
		*client_handle = (LEVEL1_HANDLE) NULL;
		return LEVEL1_SHUTDOWN;
    }
	if (-1 == client -> socket_descr) {
		ERROR_TRACE_I ("l1ssl_accept, do_accept -1 errno = %d", GetLastError ());
		shutdown (client -> socket_descr, 2);
		SOCKET_CLOSE (client -> socket_descr);
		delete_device (client);
		*client_handle = (LEVEL1_HANDLE) NULL;
		return LEVEL1_SHUTDOWN;
	}
	TRACE_I ("l1ssl_accept, socket_descr = %d", client -> socket_descr);
    client -> ssl_handle = setup_ssl_accept (client -> socket_descr);
	if (!client -> ssl_handle) {
		ERROR_TRACE_I ("l1ssl_accept, setup_ssl_accept = %d", GetLastError ());
		shutdown (client -> socket_descr, 2);
		SOCKET_CLOSE (client -> socket_descr);
		delete_device (client);
		*client_handle = (LEVEL1_HANDLE) NULL;
		return LEVEL1_SHUTDOWN;
	}
    status = accept_ssl_connection_timed (client -> ssl_handle, 20);
	if (ERROR == status) {
		ERROR_TRACE_I ("l1ssl_accept, accept_ssl_connection = %d", GetLastError ());
		close_ssl (client);
		delete_device (client);
		*client_handle = (LEVEL1_HANDLE) NULL;
		return LEVEL1_SHUTDOWN;
	}
    GetPeerCN (client -> ssl_handle,client -> ssl_cn, sizeof (client -> ssl_cn));
    if (GetGeneralPrivHost (client -> ssl_cn, "MT_") < 1) {
		strcpy (client -> ssl_dn, "No ssl_dn");
	    GetPeerDN (client -> ssl_handle, client -> ssl_dn, sizeof (client -> ssl_dn));
		last_ssl_name = client -> ssl_dn;
		ERROR_TRACE_I ("l1ssl_accept, GetGeneralPrivHost = %d", GetLastError ());
		SecurityAlarm (client -> ssl_cn, "GetGeneralPrivHost failed in Quad4", "");
		audit_fail ("GetGeneralPrivHost failed");
		last_ssl_name = NULL;
		close_ssl (client);
		delete_device (client);
		*client_handle = (LEVEL1_HANDLE) NULL;
		return LEVEL1_SHUTDOWN;
	}
    GetPeerDN (client -> ssl_handle, client -> ssl_dn, sizeof (client -> ssl_dn));
	client -> ssl_descr = SSL_get_fd (client -> ssl_handle);
	blocking = 1;
	NON_BLOCKING (client -> socket_descr, blocking);
	TRACE_I ("l1ssl_accept, ssl_descr = %d", client -> ssl_descr);
	if (!set_options (client -> ssl_descr)) {
		ERROR_TRACE ("l1ssl_accept, set_options failed");
		close_ssl (client);
		delete_device (client);
		*client_handle = (LEVEL1_HANDLE) NULL;
		return LEVEL1_NO_MEMORY;
	}
	if (!localtcp_add (client -> ssl_descr)) {
		ERROR_TRACE ("l1ssl_accept, localtcp_add failed");
		close_ssl (client);
		delete_device (client);
		*client_handle = (LEVEL1_HANDLE) NULL;
		return LEVEL1_NO_MEMORY;
	}
	last_ssl_name = client -> ssl_dn;
	audit_logon ("Quad4 Call Accepted");
	*client_handle = (LEVEL1_HANDLE) client;
	TRACE ("l1ssl_accept, connection accepted");
	return LEVEL1_NORMAL;
}

/****
*
****/

LEVEL1_STATUS l1ssl_write (LEVEL1_HANDLE		user_handle
						 , size_t				buffer_size
						 , void 				*buffer)
{
	size_t		bytes_written;
	size_t		send_size;
	DEVICE_DEF	*device;
	UCHAR		send_buffer [MAX_BUFFER_SIZE];

	TRACE_I ("l1ssl_write, buffer_size = %d", buffer_size);
	if (!user_handle) {
		ERROR_TRACE ("l1ssl_write, NULL handle");
		return LEVEL1_BAD_HANDLE;
	}
	device = (DEVICE_DEF *) user_handle;
	TRACE_S ("l1ssl_write, device = %s", device -> name);
	if (device -> escaped) {
		escape_data (buffer_size, buffer, send_buffer, &send_size);
	} else {
		memcpy (send_buffer, buffer, buffer_size);
		send_size = buffer_size;
	}
	TRACE_I ("l1ssl_write, send_size = %d", send_size);
	bytes_written = SSL_write (device -> ssl_handle, send_buffer, send_size);
	if (bytes_written != send_size) {
		ERROR_TRACE_I ("l1ssl_write, send errno = %d", errno);
		ERROR_TRACE_I ("l1ssl_write, buffer_size = %d", buffer_size);
		ERROR_TRACE_I ("l1ssl_write, bytes_written = %d", bytes_written);
		return LEVEL1_SHUTDOWN;
	}
	TRACE ("l1ssl_write, returns normal");
	return LEVEL1_NORMAL;
}

/****
*
****/

LEVEL1_STATUS l1ssl_poll (LEVEL1_HANDLE user_handle, size_t *bytes_ready)
{
	int			count;
	int			status;
	int			socket_descr;
	DEVICE_DEF	*device;

	if (!user_handle) {
		ERROR_TRACE ("l1ssl_poll, NULL handle");
		*bytes_ready = 0;
		return LEVEL1_BAD_HANDLE;
	}
	device = (DEVICE_DEF *) user_handle;
	if (device -> shutdown_ready) {
		TRACE ("l1ssl_poll, device needs to be shutdown");
		*bytes_ready = 0;
		return LEVEL1_NORMAL;
	}

	if (device -> ssl_handle) {
		socket_descr = device -> ssl_descr;
	} else {
		socket_descr = device -> socket_descr;
	}
	count = localtcp_poll (socket_descr);
	if (count == -1) {
		ERROR_TRACE_I ("l1ssl_poll, localtcp_poll = %d", GetLastError ());
		*bytes_ready = 0;
		return LEVEL1_NO_DATA;
	} else if (count == 0) {
		*bytes_ready = 0;
		return LEVEL1_NO_DATA;
	}
	TRACE_I ("l1ssl_poll, select count = %d", count);
	TRACE_I ("l1ssl_poll, socket = %d", socket_descr);
	if (!device -> ssl_handle) {
		TRACE ("l1ssl_poll, incoming call on published handle");
		*bytes_ready = 0;
		return LEVEL1_NORMAL;
	}
	status = read_socket (device);
	if (LEVEL1_NO_DATA == status) {
		*bytes_ready = 0;
		return LEVEL1_NO_DATA;
	}
	TRACE_I ("l1ssl_poll, read_socket = %d", status);
	*bytes_ready = device -> read_index;
	return LEVEL1_NORMAL;
}

/****
*
****/

LEVEL1_STATUS l1ssl_read (LEVEL1_HANDLE	user_handle
						, size_t		buffer_size
						, void			*buffer
						, size_t		*return_length)
{
	DEVICE_DEF	*device;

	TRACE ("l1ssl_read, entry point");
	if (!user_handle) {
		ERROR_TRACE ("l1ssl_read, NULL handle");
		*return_length = 0;
		return LEVEL1_BAD_HANDLE;
	}
	device = (DEVICE_DEF *) user_handle;
	if (device -> shutdown_ready) {
		TRACE ("l1ssl_read, device is shutdown");
		*return_length = 0;
		return LEVEL1_SHUTDOWN;
	}
	if (!device -> read_index) {
		TRACE ("l1ssl_read, no data");
		*return_length = 0;
		return LEVEL1_NO_DATA;
	}
	if (device -> read_index > buffer_size) {
		*return_length = buffer_size;
	} else {
		*return_length = device -> read_index;
	}
	last_ssl_name = device -> ssl_dn;
	memcpy (buffer, device -> read_buffer, *return_length);
	device -> read_index = 0;
	TRACE_B ("l1ssl_read", *return_length, buffer);
	return LEVEL1_NORMAL;
}

/****
*
****/

LEVEL1_STATUS l1ssl_clear (LEVEL1_HANDLE user_handle)
{
	DEVICE_DEF	*device;

	TRACE ("l1ssl_clear, entry point");
	if (!user_handle) {
		ERROR_TRACE ("l1ssl_clear, NULL handle");
		return LEVEL1_BAD_HANDLE;
	}
	device = (DEVICE_DEF *) user_handle;
	return LEVEL1_NOT_IMPLEMENTED;
}

/****
*
****/

LEVEL1_STATUS l1ssl_shutdown (LEVEL1_HANDLE user_handle)
{
	DEVICE_DEF	*device;

	TRACE ("l1ssl_shutdown, entry point");
	if (!user_handle) {
		ERROR_TRACE ("l1ssl_shutdown, NULL handle");
		return LEVEL1_BAD_HANDLE;
	}
	device = (DEVICE_DEF *) user_handle;
	localtcp_remove (device -> socket_descr);
	switch (device -> tcp_class) {
	case CLASS_CONNECTED:
	case CLASS_ACCEPTED:
		last_ssl_name = device -> ssl_dn;
		audit_logoff ("Quad4 Shutdown MV90");
		last_ssl_name = NULL;
		close_ssl (device);
		break;
	case CLASS_PUBLISHED:
		shutdown (device -> socket_descr, 2);
		SOCKET_CLOSE (device -> socket_descr);
		break;
	}
	delete_device (device);
	return LEVEL1_NORMAL;
}

/****
*
****/

LEVEL1_STATUS l1ssl_exit (void)
{
	DEVICE_DEF		*device;

	TRACE ("l1ssl_exit, entry point");
	while (queue_next (&root_device)) {
		TRACE ("l1ssl_exit, shutdown next handle");
		device = (DEVICE_DEF *) queue_next (&root_device);
		l1ssl_shutdown ((LEVEL1_HANDLE) device);
	}
	return LEVEL1_NORMAL;
}

/****
*
****/

static DEVICE_DEF *add_device (char *device_name, TCP_CLASS tcp_class)
{
	DEVICE_DEF	*device;
	size_t		read_buffer_size;

	device = (DEVICE_DEF *) malloc (sizeof (DEVICE_DEF));
	if (!device) {
		ERROR_TRACE_I ("add_device, malloc #1 = %d", sizeof (DEVICE_DEF));
		return (DEVICE_DEF *) NULL;
	}
	switch (tcp_class) {
	case CLASS_CONNECTED:
	case CLASS_ACCEPTED:
		read_buffer_size = ini_get_int (LEVEL1_GROUP, "MAX_FRAME", MAX_BUFFER_SIZE);
		device -> read_buffer_size = ini_get_int (device_name
												, "MAX_FRAME"
												, read_buffer_size);
		device -> read_buffer = (UCHAR *) malloc (device -> read_buffer_size);
		if (!device -> read_buffer) {
			ERROR_TRACE_I ("add_device, malloc #2 = %d"
						  , device -> read_buffer_size);
			delete_device (device);
			return (DEVICE_DEF *) NULL;
		}
		break;
	default:
		device -> read_buffer_size = 0;
		device -> read_buffer = (UCHAR *) NULL;
	}
	device -> escaped = ini_get_int (device_name, "ESCAPED", FALSE);
	device -> read_index = 0;
	device -> shutdown_ready = FALSE;
	device -> tcp_class = tcp_class;
	strcpy (device -> name, device_name);
	queue_append (&root_device, &device -> link);
	return device;
}

/****
*
****/

static void delete_device (DEVICE_DEF *device)
{
	TRACE ("delete_device, entry point");
	if (device -> read_buffer) {
		free (device -> read_buffer);
	}
	queue_remove (&root_device, &device -> link);
	free (device);
}

static void close_ssl (DEVICE_DEF *device)
{
	sslSharedCloseConn (device -> ssl_handle); 
}

/****
*
****/

static int set_options (int socket_descr)
{
	int		status;
	int		option_value;

	TRACE ("set_options, entry point");
	option_value = 1;
	status = SET_OPTION (socket_descr
					   , SOL_SOCKET
					   , SO_REUSEADDR
					   , (char *) &option_value
					   , sizeof (option_value));
	if (status) {
		ERROR_TRACE_I ("set_options, ioctl = %d", errno);
		ERROR_TRACE_I ("set_options, set_option #1 = %d", status);
		return FALSE;
	}
	option_value = 1;
	status = SET_OPTION (socket_descr
					   , SOL_SOCKET
					   , SO_KEEPALIVE
					   , (char *) &option_value
					   , sizeof (option_value));
	if (status) {
		ERROR_TRACE_I ("set_options, ioctl = %d", errno);
		ERROR_TRACE_I ("set_options, set_option #2 = %d", status);
		return FALSE;
	}
	return TRUE;
}

/****
*
****/

static int read_socket (DEVICE_DEF *device)
{
	int			count;
	int			status;
	int			max_read_size;
	UCHAR		read_buffer [MAX_BUFFER_SIZE];

	max_read_size = MIN (MAX_BUFFER_SIZE
					   , device -> read_buffer_size - device -> read_index);
	if (!max_read_size) {
		ERROR_TRACE_I ("read_socket, read_buffer_size = %d"
					, device -> read_buffer_size);
		ERROR_TRACE_I ("read_socket, read_index = %d"
					 , device -> read_index);
		device -> shutdown_ready = TRUE;
		return LEVEL1_SHUTDOWN;
	}
	count = SSL_read (device -> ssl_handle, read_buffer, max_read_size);
	if (count <= 0) {
		status = SSL_get_error (device -> ssl_handle, count);
		switch (status) {
	    case SSL_ERROR_WANT_WRITE:
	    case SSL_ERROR_NONE:
	    case SSL_ERROR_WANT_READ:
	    case SSL_ERROR_WANT_X509_LOOKUP:
			/* benign, just want to write/read for state machine */
			TRACE ("read_socket, SSL_read returns zero bytes");
			return LEVEL1_NO_DATA;
	    case SSL_ERROR_SSL:
			if (errno == EWOULDBLOCK) {
			    break;
			}
	    case SSL_ERROR_SYSCALL:
	    case SSL_ERROR_ZERO_RETURN:
	    case SSL_ERROR_WANT_CONNECT:
	    default:
			TRACE_I ("read_socket, disconnect error %d", errno);
		}
		device -> shutdown_ready = TRUE;
		return LEVEL1_SHUTDOWN;
	}
	if (device -> escaped) {
		count = insert_data (device, count, read_buffer);
		if (!count) {
			TRACE ("read_socket, all new data was discarded");
			return LEVEL1_NO_DATA;
		}
	} else {
		memcpy (&device -> read_buffer [device -> read_index]
			  , read_buffer
			  , count);
		device -> read_index += count;
	}
	TRACE_I ("read_socket, read_index = %d", device -> read_index);
	return LEVEL1_NORMAL;
}

static int insert_data (DEVICE_DEF *device, size_t buffer_size, UCHAR *buffer)
{
	size_t		out_count;
	UCHAR		*in_buffer;
	UCHAR		*out_buffer;

	out_count = 0;
	in_buffer = buffer;
	out_buffer = &device -> read_buffer [device -> read_index];
	while (buffer_size) {
		--buffer_size;
		if (ESCAPE_CHAR == device -> last_char) {
			if (ESCAPE_CHAR == *in_buffer) {
				++in_buffer;
				device -> last_char = 0;
			} else if ((0xFB == *in_buffer)
			||         (0xFD == *in_buffer)) {
				++in_buffer;
				if (out_count) {
					--out_count;
					--out_buffer;
				} else {
					if (device -> read_index) {
						--device -> read_index;
						--out_buffer;
					}
				}
				device -> last_char = ESCAPE_CHAR;
			} else {
				ERROR_TRACE ("this is where to put the telnet stuff");
				device -> last_char = *in_buffer;
				if (device -> read_index) {
					--device -> read_index;
					--out_buffer;
				}
				if (out_count) {
					--out_count;
				}
				++in_buffer;
			}
		} else if (0x0D == device -> last_char) {
			if (0x00 == *in_buffer) {
				++in_buffer;
				device -> last_char = 0;
			} else {
				*out_buffer = *in_buffer;
				device -> last_char = *in_buffer;
				++out_count;
				++out_buffer;
				++in_buffer;
			}
		} else {
			*out_buffer = *in_buffer;
			device -> last_char = *in_buffer;
			++out_count;
			++out_buffer;
			++in_buffer;
		}
	}
	device -> read_index += out_count;
	return out_count;
}

static void escape_data (size_t		in_buffer_size
					   , UCHAR		*in_buffer
					   , UCHAR		*out_buffer
					   , size_t		*return_size)
{
	int		count;

	TRACE_B ("escape_data before", in_buffer_size, in_buffer);
	count = 0;
	while (in_buffer_size) {
		if (ESCAPE_CHAR == *in_buffer) {
			*out_buffer = ESCAPE_CHAR;
			++out_buffer;
			++count;
		}
		*out_buffer = *in_buffer;
		if (0x0D == *in_buffer) {
			++out_buffer;
			++count;
			*out_buffer = 0;
		}
		++out_buffer;
		++count;
		++in_buffer;
		--in_buffer_size;
	}
	*return_size = count;
}

