/****
*
*	File name :		l1_tcp.c
*	Author :		Steve Johnson
*	Creation date :	December 1, 1985
*
*	Copyright (c) 1985,1999 Sigma Six Corporation
*
*	Provide the following entry points
*
*		l1tcp_init
*		l1tcp_connect
*		l1tcp_publish
*		l1tcp_accept
*		l1tcp_write
*		l1tcp_poll
*		l1tcp_read
*		l1tcp_clear
*		l1tcp_shutdown
*		l1tcp_exit
*
*	Physical I/O layer for using TCP/IP to a telnet session
*
****/

#include <sscommon.h>
#include <sstrace.h>
#include <ssnames.h>
#include <ssqueues.h>
#include <sslevel1.h>
#include <ssl1tcp.h>
#include <sstcpip.h>
#include <sstcplib.h>

#define LEVEL1_GROUP		"LEVEL1_TCP"

typedef enum {
	CLASS_PUBLISHED = 1,
	CLASS_ACCEPTED	= 2,
	CLASS_CONNECTED = 3,
} TCP_CLASS;

#define MAX_BUFFER_SIZE		8192
#define ESCAPE_CHAR			0xFF

/****
*
*	Device context definitions
*
****/

typedef struct {
	QUEUE_DEF	link;
	int			socket_descr;
	UCHAR		*read_buffer;
	size_t		read_buffer_size;
	size_t		read_index;
	int			escaped;
	int			shutdown_ready;
	TCP_CLASS	tcp_class;
	short		port;
	UCHAR		last_char;
	long		connect_timeout;
	char		name [MAX_STRING];
	char		address [MAX_STRING];
} DEVICE_DEF;

static QUEUE_DEF	root_device;

static DEVICE_DEF	*add_device (char *, TCP_CLASS);
static void			delete_device (DEVICE_DEF *);
static int			read_socket (DEVICE_DEF *);
static int			set_options (int);
static int			insert_data (DEVICE_DEF *, size_t, UCHAR *);
static void			escape_data (size_t, UCHAR *, UCHAR *, size_t *);

/****
*
****/

LEVEL1_STATUS l1tcp_init (void)
{
	TRACE ("l1tcp_init, entry point");
	queue_init (&root_device);
	return LEVEL1_NORMAL;
}

/****
*
****/

LEVEL1_STATUS l1tcp_connect (LEVEL1_HANDLE *user_handle, char *name)
{
	short		port;
	int			status;
	char		address [MAX_STRING];
	long		connect_timeout;
	DEVICE_DEF	*device;

	TRACE ("l1tcp_connect, entry point");
	if (strlen (name) >= MAX_STRING) {
		ERROR_TRACE_S ("l1tcp_connect, invalid device name = (%s)", name);
		return LEVEL1_BAD_NAME;
	}
	TRACE_S ("l1tcp_connect, port_name = %s", name);
	device = add_device (name, CLASS_CONNECTED);
	if (!device) {
		ERROR_TRACE ("l1tcp_connect, add_device failed");
		return LEVEL1_NO_MEMORY;
	}
	connect_timeout = ini_get_long (LEVEL1_GROUP, "CONNECT_TIMEOUT", 1000L);
	device -> connect_timeout = ini_get_long (name
										    , "CONNECT_TIMEOUT"
										    , connect_timeout);
	if (!ini_get_str (name, "TCP_ADDRESS", address)) {
		ERROR_TRACE_S ("l1tcp_connect, no TCP_ADDRESS in (%s)", name);
		delete_device (device);
		*user_handle = (LEVEL1_HANDLE) NULL;
		return LEVEL1_BAD_ADDRESS;
	}
	port = (short) ini_get_int (name, "TCP_PORT", 0);
	if (port < 1024) {
		ERROR_TRACE_S ("l1tcp_connect, no TCP_PORT in (%s)", name);
		delete_device (device);
		*user_handle = (LEVEL1_HANDLE) NULL;
		return LEVEL1_BAD_PORT;
	}
	strcpy (device -> address, address);
	device -> port = port;
	TRACE_S ("l1tcp_connect, address = (%s)", address);
	TRACE_I ("l1tcp_connect, port = %d", port);
	device -> socket_descr = socket (AF_INET, SOCK_STREAM, 0);
	if (device -> socket_descr == -1) {
		ERROR_TRACE_SYS ("l1tcp_connect, socket (%ld) = %s");
		delete_device (device);
		*user_handle = (LEVEL1_HANDLE) NULL;
		return LEVEL1_NO_MEMORY;
	}
	status = localtcp_connect (device -> socket_descr
							 , address
							 , port
							 , device -> connect_timeout);
	if (TCPIP_NORMAL != status) {
		TRACE ("l1tcp_connect, connect failed");
		shutdown (device -> socket_descr, 2);
		SOCKET_CLOSE (device -> socket_descr);
		delete_device (device);
		*user_handle = (LEVEL1_HANDLE) NULL;
		return LEVEL1_BAD_NAME;
	}
	if (!set_options (device -> socket_descr)) {
		ERROR_TRACE ("l1tcp_connect, set_options failed");
		shutdown (device -> socket_descr, 2);
		SOCKET_CLOSE (device -> socket_descr);
		delete_device (device);
		*user_handle = (LEVEL1_HANDLE) NULL;
		return LEVEL1_NO_MEMORY;
	}
	if (!localtcp_add (device -> socket_descr)) {
		ERROR_TRACE ("l1tcp_connect, localtcp_add failed");
		shutdown (device -> socket_descr, 2);
		SOCKET_CLOSE (device -> socket_descr);
		delete_device (device);
		*user_handle = (LEVEL1_HANDLE) NULL;
		return LEVEL1_NO_MEMORY;
	}
	*user_handle = (LEVEL1_HANDLE) device;
	TRACE_I ("l1tcp_connect, success descr = %d", device -> socket_descr);
	return LEVEL1_NORMAL;
}

/****
*
****/

LEVEL1_STATUS l1tcp_publish (LEVEL1_HANDLE *user_handle, char *name)
{
	int				status;
	int				backlog;
	TCP_ADDR_IN		addr;
	short			port;
	DEVICE_DEF		*device;

	TRACE ("l1tcp_publish, entry point");
	TRACE_S ("l1tcp_publish, name = %s", name);
	device = add_device (name, CLASS_PUBLISHED);
	if (!device) {
		ERROR_TRACE ("l1tcp_publish, add_device failed");
		*user_handle = (LEVEL1_HANDLE) NULL;
		return LEVEL1_NO_MEMORY;
	}
	port = (short) ini_get_int (name, "TCP_PORT", 0);
	if (port < 1024) {
		ERROR_TRACE_S ("l1tcp_publish, no port for name = %s", name);
		delete_device (device);
		*user_handle = (LEVEL1_HANDLE) NULL;
		return LEVEL1_BAD_PORT;
	}
	device -> socket_descr = socket (AF_INET, SOCK_STREAM, 0);
	if (device -> socket_descr == -1) {
		ERROR_TRACE_SYS ("l1tcp_publish, socket (%ld) = %s");
		delete_device (device);
		*user_handle = (LEVEL1_HANDLE) NULL;
		return LEVEL1_NO_MEMORY;
	}
	if (!set_options (device -> socket_descr)) {
		ERROR_TRACE ("l1tcp_publish, set_options failed");
		shutdown (device -> socket_descr, 2);
		SOCKET_CLOSE (device -> socket_descr);
		delete_device (device);
		*user_handle = (LEVEL1_HANDLE) NULL;
		return LEVEL1_NO_MEMORY;
	}
	addr . sin_family = AF_INET;
	addr . sin_port = htons (port);
	FILL_ADDRESS (addr . sin_addr, INADDR_ANY);
	memset (addr . sin_zero, '\0', sizeof (addr . sin_zero));
	status = bind (device -> socket_descr
				 , (TCP_ADDR *) &addr
				 , sizeof (addr));
	if (status == -1) {
		ERROR_TRACE_SYS ("l1tcp_publish, bind (%ld) = %s");
		shutdown (device -> socket_descr, 2);
		SOCKET_CLOSE (device -> socket_descr);
		delete_device (device);
		*user_handle = (LEVEL1_HANDLE) NULL;
		return LEVEL1_IO_ERROR;
	}

	backlog = ini_get_int (name, "TCP_BACKLOG", 2);
	status = listen (device -> socket_descr, backlog);
	if (status == -1) {
		ERROR_TRACE_SYS ("l1tcp_publish, listen (%ld) = %s");
		shutdown (device -> socket_descr, 2);
		SOCKET_CLOSE (device -> socket_descr);
		delete_device (device);
		*user_handle = (LEVEL1_HANDLE) NULL;
		return LEVEL1_IO_ERROR;
	}
	if (!localtcp_add (device -> socket_descr)) {
		ERROR_TRACE ("l1tcp_publish, localtcp_add failed");
		shutdown (device -> socket_descr, 2);
		SOCKET_CLOSE (device -> socket_descr);
		delete_device (device);
		*user_handle = (LEVEL1_HANDLE) NULL;
		return LEVEL1_NO_MEMORY;
	}
	*user_handle = (LEVEL1_HANDLE) device;
	TRACE ("l1tcp_publish, successful");
	return LEVEL1_NORMAL;
}

/****
*
****/

LEVEL1_STATUS l1tcp_accept (LEVEL1_HANDLE	*client_handle
						  , LEVEL1_HANDLE	server_handle)
{
	long			blocking;
	TCP_ADDR_LEN	addr_len;
	TCP_ADDR_IN		addr;
	DEVICE_DEF		*client;
	DEVICE_DEF		*server;

	TRACE ("l1tcp_accept, entry point");
	if (!server_handle) {
		ERROR_TRACE ("tcpip_accept, NULL handle");
		return LEVEL1_BAD_HANDLE;
	}
	server = (DEVICE_DEF *) server_handle;
	client = add_device (server -> name, CLASS_ACCEPTED);
	if (!client) {
		ERROR_TRACE ("l1tcp_accept, add_device failed");
		*client_handle = (LEVEL1_HANDLE) NULL;
		return LEVEL1_NO_MEMORY;
	}

	addr_len = sizeof (addr);
	client -> socket_descr = accept (server -> socket_descr,
									 (TCP_ADDR *) &addr,
									 &addr_len);
	if (-1 == client -> socket_descr) {
		ERROR_TRACE_SYS ("l1tcp_accept, accept (%ld) = %s");
		delete_device (client);
		*client_handle = (LEVEL1_HANDLE) NULL;
		return LEVEL1_NO_MEMORY;
	}
	TRACE_I ("l1tcp_accept, socket_descr = %d", client -> socket_descr);
	if (!set_options (client -> socket_descr)) {
		ERROR_TRACE ("l1tcp_accept, set_options failed");
		shutdown (client -> socket_descr, 2);
		SOCKET_CLOSE (client -> socket_descr);
		delete_device (client);
		*client_handle = (LEVEL1_HANDLE) NULL;
		return LEVEL1_IO_ERROR;
	}
	blocking = 1;
	NON_BLOCKING (client -> socket_descr, blocking);
	if (!localtcp_add (client -> socket_descr)) {
		ERROR_TRACE ("l1tcp_accept, localtcp_add failed");
		shutdown (client -> socket_descr, 2);
		SOCKET_CLOSE (client -> socket_descr);
		delete_device (client);
		*client_handle = (LEVEL1_HANDLE) NULL;
		return LEVEL1_NO_MEMORY;
	}
	*client_handle = (LEVEL1_HANDLE) client;
	TRACE ("l1tcp_accept, connection accepted");
	return LEVEL1_NORMAL;
}

/****
*
****/

LEVEL1_STATUS l1tcp_write (LEVEL1_HANDLE		user_handle
						 , size_t				buffer_size
						 , void 				*buffer)
{
	size_t		bytes_written;
	size_t		send_size;
	DEVICE_DEF	*device;
	UCHAR		send_buffer [MAX_BUFFER_SIZE];

	TRACE_I ("l1tcp_write, buffer_size = %d", buffer_size);
	if (!user_handle) {
		ERROR_TRACE ("l1tcp_write, NULL handle");
		return LEVEL1_BAD_HANDLE;
	}
	device = (DEVICE_DEF *) user_handle;
	TRACE_S ("l1tcp_write, device = %s", device -> name);
	if (device -> escaped) {
		escape_data (buffer_size, buffer, send_buffer, &send_size);
	} else {
		memcpy (send_buffer, buffer, buffer_size);
		send_size = buffer_size;
	}
	TRACE_B ("l1tcp_write, out", send_size, send_buffer);
	TRACE_I ("l1tcp_write, socket_descr = %d", device -> socket_descr);
	bytes_written = send (device -> socket_descr, send_buffer, send_size, 0);
	if (bytes_written != send_size) {
		ERROR_TRACE_SYS ("l1tcp_write, send (%ld) = %s");
		ERROR_TRACE_I ("l1tcp_write, send_size = %d", send_size);
		ERROR_TRACE_I ("l1tcp_write, bytes_written = %d", bytes_written);
		return LEVEL1_SHUTDOWN;
	}
	TRACE ("l1tcp_write, returns normal");
	return LEVEL1_NORMAL;
}

/****
*
****/

LEVEL1_STATUS l1tcp_poll (LEVEL1_HANDLE user_handle, size_t *bytes_ready)
{
	int			count;
	int			status;
	DEVICE_DEF	*device;

	if (!user_handle) {
		ERROR_TRACE ("l1tcp_poll, NULL handle");
		*bytes_ready = 0;
		return LEVEL1_BAD_HANDLE;
	}
	device = (DEVICE_DEF *) user_handle;
	if (device -> shutdown_ready) {
		TRACE ("l1tcp_poll, device needs to be shutdown");
		*bytes_ready = 0;
		return LEVEL1_NORMAL;
	}
	count = localtcp_poll (device -> socket_descr);
	if (count == -1) {
		count = localtcp_poll (device -> socket_descr);
		if (count == -1) {
			ERROR_TRACE_SYS ("l1tcp_poll, localtcp_poll (%ld) = %s");
			device -> shutdown_ready = TRUE;
			*bytes_ready = 0;
			return LEVEL1_NORMAL;
		}
		*bytes_ready = 0;
		return LEVEL1_NO_DATA;
	} else if (count == 0) {
		*bytes_ready = 0;
		return LEVEL1_NO_DATA;
	}

	TRACE_I ("l1tcp_poll, select count = %d", count);
	TRACE_I ("l1tcp_poll, socket = %d", device -> socket_descr);
	if (CLASS_PUBLISHED == device -> tcp_class) {
		TRACE ("l1tcp_poll, published socket");
		*bytes_ready = 0;
		return LEVEL1_NORMAL;
	}
	status = read_socket (device);
	if (LEVEL1_NO_DATA != status) {
		TRACE_I ("l1tcp_poll, read_socket = %d", status);
		*bytes_ready = device -> read_index;
		return LEVEL1_NORMAL;
	}
	*bytes_ready = 0;
	return LEVEL1_NO_DATA;
}

/****
*
****/

LEVEL1_STATUS l1tcp_read (LEVEL1_HANDLE	user_handle
						, size_t		buffer_size
						, void			*buffer
						, size_t		*return_length)
{
	DEVICE_DEF	*device;

	TRACE ("l1tcp_read, entry point");
	if (!user_handle) {
		ERROR_TRACE ("l1tcp_read, NULL handle");
		*return_length = 0;
		return LEVEL1_BAD_HANDLE;
	}
	device = (DEVICE_DEF *) user_handle;
	if (device -> shutdown_ready) {
		TRACE ("l1tcp_read, device is shutdown");
		*return_length = 0;
		return LEVEL1_SHUTDOWN;
	}
	if (!device -> read_index) {
		TRACE ("l1tcp_read, no data");
		*return_length = 0;
		return LEVEL1_NO_DATA;
	}
	if (device -> read_index > buffer_size) {
		*return_length = buffer_size;
	} else {
		*return_length = device -> read_index;
	}
	memcpy (buffer, device -> read_buffer, *return_length);
	device -> read_index = 0;
	TRACE_I ("l1tcp_read, *return_length = %d", *return_length);
	return LEVEL1_NORMAL;
}

/****
*
****/

LEVEL1_STATUS l1tcp_clear (LEVEL1_HANDLE user_handle)
{
	DEVICE_DEF	*device;

	TRACE ("l1tcp_clear, entry point");
	if (!user_handle) {
		ERROR_TRACE ("l1tcp_clear, NULL handle");
		return LEVEL1_BAD_HANDLE;
	}
	device = (DEVICE_DEF *) user_handle;
	return LEVEL1_NOT_IMPLEMENTED;
}

/****
*
****/

LEVEL1_STATUS l1tcp_shutdown (LEVEL1_HANDLE user_handle)
{
	int			status;
	DEVICE_DEF	*device;

	TRACE ("l1tcp_shutdown, entry point");
	if (!user_handle) {
		ERROR_TRACE ("l1tcp_shutdown, NULL handle");
		return LEVEL1_BAD_HANDLE;
	}
	device = (DEVICE_DEF *) user_handle;
	localtcp_remove (device -> socket_descr);
	switch (device -> tcp_class) {
	case CLASS_ACCEPTED:
	case CLASS_CONNECTED:
		status = shutdown (device -> socket_descr, 2);
		if (status) {
			TRACE_I ("l1tcp_shutdown, status = %d", status);
			TRACE_SYS ("l1tcp_shutdown, shutdown (%ld) = %s");
		}
		break;
	default:
		break;
	}
	SOCKET_CLOSE (device -> socket_descr);
	delete_device (device);
	return LEVEL1_NORMAL;
}

/****
*
****/

LEVEL1_STATUS l1tcp_exit (void)
{
	DEVICE_DEF		*device;

	TRACE ("l1tcp_exit, entry point");
	while (queue_next (&root_device)) {
		TRACE ("l1tcp_exit, shutdown next handle");
		device = (DEVICE_DEF *) queue_next (&root_device);
		l1tcp_shutdown ((LEVEL1_HANDLE) device);
	}
	return LEVEL1_NORMAL;
}

/****
*
****/

static DEVICE_DEF *add_device (char *device_name, TCP_CLASS tcp_class)
{
	DEVICE_DEF	*device;
	size_t		read_buffer_size;

	device = (DEVICE_DEF *) malloc (sizeof (DEVICE_DEF));
	if (!device) {
		ERROR_TRACE_I ("add_device, malloc #1 = %d", sizeof (DEVICE_DEF));
		return (DEVICE_DEF *) NULL;
	}
	switch (tcp_class) {
	case CLASS_CONNECTED:
	case CLASS_ACCEPTED:
		read_buffer_size = ini_get_int (LEVEL1_GROUP, "MAX_FRAME", MAX_BUFFER_SIZE);
		device -> read_buffer_size = ini_get_int (device_name
												, "MAX_FRAME"
												, read_buffer_size);
		device -> read_buffer = (UCHAR *) malloc (device -> read_buffer_size);
		if (!device -> read_buffer) {
			ERROR_TRACE_I ("add_device, malloc #2 = %d"
						  , device -> read_buffer_size);
			delete_device (device);
			return (DEVICE_DEF *) NULL;
		}
		break;
	default:
		device -> read_buffer_size = 0;
		device -> read_buffer = (UCHAR *) NULL;
	}
	device -> escaped = ini_get_int (device_name, "ESCAPED", FALSE);
	device -> read_index = 0;
	device -> shutdown_ready = FALSE;
	device -> tcp_class = tcp_class;
	strcpy (device -> name, device_name);
	queue_append (&root_device, &device -> link);
	return device;
}

/****
*
****/

static void delete_device (DEVICE_DEF *device)
{
	TRACE ("delete_device, entry point");
	if (device -> read_buffer) {
		free (device -> read_buffer);
	}
	queue_remove (&root_device, &device -> link);
	free (device);
}

/****
*
****/

static int set_options (int socket_descr)
{
	int		status;
	int		option_value;

	TRACE ("set_options, entry point");
	option_value = 1;
	status = SET_OPTION (socket_descr
					   , SOL_SOCKET
					   , SO_REUSEADDR
					   , (char *) &option_value
					   , sizeof (option_value));
	if (status) {
		ERROR_TRACE_SYS ("set_options, ioctl (%ld) = %s");
		ERROR_TRACE_I ("set_options, set_option #1 = %d", status);
		return FALSE;
	}
	option_value = 1;
	status = SET_OPTION (socket_descr
					   , SOL_SOCKET
					   , SO_KEEPALIVE
					   , (char *) &option_value
					   , sizeof (option_value));
	if (status) {
		ERROR_TRACE_SYS ("set_options, ioctl (%ld) = %s");
		ERROR_TRACE_I ("set_options, set_option #2 = %d", status);
		return FALSE;
	}
	return TRUE;
}

/****
*
****/

static int read_socket (DEVICE_DEF *device)
{
	int			count;
	int			max_read_size;
	UCHAR		read_buffer [MAX_STRING];

	max_read_size = MIN (MAX_STRING
					   , device -> read_buffer_size - device -> read_index);
	if (!max_read_size) {
		ERROR_TRACE_I ("read_socket, read_buffer_size = %d"
					, device -> read_buffer_size);
		ERROR_TRACE_I ("read_socket, read_index = %d"
					 , device -> read_index);
		device -> shutdown_ready = TRUE;
		return LEVEL1_SHUTDOWN;
	}
	count = recv (device -> socket_descr
				, read_buffer
				, max_read_size
				, 0);
	if (count < 0) {
		if (ECONNRESET == GetLastTcpError ()) {
			ERROR_TRACE ("read_socket, ECONNRESET");
			device -> shutdown_ready = TRUE;
			return LEVEL1_SHUTDOWN;
		} else {
			ERROR_TRACE_I ("read_socket, count = %d", count);
			ERROR_TRACE_SYS ("read_socket, recv (%ld) = %s");
			device -> shutdown_ready = TRUE;
			return LEVEL1_SHUTDOWN;
		}
	} else if (count == 0) {
		TRACE_I ("read_socket, recv count = 0, normal shutdown on %d"
				, device -> socket_descr);
		device -> shutdown_ready = TRUE;
		return LEVEL1_SHUTDOWN;
	}
	if (device -> escaped) {
		count = insert_data (device, count, read_buffer);
		if (!count) {
			TRACE ("read_socket, all new data was discarded");
			return LEVEL1_NO_DATA;
		}
	} else {
		memcpy (&device -> read_buffer [device -> read_index]
			  , read_buffer
			  , count);
		device -> read_index += count;
	}
	TRACE_I ("read_socket, read_index = %d", device -> read_index);
	return LEVEL1_NORMAL;
}

static int insert_data (DEVICE_DEF *device, size_t buffer_size, UCHAR *buffer)
{
	size_t		out_count;
	UCHAR		*in_buffer;
	UCHAR		*out_buffer;

	out_count = 0;
	in_buffer = buffer;
	out_buffer = &device -> read_buffer [device -> read_index];
	while (buffer_size) {
		--buffer_size;
		if (ESCAPE_CHAR == device -> last_char) {
			if (ESCAPE_CHAR == *in_buffer) {
				++in_buffer;
				device -> last_char = 0;
			} else if ((0xFB == *in_buffer)
			||         (0xFD == *in_buffer)) {
				++in_buffer;
				if (out_count) {
					--out_count;
					--out_buffer;
				} else {
					if (device -> read_index) {
						--device -> read_index;
						--out_buffer;
					}
				}
				device -> last_char = ESCAPE_CHAR;
			} else {
				ERROR_TRACE ("this is where to put the telnet stuff");
				device -> last_char = *in_buffer;
				if (device -> read_index) {
					--device -> read_index;
					--out_buffer;
				}
				if (out_count) {
					--out_count;
				}
				++in_buffer;
			}
		} else if (0x0D == device -> last_char) {
			if (0x00 == *in_buffer) {
				++in_buffer;
				device -> last_char = 0;
			} else {
				*out_buffer = *in_buffer;
				device -> last_char = *in_buffer;
				++out_count;
				++out_buffer;
				++in_buffer;
			}
		} else {
			*out_buffer = *in_buffer;
			device -> last_char = *in_buffer;
			++out_count;
			++out_buffer;
			++in_buffer;
		}
	}
	device -> read_index += out_count;
	return out_count;
}

static void escape_data (size_t		in_buffer_size
					   , UCHAR		*in_buffer
					   , UCHAR		*out_buffer
					   , size_t		*return_size)
{
	int		count;

	TRACE_B ("escape_data before", in_buffer_size, in_buffer);
	count = 0;
	while (in_buffer_size) {
		if (ESCAPE_CHAR == *in_buffer) {
			*out_buffer = ESCAPE_CHAR;
			++out_buffer;
			++count;
		}
		*out_buffer = *in_buffer;
		if (0x0D == *in_buffer) {
			++out_buffer;
			++count;
			*out_buffer = 0;
		}
		++out_buffer;
		++count;
		++in_buffer;
		--in_buffer_size;
	}
	*return_size = count;
}

