/****
*
*	File name :		l1_unix.c
*	Author :		Steve Johnson
*	Creation date :	December 1, 1985
*
*	Copyright (c) 1985,1999 Sigma Six Corporation
*
*	Provide the following entry points
*
*		unix_init
*		unix_connect
*		unix_publish
*		unix_accept
*		unix_write
*		unix_poll
*		unix_read
*		unix_clear
*		unix_shutdown
*		unix_exit
*
*	Physical I/O layer for UNIX tty type devices
*
****/

#include <sscommon.h>
#include <sstrace.h>
#include <ssnames.h>
#include <ssqueues.h>
#include <sslevel1.h>
#include <ssl1unix.h>

#define LEVEL1_GROUP		"LEVEL1_UNIX"
#define DEFAULT_MAX_FRAME	2048

/****
*
*	Device context definitions
*
****/

typedef struct {
	QUEUE_DEF	link;
	char		*user_buffer;
	UCHAR		*read_buffer;
	size_t		max_frame;
	size_t		read_count;
	size_t		read_index;
	int			write_port;
	int			read_port;
	char		name [MAX_STRING];
} DEVICE_DEF;

static QUEUE_DEF	root_device;

static DEVICE_DEF	*add_device (char *);
static void			delete_device (DEVICE_DEF *);
static int			get_char (DEVICE_DEF *, UCHAR *);

/****
*
****/

LEVEL1_STATUS unix_init (void)
{
	TRACE ("unix_init, entry point");
	queue_init (&root_device);
	return LEVEL1_NORMAL;
}

/****
*
****/

LEVEL1_STATUS unix_connect (LEVEL1_HANDLE *user_handle, char *name)
{
	char		port_name [MAX_STRING];
	long		local_baud_rate;
	long		default_baud_rate;
	size_t		default_max_frame;
	DEVICE_DEF	*device;

	TRACE ("unix_connect, entry point");
	if (strlen (name) >= MAX_STRING) {
		ERROR_TRACE_S ("unix_connect, invalid device name = (%s)", name);
		return LEVEL1_BAD_NAME;
	}
	if (!ini_get_str (name, "TYCO_PORT", port_name)) {
		strcpy (port_name, name);
	}
	TRACE_S ("unix_connect, port_name = %s", port_name);
	device = add_device (port_name);
	if (!device) {
		ERROR_TRACE ("unix_connect, add_device failed");
		return LEVEL1_NO_MEMORY;
	}
	device -> read_port = open (port_name, O_RDONLY, 0777);
	if (ERROR == device -> read_port) {
		ERROR_TRACE_I ("unix_connect, fopen (r) errno = %d", errno);
		ERROR_TRACE_S ("unix_connect, port_name = %s", port_name);
		delete_device (device);
		return LEVEL1_IO_ERROR;
	}
	device -> write_port = open (port_name, O_WRONLY, 0777);
	if (ERROR == device -> write_port) {
		ERROR_TRACE_I ("unix_connect, fopen (W) errno = %d", errno);
		ERROR_TRACE_S ("unix_connect, port_name = %s", port_name);
		close (device -> read_port);
		delete_device (device);
		return LEVEL1_IO_ERROR;
	}
	default_baud_rate = ini_get_long (LEVEL1_GROUP, "BAUD_RATE", 4800L);
	local_baud_rate = ini_get_long (name, "BAUD_RATE", default_baud_rate);
	default_max_frame = ini_get_int (LEVEL1_GROUP, "MAX_FRAME", DEFAULT_MAX_FRAME);
	device -> max_frame = ini_get_int (name, "MAX_FRAME", default_max_frame);
	device -> read_buffer = (UCHAR *) malloc (device -> max_frame);
	if (!device -> read_buffer) {
		ERROR_TRACE_S ("unix_connect, port_name = %s", port_name);
		ERROR_TRACE_I ("unix_connect, malloc size = %d", device -> max_frame);
		close (device -> read_port);
		close (device -> write_port);
		delete_device (device);
		return LEVEL1_NO_MEMORY;
	}
	(void) ioctl (device -> read_port, FIOBAUDRATE, local_baud_rate);
	(void) ioctl (device -> read_port, FIOSETOPTIONS, OPT_RAW);
	(void) ioctl (device -> write_port, FIOBAUDRATE, local_baud_rate);
	(void) ioctl (device -> write_port, FIOSETOPTIONS, OPT_RAW);

	device -> read_count = 0;
	device -> read_index = 0;
	*user_handle = (LEVEL1_HANDLE) device;
	TRACE ("unix_connect, successful");
	return LEVEL1_NORMAL;
}

/****
*
****/

LEVEL1_STATUS unix_publish (LEVEL1_HANDLE *user_handle, char *name)
{
	ERROR_TRACE_S ("unix_publish, not supported (%s)", name);
	return LEVEL1_NOT_IMPLEMENTED;
}

/****
*
****/

LEVEL1_STATUS unix_accept (LEVEL1_HANDLE	*client_handle
						 , LEVEL1_HANDLE	server_handle)
{
	ERROR_TRACE ("unix_accept, not supported");
	return LEVEL1_NOT_IMPLEMENTED;
}

/****
*
****/

LEVEL1_STATUS unix_write (LEVEL1_HANDLE	user_handle
						, size_t		buffer_size
						, void			*buffer)
{
	size_t		bytes_written;
	DEVICE_DEF	*device;

	TRACE_B ("unix_write", buffer_size, buffer);
	if (!user_handle) {
		ERROR_TRACE ("unix_write, NULL handle");
		return LEVEL1_BAD_HANDLE;
	}
	device = (DEVICE_DEF *) user_handle;
	TRACE_S ("unix_write, device = %s", device -> name);
	bytes_written = write (device -> write_port, buffer, buffer_size);
	if (bytes_written != buffer_size) {
		ERROR_TRACE_I ("unix_write, write error = %d", errno);
		ERROR_TRACE_I ("unix_write, buffer_size = %d", buffer_size);
		ERROR_TRACE_I ("unix_write, bytes_written = %d", bytes_written);
		return LEVEL1_SHUTDOWN;
	}
	TRACE ("unix_write, returns normal");
	return LEVEL1_NORMAL;
}

/****
*
****/

LEVEL1_STATUS unix_poll (LEVEL1_HANDLE user_handle, size_t *bytes_ready)
{
	int			count;
	DEVICE_DEF	*device;

	if (!user_handle) {
		ERROR_TRACE ("unix_poll, NULL handle");
		*bytes_ready = 0;
		return LEVEL1_BAD_HANDLE;
	}
	device = (DEVICE_DEF *) user_handle;
	if (!device -> read_count) {
		(void) ioctl (device -> read_port, FIONREAD, (int) &count);
		if (!count) {
			*bytes_ready = 0;
			return LEVEL1_NO_DATA;
		}
		TRACE_I ("unix_poll, %d bytes ioctl ready", count);
		*bytes_ready = count;
	} else {
		*bytes_ready = device -> read_count;
	}
	return LEVEL1_NORMAL;
}

/****
*
****/

LEVEL1_STATUS unix_read (LEVEL1_HANDLE	user_handle
					   , size_t			buffer_size
					   , void			*buffer
					   , size_t			*return_length)
{
	int			status;
	size_t		count;
	UCHAR		*char_buffer;
	DEVICE_DEF	*device;

	if (!user_handle) {
		ERROR_TRACE ("unix_read, NULL handle");
		return LEVEL1_BAD_HANDLE;
	}
	device = (DEVICE_DEF *) user_handle;
	count = 0;
	char_buffer = (UCHAR *) buffer;
	while (count < buffer_size) {
		status = get_char (device, char_buffer);
		if (LEVEL1_NO_DATA == status) {
			*return_length = count;
			if (count) {
				TRACE ("unix_read, no more data");
				return LEVEL1_NORMAL;
			} else {
				TRACE ("unix_read, no data");
				return LEVEL1_NO_DATA;
			}
		} else if (LEVEL1_NORMAL != status) {
			TRACE_I ("unix_read, get_char = %d", status);
			return status;
		}
		++count;
		++char_buffer;
	}
	*return_length = count;
	return LEVEL1_NORMAL;
}

/****
*
****/

LEVEL1_STATUS unix_clear (LEVEL1_HANDLE user_handle)
{
	DEVICE_DEF	*device;

	TRACE ("unix_clear, entry point");
	if (!user_handle) {
		ERROR_TRACE ("unix_clear, NULL handle");
		return LEVEL1_BAD_HANDLE;
	}
	device = (DEVICE_DEF *) user_handle;
	return LEVEL1_NOT_IMPLEMENTED;
}

/****
*
****/

LEVEL1_STATUS unix_shutdown (LEVEL1_HANDLE user_handle)
{
	DEVICE_DEF	*device;

	TRACE ("unix_shutdown, entry point");
	if (!user_handle) {
		ERROR_TRACE ("unix_shutdown, NULL handle");
		return LEVEL1_BAD_HANDLE;
	}
	device = (DEVICE_DEF *) user_handle;
	if (device -> read_buffer) {
		free (device -> read_buffer);
	}
	close (device -> read_port);
	close (device -> write_port);
	delete_device (device);
	return LEVEL1_NORMAL;
}

/****
*
****/

LEVEL1_STATUS unix_exit (void)
{
	DEVICE_DEF		*device;

	TRACE ("unix_exit, entry point");
	while (queue_next (&root_device)) {
		TRACE ("unix_exit, shutdown next handle");
		device = (DEVICE_DEF *) queue_next (&root_device);
		level1_shutdown ((LEVEL1_HANDLE) device);
	}
	return LEVEL1_NORMAL;
}

/****
*
****/

static DEVICE_DEF *add_device (char *device_name)
{
	DEVICE_DEF	*device;

	device = (DEVICE_DEF *) malloc (sizeof (DEVICE_DEF));
	if (!device) {
		ERROR_TRACE_I ("add_device, malloc size = %d",
				sizeof (DEVICE_DEF));
		return NULL;
	}
	strcpy (device -> name, device_name);
	device -> read_buffer = (UCHAR *) NULL;
	queue_append (&root_device, &device -> link);
	return device;
}

/****
*
****/

static void delete_device (DEVICE_DEF *device)
{
	TRACE ("delete_device, entry point");
	queue_remove (&root_device, &device -> link);
	free (device);
}

/*******
*	
*	reads the 8250 port status
*	does not wait for one if there isn't one available
*	returns the character in the char pointed to by chartoreturn
*
*******/

static int get_char (DEVICE_DEF *device, UCHAR *inchar)
{
	size_t		bytes_ready;
	size_t		bytes_read;

	if (!device -> read_count) {
		(void) ioctl (device -> read_port, FIONREAD, (int) &bytes_ready);
		if (!bytes_ready) {
			TRACE ("get_char, no data");
			return LEVEL1_NO_DATA;
		}
		if (bytes_ready > device -> max_frame) {
			ERROR_TRACE ("get_char, overflow");
			ERROR_TRACE_I ("get_char, max_frame = %d", device -> max_frame);
			ERROR_TRACE_I ("get_char, bytes_ready = %d", bytes_ready);
			bytes_ready = device -> max_frame;
		}
		bytes_read = read (device -> read_port
						 , device -> read_buffer
						 , bytes_ready);
		if (bytes_read != bytes_ready) {
			ERROR_TRACE_I ("get_char, read failed errno = %d", errno);
			ERROR_TRACE_I ("get_char, bytes_read = %d", bytes_read);
			ERROR_TRACE_I ("get_char, bytes_ready = %d", bytes_ready);
		}
		TRACE_B ("get_char", bytes_read, device -> read_buffer);
		device -> read_count = bytes_read;
		device -> read_index = 0;
	} 
	*inchar = device -> read_buffer [device -> read_index];
	++device -> read_index;
	--device -> read_count;
	return LEVEL1_NORMAL;
}

/****
*
*	select style poll is obsolete
*
*	int				file_descr;
*	FD_BIT_MASK		write_mask;
*	FD_BIT_MASK		except_mask;
*	FD_BIT_MASK		read_mask;
*	DEVICE_DEF		*device;
*	struct timeval	time_value;
*
*	selectInit ();
*	file_descr = fileno (device -> read_port);
*	FD_ZERO (&write_mask);
*	FD_ZERO (&except_mask);
*	FD_ZERO (&read_mask);
*	FD_SET (file_descr, &except_mask);
*	FD_SET (file_descr, &read_mask);
*	time_value . tv_sec = 0;
*	time_value . tv_usec = 0;
*	count = select (FD_SETSIZE,
*					&read_mask,
*					&write_mask,
*					&except_mask,
*					&time_value);
*	if (count == -1) {
*		ERROR_TRACE_I ("basic_connect, select = %d", GetLastError ());
*		return LEVEL1_ABORTED;
*	} else if (count) {
*		if (FD_ISSET (file_descr, &except_mask)) {
*			ERROR_TRACE_I ("unix_poll, select = %d", GetLastError ());
*			return LEVEL1_ABORTED;
*		}
*		if (FD_ISSET (file_descr, &read_mask)) {
*			TRACE ("unix_poll, successful");
*			*bytes_ready = 1;
*			return LEVEL1_NORMAL;
*		} else {
*			ERROR_TRACE ("unix_poll, unsuccessful");
*		}
*	}
*
****/
