/****
*
*	File name :		level1.c
*	Author :		Steve Johnson
*	Creation date :	December 1, 1985
*
*	Copyright (c) 1985,1999 Sigma Six Corporation
*
*	Provides the following entry points
*
*		level1_init
*		level1_publish
*		level1_connect
*		level1_accept
*		level1_write
*		level1_read
*		level1_select
*		level1_clear
*		level1_shutdown
*		level1_exit
*
****/

#include <sscommon.h>
#include <sstrace.h>
#include <ssqueues.h>
#include <ssnames.h>
#include <sslevel1.h>
#include <ssl1tcp.h>
#include <ssl18250.h>
#include <ssl1win.h>
#include <ssl1ssl.h>
#include <ssl1digi.h>
#include <ssl1unix.h>

/****
*
****/

typedef struct {
	char		*name;
	QUEUE_DEF	*root_handle;
	int			on_line;
	int			(*init_procedure) (void);
	int			(*publish_procedure) (LEVEL1_HANDLE *, char *);
	int			(*connect_procedure) (LEVEL1_HANDLE *, char *);
	int			(*accept_procedure) (LEVEL1_HANDLE *, LEVEL1_HANDLE);
	int			(*write_procedure) (LEVEL1_HANDLE, size_t, void *);
	int			(*read_procedure) (LEVEL1_HANDLE, size_t, void *, size_t *);
	int			(*poll_procedure) (LEVEL1_HANDLE, size_t *);
	int			(*clear_procedure) (LEVEL1_HANDLE);
	int			(*shutdown_procedure) (LEVEL1_HANDLE);
	int			(*exit_procedure) (void);
} DEVICE_DEF;

static DEVICE_DEF device_table [] = {
	{ "L1_TCP",
		NULL,
		FALSE,
		l1tcp_init,
		l1tcp_publish,
		l1tcp_connect,
		l1tcp_accept,
		l1tcp_write,
		l1tcp_read,
		l1tcp_poll,
		l1tcp_clear,
		l1tcp_shutdown,
		l1tcp_exit },
	{ "L1_SSL",
		NULL,
		FALSE,
		l1ssl_init,
		l1ssl_publish,
		l1ssl_connect,
		l1ssl_accept,
		l1ssl_write,
		l1ssl_read,
		l1ssl_poll,
		l1ssl_clear,
		l1ssl_shutdown,
		l1ssl_exit },
	{ "L1_8250",
		NULL,
		FALSE,
		pc8250_init,
		pc8250_publish,
		pc8250_connect,
		pc8250_accept,
		pc8250_write,
		pc8250_read,
		pc8250_poll,
		pc8250_clear,
		pc8250_shutdown,
		pc8250_exit },
	{ "L1_WIN32",
		NULL,
		FALSE,
		wincom_init,
		wincom_publish,
		wincom_connect,
		wincom_accept,
		wincom_write,
		wincom_read,
		wincom_poll,
		wincom_clear,
		wincom_shutdown,
		wincom_exit },
	{ "L1_DIGI",
		NULL,
		FALSE,
		digi_init,
		digi_publish,
		digi_connect,
		digi_accept,
		digi_write,
		digi_read,
		digi_poll,
		digi_clear,
		digi_shutdown,
		digi_exit },
	{ "L1_UNIX",
		NULL,
		FALSE,
		unix_init,
		unix_publish,
		unix_connect,
		unix_accept,
		unix_write,
		unix_read,
		unix_poll,
		unix_clear,
		unix_shutdown,
		unix_exit },
};
#define NUM_DEVICES	(sizeof (device_table) / sizeof (*device_table))

/****
*
****/

typedef struct {
	QUEUE_DEF		link;
	LEVEL1_HANDLE	lower_handle;
	DEVICE_DEF		*device;
	char			name [MAX_STRING];
} HANDLE_DEF;

static void				delete_handle (HANDLE_DEF *);
static HANDLE_DEF		*add_handle (DEVICE_DEF *, char *);
static DEVICE_DEF		*get_device_by_name (char *);

/****
*
****/

LEVEL1_STATUS level1_init (void)
{
	int				index;
	LEVEL1_STATUS	status;
	DEVICE_DEF		*device;
	int				(*init_procedure) (void);

	TRACE ("level1_init, entry point");
	for (index = 0; index < NUM_DEVICES; ++index) {
		device = &device_table [index];
		device -> root_handle = (QUEUE_DEF *) malloc (sizeof (QUEUE_DEF));
		if (!device -> root_handle) {
			ERROR_TRACE ("level1_init, malloc QUEUE_DEF");
			device -> on_line = FALSE;
			level1_exit ();
			return LEVEL1_NO_MEMORY;
		}
		device -> on_line = FALSE;
		queue_init (device -> root_handle);
	}
	for (index = 0; index < NUM_DEVICES; ++index) {
		device = &device_table [index];
		init_procedure = device -> init_procedure;
		status = (*init_procedure) ();
		if (LEVEL1_NORMAL == status) {
			TRACE_I ("level1_init, good index = %d", index);
			device -> on_line = TRUE;
		} else if (LEVEL1_NOT_IMPLEMENTED == status) {
			TRACE_I ("level1_init, not implemented = %d", index);
			device -> on_line = FALSE;
		} else if (LEVEL1_OFFLINE == status) {
			TRACE_I ("level1_init, device offline = %d", index);
			device -> on_line = FALSE;
		} else {
			ERROR_TRACE_I ("level1_init, status = %d", status);
			ERROR_TRACE_I ("level1_init, index = %d", index);
			device -> on_line = FALSE;
			level1_exit ();
			return status;
		}
	}
	TRACE ("level1_init, normal return");
	return LEVEL1_NORMAL;
}

/****
*
***/

LEVEL1_STATUS level1_publish (LEVEL1_HANDLE *user_handle, char *name)
{
	LEVEL1_STATUS	status;
	HANDLE_DEF		*handle;
	DEVICE_DEF		*device;
	int				(*publish_procedure) (LEVEL1_HANDLE *, char *);

	TRACE ("level1_publish, entry point");
	TRACE_S ("level1_publish, name = %s", name);
	device = get_device_by_name (name);
	if (!device) {
		ERROR_TRACE_S ("level1_publish, unknown device = %s", name);
		return LEVEL1_BAD_NAME;
	}
	if (!device -> on_line) {
		ERROR_TRACE_S ("level1_publish, device unavailable on = %s", name);
		return LEVEL1_OFFLINE;
	}
	TRACE_S ("level1_publish, base name = %s", name);
	handle = add_handle (device, name);
	if (!handle) {
		ERROR_TRACE ("level1_publish, add_handle failed");
		ERROR_TRACE_S ("level1_publish, name = %s", name);
		return LEVEL1_NO_MEMORY;
	}
	publish_procedure = device -> publish_procedure;
	status = (*publish_procedure) (&handle -> lower_handle, name);
	if (LEVEL1_NORMAL != status) {
		ERROR_TRACE_I ("level1_publish, return status = %d", status);
		delete_handle (handle);
		return status;
	}
	*user_handle = (LEVEL1_HANDLE) handle;
	TRACE_I ("level1_publish, status = %d", status);
	return status;
}

/****
*
****/

LEVEL1_STATUS level1_connect (LEVEL1_HANDLE *user_handle, char *name)
{
	LEVEL1_STATUS	status;
	HANDLE_DEF		*handle;
	DEVICE_DEF		*device;
	int				(*connect_procedure) (LEVEL1_HANDLE *, char *);

	TRACE ("level1_connect, entry point");
	TRACE_S ("level1_connect, name = %s", name);
	device = get_device_by_name (name);
	if (!device) {
		ERROR_TRACE_S ("level1_connect, device unknown = %s", name);
		return LEVEL1_BAD_NAME;
	}
	if (!device -> on_line) {
		ERROR_TRACE ("level1_connect, device unavailable");
		ERROR_TRACE_S ("level1_connect, name = %s", name);
		return LEVEL1_OFFLINE;
	}
	handle = add_handle (device, name);
	if (!handle) {
		ERROR_TRACE ("level1_connect, add_handle failed");
		return LEVEL1_NO_MEMORY;
	}
	connect_procedure = device -> connect_procedure;
	status = (*connect_procedure) (&handle -> lower_handle, name);
	if (LEVEL1_NORMAL != status) {
		TRACE_I ("level1_connect, return status = %d", status);
		delete_handle (handle);
		return status;
	}
	TRACE_I ("level1_connect, return status = %d", status);
	*user_handle = (LEVEL1_HANDLE) handle;
	return status;
}

/****
*
****/

LEVEL1_STATUS level1_accept (LEVEL1_HANDLE	*client_handle
						   , LEVEL1_HANDLE	server_handle)
{
	LEVEL1_STATUS	status;
	HANDLE_DEF 		*server;
	HANDLE_DEF 		*client;
	int				(*accept_procedure) (LEVEL1_HANDLE *, LEVEL1_HANDLE);

	TRACE ("level1_accept, entry point");
	if (!server_handle) {
		ERROR_TRACE ("level1_accept, NULL server handle");
		*client_handle = (LEVEL1_HANDLE) NULL;
		return LEVEL1_BAD_HANDLE;
	}
	server = (HANDLE_DEF *) server_handle;
	client = add_handle (server -> device, server -> name);
	if (!client) {
		ERROR_TRACE ("level1_accept, add_handle failed");
		*client_handle = (LEVEL1_HANDLE) NULL;
		return LEVEL1_NO_MEMORY;
	}
	*client_handle = (LEVEL1_HANDLE) client;
	accept_procedure = client -> device -> accept_procedure;
	status = (*accept_procedure) (&client -> lower_handle
								, server -> lower_handle);
	if (LEVEL1_NORMAL != status) {
		ERROR_TRACE_I ("level1_accept, return status = %d", status);
		delete_handle (client);
		*client_handle = (LEVEL1_HANDLE) NULL;
		return status;
	}
	TRACE_I ("level1_accept, return status = %d", status);
	return status;
}

/****
*
****/

LEVEL1_STATUS level1_write (LEVEL1_HANDLE	user_handle
						  , size_t			buffer_size
						  , void			*buffer)
{
	LEVEL1_STATUS	status;
	HANDLE_DEF		*handle;
	int				(*write_procedure) (LEVEL1_HANDLE, size_t, void *);

	TRACE ("level1_write, entry point");
	if (!user_handle) {
		ERROR_TRACE ("level1_write, NULL handle");
		return LEVEL1_BAD_HANDLE;
	}
	handle = (HANDLE_DEF *) user_handle;

	write_procedure = handle -> device -> write_procedure;
	status = (*write_procedure) (handle -> lower_handle,
							     buffer_size,
							     buffer);
	TRACE_I ("level1_write, return status = %d", status);
	return status;
}

/****
*
****/

LEVEL1_STATUS level1_read (LEVEL1_HANDLE	user_handle
						 , size_t			buffer_size
						 , void				*buffer
						 , size_t			*return_length)
{
	LEVEL1_STATUS	status;
	HANDLE_DEF		*handle;
	int				(*read_procedure) (LEVEL1_HANDLE, size_t, void *, size_t *);

	if (!user_handle) {
		ERROR_TRACE ("level1_read, NULL handle");
		return LEVEL1_BAD_HANDLE;
	}
	handle = (HANDLE_DEF *) user_handle;
	read_procedure = handle -> device -> read_procedure;
	status = (*read_procedure) (handle -> lower_handle,
							    buffer_size,
							    buffer,
							    return_length);
	return status;
}

/****
*
****/

LEVEL1_STATUS level1_poll (LEVEL1_HANDLE user_handle, size_t *bytes_ready)
{
	LEVEL1_STATUS	status;
	HANDLE_DEF		*handle;
	int				(*poll_procedure) (LEVEL1_HANDLE, size_t *);

	if (!user_handle) {
		ERROR_TRACE ("level1_poll, NULL handle");
		return LEVEL1_BAD_HANDLE;
	}
	handle = (HANDLE_DEF *) user_handle;
	poll_procedure = handle -> device -> poll_procedure;
	status = (*poll_procedure) (handle -> lower_handle, bytes_ready);
	if (LEVEL1_NO_DATA == status) {
		return LEVEL1_NO_DATA;
	}
	if (LEVEL1_NORMAL != status) {
		ERROR_TRACE_S ("level1_poll, handle = %s", handle -> name);
		ERROR_TRACE_I ("level1_poll, status = %d", status);
		return status;
	}
	return LEVEL1_NORMAL;
}

/****
*
****/

LEVEL1_STATUS level1_clear (LEVEL1_HANDLE user_handle)
{
	int			status;
	HANDLE_DEF	*handle;
	int			(*clear_procedure) (LEVEL1_HANDLE);

	TRACE ("level1_clear, entry point");
	if (!user_handle) {
		ERROR_TRACE ("level1_clear, NULL handle");
		return LEVEL1_BAD_HANDLE;
	}
	handle = (HANDLE_DEF *) user_handle;
	clear_procedure = handle -> device -> clear_procedure;
	status = (*clear_procedure) (handle -> lower_handle);
	TRACE_I ("level1_clear return status = %d", status);
	return status;
}

/****
*
****/

LEVEL1_STATUS level1_shutdown (LEVEL1_HANDLE user_handle)
{
	int			status;
	HANDLE_DEF	*handle;
	int			(*shutdown_procedure) (LEVEL1_HANDLE);

	TRACE ("level1_shutdown, entry point");
	if (!user_handle) {
		ERROR_TRACE ("level1_shutdown, NULL handle");
		return LEVEL1_BAD_HANDLE;
	}
	handle = (HANDLE_DEF *) user_handle;
	shutdown_procedure = handle -> device -> shutdown_procedure;
	status = (*shutdown_procedure) (handle -> lower_handle);
	delete_handle (handle);
	TRACE_I ("level1_shutdown return status = %d", status);
	return status;
}

/****
*
****/

LEVEL1_STATUS level1_exit (void)
{
	LEVEL1_STATUS	status;
	int				last_error;
	int				index;
	QUEUE_DEF		*root_handle;
	HANDLE_DEF		*loop_handle;
	int				(*exit_procedure) (void);

	TRACE ("level1_exit, entry point");
	last_error = LEVEL1_NORMAL;
	for (index = 0; index < NUM_DEVICES; ++index) {
		root_handle = device_table [index] . root_handle;
		while (queue_next (root_handle)) {
			TRACE ("level1_exit, shutdown one handle");
			loop_handle = (HANDLE_DEF *) queue_next (root_handle);
			level1_shutdown ((LEVEL1_HANDLE) loop_handle);
		}
		TRACE ("level1_exit, all handles closed");
		if (device_table [index] . on_line) {
			exit_procedure = device_table [index].exit_procedure;
			status = (*exit_procedure) ();
			if (LEVEL1_NORMAL != status) {
				ERROR_TRACE_I ("level1_exit, exit=%d",status);
				ERROR_TRACE_I ("level1_exit,index=%d",index);
				last_error = status;
			}
		}
		if (device_table [index] . root_handle) {
			free (device_table [index] . root_handle);
		} else {
			ERROR_TRACE ("level1_exit, no root handle");
		}
	}
	TRACE_I ("level1_exit return status = %d", last_error);
	return last_error;
}

static DEVICE_DEF *get_device_by_name (char *name)
{
	int		index;
	char	device_name [MAX_STRING];

	if (!ini_get_str (name, "DEVICE_TYPE", device_name)) {
		ERROR_TRACE_S ("get_device_by_name, no DEVICE_TYPE (%s)", name);
		return (DEVICE_DEF *) NULL;
	}
	for (index = 0; index < NUM_DEVICES; ++index) {
		if (case_compare (device_name, device_table [index] . name)) {
			return &device_table [index];
		}
	}
	ERROR_TRACE_S ("get_device_by_name, device_name = (%s)", device_name);
	return (DEVICE_DEF *) NULL;
}

/****
*
****/

static HANDLE_DEF *add_handle (DEVICE_DEF *device, char *name)
{
	HANDLE_DEF	*handle;

	handle = (HANDLE_DEF *) malloc (sizeof (HANDLE_DEF));
	if (!handle) {		
		ERROR_TRACE_I ("add_handle, malloc size = %d", sizeof (HANDLE_DEF));
		return (HANDLE_DEF *) NULL;
	}
	handle -> device = device;
	strcpy (handle -> name, name);
	queue_append (device -> root_handle, &handle -> link);
	return handle;
}

/****
*
****/

static void delete_handle (HANDLE_DEF *handle)
{
	TRACE ("delete_handle, (level1) entry point");
	queue_remove (handle -> device -> root_handle, &handle -> link);
	free (handle);
}

