/****
*
*	File name :		queues.c
*	Author :		Steve Johnson
*	Creation date :	December 1, 1985
*
*	Copyright (c) 1985,1999 Sigma Six Corporation
*
*	Provides the following entry points:
*
*		queue_init
*		queue_next
*		queue_tail
*		queue_get
*		queue_append
*		queue_new_head
*		queue_remove
*		queue_count
*		queue_search
*
*	This is a generic link list management layer
*
****/

#include <sscommon.h>
#include <sstrace.h>
#include <ssqueues.h>

/****
*
*	return the next item on the given queue, don't remove it
*
****/

void queue_init (QUEUE_DEF *queue_head)
{
	queue_head -> forward = (QUEUE_DEF *) NULL;
	queue_head -> backward = (QUEUE_DEF *) NULL;
}

/****
*
*	return the next item on the given queue, don't remove it
*
****/

void *queue_next (QUEUE_DEF *item)
{
	QUEUE_DEF	*queue_item;

	queue_item = (QUEUE_DEF *) item;
	return queue_item -> forward;
}

/****
*
*	return the last item on the given queue, don't remove it
*
****/

void *queue_tail (QUEUE_DEF *item)
{
	QUEUE_DEF	*queue_item;

	queue_item = (QUEUE_DEF *) item;
	return queue_item -> backward;
}

/****
*
*	add the given item to the tail of the given queue
*
****/

void queue_append (QUEUE_DEF *queue_head, QUEUE_DEF *queue_item)
{
	QUEUE_DEF	*last_item;

	if (queue_item) {
		DISABLE ();
		if (queue_head -> forward) {
			last_item = queue_head -> backward;
			if (!last_item) {
				ENABLE ();
				ERROR_TRACE ("queue_append, corrupted #1");
				return;
			}
			if (last_item -> forward) {
				ENABLE ();
				ERROR_TRACE ("queue_append, corrupted #2");
				return;
			}
			last_item -> forward = queue_item;
			queue_item -> forward = (QUEUE_DEF *) NULL;
			queue_item -> backward = last_item;
			queue_head -> backward = queue_item;
		} else {
			queue_item -> forward = (QUEUE_DEF *) NULL;
			queue_item -> backward = (QUEUE_DEF *) NULL;
			queue_head -> forward = queue_item;
			queue_head -> backward = queue_item;
		}
		ENABLE ();
	} else {
		ERROR_TRACE ("queue_append, null item to append");
	}
}

/****
*
*	add the given item to the head of the given queue
*
****/

void queue_new_head (QUEUE_DEF *queue_head, QUEUE_DEF *queue_item)
{
	QUEUE_DEF	*first_item;

	if (queue_item) {
		DISABLE ();
		if (queue_head -> forward) {
			first_item = queue_head -> forward;
			queue_head -> forward = queue_item;
			queue_item -> forward = first_item;
			queue_item -> backward = (QUEUE_DEF *) NULL;
			first_item -> backward = queue_item;
		} else {
			queue_item -> forward = (QUEUE_DEF *) NULL;
			queue_item -> backward = (QUEUE_DEF *) NULL;
			queue_head -> forward = queue_item;
			queue_head -> backward = queue_item;
		}
		ENABLE ();
	} else {
		ERROR_TRACE ("queue_new_head, null item to append");
	}
}

/****
*
*	move the given item to the end of the queue
*
****/

void queue_move_to_tail (QUEUE_DEF *queue_head, QUEUE_DEF *queue_item)
{
	queue_remove (queue_head, queue_item);
	queue_append (queue_head, queue_item);
}

/****
*
*	return the first item on the given queue and remove it
*
****/

void *queue_get (QUEUE_DEF *queue_head)
{
	QUEUE_DEF	*item;
	QUEUE_DEF	*next_item;

	DISABLE ();
	item = queue_head -> forward;
	if (item) {
		next_item = item -> forward;
		queue_head -> forward = next_item;
		if (next_item) {
			next_item -> backward = (QUEUE_DEF *) NULL;
		} else {
			queue_head -> backward = (QUEUE_DEF *) NULL;
		}
	}
	ENABLE ();
	return item;
}

/****
*
*	remove the given item from anywhere in the queue
*
****/

void queue_remove (QUEUE_DEF *queue_head, QUEUE_DEF *queue_item)
{
	QUEUE_DEF	*this_item;

	DISABLE ();
	this_item = queue_head -> forward;
	while (this_item) {
		if (this_item == queue_item) {
			if ((queue_item -> backward) && (queue_item -> forward)) {
				queue_item -> forward -> backward = queue_item -> backward;
				queue_item -> backward -> forward = queue_item -> forward;
			} else if (queue_item -> forward) {
				queue_head -> forward = queue_item -> forward;
				queue_item -> forward -> backward = (QUEUE_DEF *) NULL;
			} else if (queue_item -> backward) {
				queue_head -> backward = queue_item -> backward;
				queue_item -> backward -> forward = (QUEUE_DEF *) NULL;
			} else {
				queue_head -> forward = (QUEUE_DEF *) NULL;
				queue_head -> backward = (QUEUE_DEF *) NULL;
			}
			ENABLE ();
			return;
		}
		this_item = this_item -> forward;
	}
	ENABLE ();
	ERROR_TRACE ("queue_remove, item is not on queue");
}

/****
*
*	count the number of items in the queue
*
****/

int queue_count (QUEUE_DEF *queue_head)
{
	int			count;
	QUEUE_DEF	*this_item;

	count = 0;
	DISABLE ();
	this_item = queue_head -> forward;
	while (this_item) {
		++count;
		this_item = this_item -> forward;
	}
	ENABLE ();
	TRACE_I ("queue_count, returning count = %d", count);
	return count;
}

/****
*
*       search the queue for a match on the given queue item
*
****/

int queue_search (QUEUE_DEF *queue_head, QUEUE_DEF *queue_item)
{
	QUEUE_DEF	*this_item;

	DISABLE ();
	this_item = queue_head -> forward;
	while (this_item) {
		if (this_item == queue_item) {
			ENABLE ();
			return TRUE;
		}
		this_item = this_item -> forward;
	}
	ENABLE ();
	TRACE ("queue_search, item is not on queue");
	return FALSE;
}

