/*timLib.c,v 1.1.1.1 1999/08/03 14:44:59 tis4000 Exp */

/* timLib.c -- time-out library for VxWorks
 *
 *	9/20/91	@(#)timLib.c	1.2
 *
 * History:
 * 20jun90	ajk - original issue.
 * 090591	joh fixed timLibInit() declared void but still returns a value
 * 091191	added WAIT_FOREVER to semTake()
 * 03-31-94	ems	added semLib.h so compiler knows what semCreate returns
 *
 */

#include	<vxWorks.h>
#include	<wdLib.h>
#include	<semLib.h>
#include	"timLib.h"

LOCAL	LIST	timList;
LOCAL	WDOG_ID	timToWD;	/* Watchdog */
LOCAL	SEM_ID	timToSem;	/* Time-out semaphore */
LOCAL	SEM_ID	timLockSem;	/* Lock semaphore */

#define abort() taskSuspend(taskIdSelf())

/* Initiate time-out task */
int timLibInit(void)
{

	return taskSpawn("time-out", 10, 0, 2000, timTask);
}

/* Time-out task */
VOID timTask(void)
{
	TOUT	*pTim, *pPrevTim;
	ULONG	time;

	lstInit(&timList);

	/* Create internal semaphores */
	timLockSem = semCreate();
	semGive(timLockSem);
	timToSem = semCreate();
	semGive(timToSem);
	timToWD = wdCreate();

	for (;;)
	{
#		ifdef V5_vxWorks
			semTake(timToSem, WAIT_FOREVER);
#		else
			semTake(timToSem);
#		endif
		/* Watchdog caused us to wake up */
		time = tickGet();
		/* Lock list access */
#		ifdef V5_vxWorks
			semTake(timLockSem, WAIT_FOREVER);
#		else
			semTake(timLockSem);
#		endif
		pTim = (TOUT *)lstFirst(&timList);
		while (pTim != NULL)
		{
			if (time >= pTim->timeout)
			{	/* time-out expired */
				semGive(pTim->sem_id);
				pTim->timeout = 0;
				pTim->expired = TRUE;
				/* Remove this element from front of list */
				lstDelete(&timList, &pTim->timNode);
				pTim = (TOUT *)lstFirst(&timList);
			}
			else
				pTim = NULL; /* don't process rest of list */
		}
		/* Re-start watchdog */
		pTim = (TOUT *)lstFirst(&timList);
		if (pTim != NULL)
		{
			wdStart(timToWD, pTim->timeout - time,
			 semGive, timToSem);
		}
		/* Unlock list access */
		semGive(timLockSem);
	}
}

/* timCreate:  create a time-out element */
TIM_ID	timCreate(SEM_ID sem_id)	/* semaphore to be "given" at time-out */
{
	TIM_ID	tim_id;

	tim_id = (TIM_ID)malloc(sizeof(TOUT));
	if (tim_id != NULL)
	{
		tim_id->timeout = 0;
		tim_id->expired = TRUE;
		tim_id->sem_id = sem_id;
	}
	return tim_id;
}

/* timInit:  initialize a time-out element.
 * Caution - DON'T use timDelete in caller's structure */
VOID	timInit(tim_id, sem_id)
TIM_ID	tim_id;	/* user-supplied structure */
SEM_ID	sem_id;	/* semaphore to be "given" at time-out */
{
	tim_id->timeout = 0;
	tim_id->expired = TRUE;
	tim_id->sem_id = sem_id;
	return;
}

/* Start a timer */
VOID timStart(tim_id, timeout)
TIM_ID	tim_id;
ULONG	timeout;
{
	ULONG	time;
	TOUT	*pTim, *pPrevTim;
	
	/* If already in list, remove it *.
	if (tim_id->expired)
		timCancel(tim_id);

	/* Lock list access */
#	ifdef V5_vxWorks
		semTake(timLockSem, WAIT_FOREVER);
#	else
		semTake(timLockSem);
#	endif
	/* Insert node */
	time = tickGet();
	tim_id->timeout = timeout;
	if (timeout <= time)
	{	/* Already expired */
		semGive(tim_id->sem_id);
		tim_id->expired = TRUE;
	}
	else
	{	/* Place time-out element into linked list */
		tim_id->expired = FALSE;
		/* List is ordered by increasing time-out value */
		pPrevTim = NULL;
		pTim = (TOUT *)lstFirst(&timList);
		while (pTim != NULL)
		{
			/* if time is smaller than that in list, insert */
			if (tim_id->timeout < pTim->timeout)
				break;
			pPrevTim = pTim;
			pTim = (TOUT *)lstNext(&pTim->timNode);
		}
		lstInsert(&timList, &pPrevTim->timNode, &tim_id->timNode);
		/* Start the watchdog if necessary */
		semGive(timToSem);
	}
	/* Unlock list */
	semGive(timLockSem);

	return;
}

/* Cancel a time-out */
VOID timCancel(tim_id)
TIM_ID	tim_id;
{
	/* Lock access to list */
#	ifdef V5_vxWorks
		semTake(timLockSem, WAIT_FOREVER);
#	else
		semTake(timLockSem);
#	endif
	if (!tim_id->expired)
	{
		/* Delete element from list */
		lstDelete(&timList, &tim_id->timNode);
		tim_id->expired = TRUE;
	}
	/* Unlock list */
	semGive(timLockSem);
}

/* Delete a timeout element */
VOID timDelete(tim_id)
TIM_ID	tim_id;
{
	free(tim_id);
}

/* Find out if a time-out expired */
BOOL	timExpired(tim_id)
TIM_ID	tim_id;
{
	return tim_id->expired;
}

/* Time-out list querry */
VOID timQry(void)
{
	TOUT	*pTim;
	
	/* Lock list access */
#	ifdef V5_vxWorks
		semTake(timLockSem, WAIT_FOREVER);
#	else
		semTake(timLockSem);
#	endif

	/* insert node */
	pTim = (TOUT *)lstFirst(&timList);
	while (pTim != NULL)
	{
		printf("pTim=%d, sem_id=%d, timeout=%d\n",
		 pTim, pTim->sem_id, pTim->timeout);
		pTim = (TOUT *)lstNext(&pTim->timNode);
	}
	/* Unlock list */
	semGive(timLockSem);
	return;
}

